/*	Read new Lick camera image, descramble the image as	*/
/*	needed and place it into shared memory, and send	*/
/*	messages to image processing programs to do something	*/
/*	with the image.						*/ 
/*	camread [-200] [fiber_name]				*/

/*	If -200 is supplied on the command line it must come	*/
/*	before the fiber_name parameter.			*/
/*	-200 specifies the fast fiber.				*/
/*	fiber_name is translated by the services file into the	*/
/*	device name in /dev.					*/

/*	Copyright June, 2000 Richard Stover			*/
/*	University of California at Santa Cruz			*/
/*	Santa Cruz, CA 95064					*/

/*	richard@ucolick.org					*/

/*
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <pwd.h>
#include <sys/io.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/msg.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/select.h>
#include <signal.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <sys/ioctl.h>
#include <pthread.h>
#include <semaphore.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include "traffic.h"
#include "local_traffic.h"
#include "guiderxv.h"
#include "guider.h"
#include "pp.h"
#include "cyport.h"
#include "camera.h"
#include "notify.h"
#include "camread.h"
#include "sqlget.h"
#include "usbfo.h"

#ifdef CYP_DEBUG_INFO
#undef CYP_DEBUG_INFO
#define CYP_DEBUG_INFO 3
#endif

#define RCSID static char *rcsid __attribute__ ((unused))
RCSID = "$Id: camread.c,v 1.71 2010/05/18 19:26:13 ccdev Exp $";
/*
$Log: camread.c,v $
Revision 1.71  2010/05/18 19:26:13  ccdev
Fixed bugs in the image type 48 descrambling.

Revision 1.70  2010/05/11 22:17:49  ccdev
wait_for_image_to_guide creates the thread wait_for_guider_sock.
After the thread is created we wait 1 second. This didn't seem to
be needed before but under Fedora 12 on explorer this wait is needed
to allow the thread to get started for some reason. Without the wait
a segv happens. With the wait the segv does not happen. It is not
apparent where the segv is actually happening in the code.

Revision 1.69  2010/05/07 22:15:41  ccdev
Mods for type 48 geometry, 8-amp output.

Revision 1.68  2010/04/10 00:11:39  ccdev
Fixed the routine that handles the image flip notifier messages to
look for column0 and row0 instead of column and row.

Revision 1.67  2010/04/10 00:06:57  ccdev
Debug notifier image flipping messages.

Revision 1.66  2010/04/09 23:37:41  ccdev
Fixed case 5 of image descramble. It was not doing anything correctly.

Revision 1.65  2010/04/08 22:15:41  ccdev
fixed bug in log message.

Revision 1.64  2009/09/09 23:35:56  ccdev
Changed strndup to local_strndup to avoid conflict with an include
file definition of strndup.

Revision 1.63  2009/07/23 18:04:15  ccdev
Removed a number of diagnostic flogerror statements.

Revision 1.62  2009/06/16 21:36:01  ccdev
If we can't make a network connection using our hostname then try
localhost.

Revision 1.61  2009/06/12 17:40:04  ccdev
Log our initial request for memory so we can tell if we got all
we ask for initially.

Revision 1.60  2009/05/04 17:38:26  ccdev
Removed EPOCH comment card, which was wrong.

Revision 1.59  2009/04/23 23:49:11  ccdev
Fixed a couple bugs in the abort notification mechanism.

Revision 1.58  2009/04/05 22:13:09  ccdev
Issue a warning rather than an error message if we get a timeout
after the user does an Abort.

Revision 1.57  2009/03/31 19:01:04  ccdev
Added the save_error_status facility to post critical error messages
to the database. These messages can be posted to a popup window
by the user gui.

Revision 1.56  2009/03/28 22:34:24  ccdev
Write a status message to the database if an error occurs during
image readout.

Revision 1.55  2009/03/24 21:33:53  ccdev
Reduced the restart wait in re_camread from 5 seconds to 1 second.
I'm not sure any wait is needed.

Revision 1.54  2009/03/12 21:18:54  ccdev
Removed some debug statements.

Revision 1.53  2009/03/12 21:15:30  ccdev
Added hdata->type,"IMAGE READOUT GEOMETRY" to FITS header.

Revision 1.52  2009/02/11 19:59:18  ccdev
Added more checks for errors while waiting for image data.
Added a description of the child processes for the re_camread function.

Revision 1.51  2009/01/06 18:36:35  ccdev
Modified some flogerror statements and made sure all forks and threads
log their pid when they start.

Revision 1.50  2008/12/23 18:22:04  ccdev
Change the way the CRVALX values are computed to properly account
for the row or column overscan. With the new computation the
row and column numbers of the imaging area remain the same regardless
of the row or column overscan number.
Also a little cleanup.

Revision 1.49  2008/09/25 18:15:06  ccdev
Corrected bug that resulted in over-padding fits header and
another bug that resulted in over-padding the image.

Revision 1.48  2008/09/22 19:02:04  ccdev
Added date/time info and gui indentification to TrackingDialog messages.

Revision 1.47  2008/09/21 07:23:31  ccdev
Added column3 and row3 to image flip parameters to control the behavior
of FITS card generation.

Revision 1.46  2008/09/21 04:28:13  ccdev
Fixed typo which was leaving the timer interval used in the setitimer calls
uninitialized.

Revision 1.45  2008/09/17 17:47:16  ccdev
Removed some debugging statements.

Revision 1.44  2008/09/17 00:45:07  ccdev
Fixed math bug in address computations in four-quad (type 8) image
descramble.

Revision 1.43  2008/09/03 16:06:09  ccdev
Fixed typo.

Revision 1.42  2008/07/14 18:03:48  ccdev
Fixed typo.

Revision 1.41  2008/06/05 16:21:57  ccdev
Corrected a typo in a printf statement.

Revision 1.40  2008/05/13 22:45:25  ccdev
Added periodic image display update messages.

Revision 1.39  2008/04/08 20:04:59  ccdev
Include latest additions from LAMOST team (mostly case 50 descramble).

Revision 1.5  2008/02/25 18:11:31  cvs
add codes to adjust auto baseline. add gmedian to dtakeservice file.

Revision 1.4  2008/02/22 00:06:43  cvs
Initialize old timer value to prevent error in later setting of new timer
value.

Revision 1.3  2008/02/07 00:05:50  cvs
Added parameter "port" for connect_to_notifier function.

Revision 1.2  2008/02/01 23:23:31  cvs
Add disablegui procedure
Add 4 read out modes

Revision 1.1.1.1  2007/12/19 18:22:47  cvs
Initial checkin

Revision 1.37  2007/11/16 01:13:45  ccdev
Explicitly set the file mode for new image file to make sure the sticky
bit is turned off. The sticky bit is used by the Mt. Hamilton archiving
system to indicate that a file has been written to the archive, so the
bit needs to be off for a new file. (The sticky bit might be on if
we are in fact overwriting an old image on which the sticky bit was
previously set.)

Revision 1.36  2007/09/14 23:31:22  ccdev
Added extra flogerror message about not being able to record image.

Revision 1.35  2007/09/11 21:34:06  ccdev
Made a separate function from the code in update_fits_cards() that
acturally creates a FITS card image. This new routine, fits_card_image(),
is used by some of the camread loadable modules.

Revision 1.34  2007/07/18 21:44:19  ccdev
Removed debug statement.

Revision 1.33  2007/07/18 21:37:56  ccdev
Loosened permission on created semaphore.

Revision 1.32  2007/07/18 21:16:53  ccdev
Development for image geometry codes 8, 9, and 10.

Revision 1.31  2007/07/07 00:06:08  ccdev
Mods to comments.

Revision 1.30  2007/07/05 20:17:56  ccdev
Support for various configurations of four amplifier readout.

Revision 1.29  2007/07/02 20:27:27  ccdev
Include the overscan pixels in the calculation of the image origin
pixel number.

Revision 1.28  2007/06/20 20:34:48  ccdev
Increased maximum row or column size.

Revision 1.27  2007/04/19 21:48:31  ccdev
Added code to get sensor number of rows and columns from the
telefocus.configs database and then use these numbers as necessary
to compute the row and/or column original of the image.

Revision 1.26  2007/03/02 22:39:31  ccdev
Use strdup_envar() to translate log filenames read from the database
which may have the @VAR@ constructs in them.

Revision 1.25  2007/02/22 19:54:50  ccdev
Call send_progress_report with the total number of pixels in the raw
image rather than the total number of pixels in the saved image.

Revision 1.24  2007/02/07 21:27:59  ccdev
Fixed bug in update_fits_card() in case of a character string parameter
where the generated FITS card was 82 characters long (for an 80-character
buffer!!!).

Also call a routine to translate atXXXat (at=@) strings in the
module file names retrieved from the database.

Also added missing comment string ending characters (* /) in
one section of code.

Revision 1.23  2006/12/20 22:23:00  kibrick
Move creation of semaphore pair into top of camread prior to forking the
subprocesses, and explicitly set both semaphores to zero prior to the forks.

Revision 1.22  2006/12/20 21:11:51  ccdev
Fixed a couple bugs from last commit.   rjs

Revision 1.21  2006/12/20 21:01:42  ccdev
Added second semaphore for process synchronization.

Revision 1.20  2006/12/20 01:12:35  ccdev
Added a semaphore to synchronize the image reading process and the
image descramble process.

Revision 1.19  2006/12/18 23:52:58  kibrick
There is a nasty race condition which shows up on faster machines and which
requires a nanosleep delay of 1 millisecond.  Without this delay, reading
of the next image steps on the byte count used for processing the previous
image.

Revision 1.18  2006/11/06 19:32:35  sla
sync with CCDLAB

Revision 1.58  2006/09/14 23:30:14  ccdev
Removed an obsolete reference to grecord and substituted reference to
__FUNCTION__ instead.

Revision 1.57  2006/08/24 18:35:39  ccdev
Modified so the recorded image properly uses the image flip data
in constructing the FITS header, and fixed fits_cards() to correctly
use the binning to derive the correct CRVAL1 and CRVAL2 values from
the image header data.

Revision 1.56  2006/08/24 00:05:38  ccdev
Added code to retrieve image binning information from the header
if it is provided. Also fix the image header decoding so it properly
calculates the values that will become CRVAL1 and CRVAL2 in the FITs
headers. Also, fix a bug in retrieving the binning from FITS cards.
This is needed for older versions of the camera software which do
not pass the binning in the header.

Revision 1.55  2006/07/19 19:06:15  ccdev
Modified simulated reads to ignore image flipping, which just
confuses things when one is trying to debug other software.
Also did a little additional cleanup of shared memory segments.

Revision 1.54  2006/07/19 00:28:30  ccdev
Increased size of image simulation buffer to handle 4Kx4K image simulation.
Put properly formatted END card at end of FITS header buffer.
Disconnect from controlling tty.

Revision 1.53  2006/07/12 18:31:54  ccdev
Moved FITS header to be at the beginning of the descrambled image
instead of at the end to make it possible to hand this image
directly to ds9.

Revision 1.12  2006/07/07 22:50:55  sla
updates from prober

Revision 1.52  2006/06/20 16:52:17  ccdev
Added code to support the telefocus.configs.%version.iflag database
value, which defines for the USB interface which fiber channel to use.

Revision 1.51  2006/06/15 23:22:49  ccdev
Added checks for FITS keywords longer than 8 characters.
Also reset the pointer to the FITS_OTHER card buffer in function
fits_cards() so that the search for keyword BIN2 would work correctly.
Also added a check for duplicate FITS keywords as they are being read
from the database.

Revision 1.50  2006/06/05 18:09:46  ccdev
Lots of code added for background subtraction of guider images.
Also fixed counting error in descramble() routine readout_geometry
values 6 and 7.

Revision 1.10  2006/05/31 20:11:02  sla
sync with CCDLAB

Revision 1.49  2006/04/11 17:23:12  ccdev
Added AMPSROW and AMPSCOL keywords to FITS header to record number
of amplifiers per row and column.

Revision 1.48  2006/04/10 21:47:46  ccdev
Added OBSNUM fits keyword to FITS header.

Revision 1.47  2006/03/24 20:28:54  ccdev
Fixed descramble geometries 2 and 3.   rjs

Revision 1.46  2006/03/14 22:34:17  ccdev
Added hack of allocating a block of memory and then freeing it
in order to grow our process size. Without this the dynamic linker
takes a segv on some systems.
Also added the -c command line option to record any images with a
checksum error.

Revision 1.45  2006/02/03 23:34:33  ccdev
Changed #if 0/#endif to a comment to avoid gcc complaints about
unbalanced single quote character.

Revision 1.44  2006/01/19 01:05:18  ccdev
In wait_for_image_to_record() we now allocate a buffer for
swapping the image bytes before we try to lock ourselves in
memory. In pre-FedoraCore3 linux the lock would fail if we aren't
root. In FC3 the lock succeeds. But when locked in core it is
not possible for the process to grow any further. So allocating
a swap buffer later (in record_image()) fails.

Revision 1.43  2006/01/09 18:29:13  ccdev
Added VERSION to initial FITS cards.

Revision 1.42  2006/01/05 22:59:46  ccdev
Now retrieve the version name before attempting to get first
log filename because the filename depends on the version.

Revision 1.41  2005/09/02 23:16:51  ccdev
Moved a logrecord() call so it correctly reflects the current
record count.

Revision 1.40  2005/09/02 00:07:58  ccdev
Added -i option to specify "instrument mode." In instrument mode we do
not fork a new process to coordinate with an external guider process
and we do not issue the GUIDER_NEWIM message.

Revision 1.39  2005/09/01 17:41:32  ccdev
Removed a single quote from the previous log messages because
gcc version 3.2.2 20030222 (Red Hat Linux 3.2.2-5) chokes.
Also had to throw in a blank line in the text outside the change Log
because changes inside the Log are not considered changes to the file
and so ci would not check it back in without some sort of change.

Revision 1.38  2005/09/01 17:37:04  ccdev
Removed include of read_progress.h, which is not needed.

Revision 1.37  2005/08/31 20:55:06  ccdev
Added capability to report readout progress via music messages.
Additional error checking in the image writing routine.

Revision 1.36  2005/08/14 16:57:38  ccdev
Removed a short wait. (Should not be needed.)
Removed explicit definition of dbcmd pathname.

Revision 1.35  2005/07/29 23:39:04  ccdev
Now use update_fits_card() routine to save checksum FITS cards.
Had to add format type 'T' (logical) to update_fits_card().

Revision 1.34  2005/07/29 23:05:14  ccdev
Added checksum information to basic FITS header creation.

Revision 1.33  2005/07/29 22:24:24  ccdev
Format for DATE_OBS keyword corrected.

Revision 1.32  2005/07/28 23:08:39  ccdev
Initialize two instances for run-time allocated 'struct tuple value[2]'.

Revision 1.31  2005/07/28 22:05:03  ccdev
Modified get_keword_list to use for() loop instead of while() loop because
tuplep needs to be incremented on each loop. This was not happening if an
error occurred or it the version list for a keyword did not match the
current version.

Revision 1.29  2005/07/22 21:28:16  ccdev
Removed checks for only one pixel to read before wait_for_data()
because the bug in the camera that may cause this condition is fixed.

Revision 1.28  2005/05/24 21:57:58  ccdev
Now make sure that all instances of camread that need to swap
btyes for writing FITS images actually have a swap buffer.

Revision 1.27  2005/05/13 23:56:14  ccdev
Actually added needswab() function.

Revision 1.26  2005/05/13 23:42:59  ccdev
Now allocate a buffer for swapping FITS image bytes if needed and do
this early in the process before the process is locked in memory.
Once locked in memory the program cannot grow further.
Added the needswab() function to dynamically determine if swapping
is needed or not.
Also enhanced several flogerror messages.

Revision 1.25  2005/04/01 00:27:05  ccdev
Fixed a couple instances if variables not having the correct cast.

Revision 1.24  2005/03/15 23:42:12  ccdev
Now uses one big image buffer (possibly increased in size by
command line arguments) to buffer image. The buffer acts by a
big circular fifo.
Lots of mods to better support setuid root operation with access
to the database (which does not have root as a known host).
Now use sqlget for faster direct access to database and dbshow
in only a few places.

Revision 1.23  2005/03/03 22:38:57  ccdev
Modified to do checksum calculations and much more direct (and faster)
access to the database. The faster access db access makes startup
much faster.

Revision 1.22  2005/02/08 18:59:04  ccdev
Cleaned up many function prototypes, unused variables, obsolete code
and other -Wall warnings.

Revision 1.21  2004/05/05 22:35:26  ccdev
Now use table named teleposition instead of position. position was a
reserved word.

Revision 1.20  2004/03/16 23:54:14  ccdev
Changed DATE-OBS for the START OF OBSERVATION to DATE-STA. There was
already at DATE-OBS which records the beginning of readout time.

Revision 1.19  2003/12/18 19:02:31  ccdev
Removed a debug statement.

Revision 1.18  2003/12/16 19:51:07  ccdev
Now writes FITS images with BSCALE and BZERO to correctly  handle
images with a range of 0 to 65535.

Revision 1.17  2003/12/16 17:52:13  ccdev
Comment out call to logdecoded().

Revision 1.16  2003/12/15 18:36:41  ccdev
Added the 'int' specification to the static m[] array.

Revision 1.15  2003/12/02 17:40:34  ccdev
Corrected one typo.

Revision 1.14  2003/11/27 00:44:49  ccdev
Now get the fiber port device name from the database, based on the
version of the software running.

Revision 1.13  2003/11/25 01:05:17  ccdev
if ioctl(im_fd,CYP_WAIT_READ,&nr) fails just log it and pause, waiting for
parent process to restart everything.

Revision 1.12  2003/11/24 22:41:13  ccdev
Tries to reconnect to the image reading port. This makes it possible for
a USB connection to be unplugged and plugged back in without manually
restarting the guider software.

Revision 1.11  2003/11/21 23:54:15  ccdev
Corrected several bugs in the FITS cards for DATE-OBS and related.

Revision 1.10  2003/11/19 23:55:13  ccdev
Added DATE-OBS to the basic FITS header.

Revision 1.9  2003/11/19 22:14:25  root
Added double buffering for incoming images and manipulation of
the environment so that 'ps alx' shows different args for each fork
of the program.

Revision 1.8  2003/10/15 17:05:59  ccdev
Redefine IFCMD_200MHZ only if it is not already defined.

Revision 1.7  2003/10/15 16:50:32  ccdev
Many modifications to handle image flipping while descrambling.

Revision 1.6  2003/06/18 22:39:06  ccdev
Modified to send a music broadcast about a newly recorded image no
more than 5 times per second to avoid flooding the user gui with
too many messages. This also required catching the EINTR error in
a couple places and restarting the interrupted (by the timer) system
call.

Revision 1.5  2003/04/29 19:39:06  ccdev
Added image simulation capability.

Revision 1.4  2003/03/08 00:52:23  ccdev
Now reads version and telescope position port from database.

Revision 1.3  2003/03/03 22:00:58  ccdev
Support for larger header and the additional image origin information
in the header. Support for putting RA, Dec and HA into fits header.
A couple bugs corrected.

Revision 1.2  2003/01/03 19:24:45  ccdev
Lots and lots of development of FITS parameters, image writing, and
code to select different fiber ports.

Revision 1.1  2002/04/24 22:09:52  ccdev
Initial revision

Revision 1.1  2002/02/14 00:18:23  ccdev
Initial revision

Revision 1.5  2002/01/17 21:48:12  ccdev
Mods to handle more complex image header.
Mods to handle more complex image descrambling.

Revision 1.4  2002/01/03 01:00:58  ccdev
Added code for fast fiber selection and for setting the max
row and column size and for getting the header size from the header
itself.

Revision 1.3  2001/05/10 21:37:35  ccdev
Get device name from dtakeservice file under label guider_fiberport
in the "network" portion of the dtakeservice file.

Revision 1.2  2001/05/10 18:37:33  ccdev
Locks process into memory.
WCMD ioctls now pass int instead of char.

Revision 1.1  2000/08/22 18:41:01  ccdev
Initial revision

*/

extern char *log_filename;
int im_fd;			/* Descriptor to fiber interface.	*/
int shid;			/* Shared memory ID for descrambled mem	*/
char *shad;			/* Address of descrambled memory buffer	*/
int shid2;			/* Shared memory ID for raw data buffer	*/
char *shad2;			/* Address of raw data shared memory	*/
unsigned short *image;		/* Same as shad2, cast as (short *)	*/
unsigned short *rimage;		/* Either shad2 or shad2a, used by the	*/
				/* image descramble routine.		*/
char *mem_image;		/* Same as shad, used in descrambling	*/

char *process_image;		/* Address of image to be used for	*/
				/* subsequent processing.		*/
char *process_header;		/* Address of header associated with	*/
				/* the process_image.			*/
char *process_buffer;		/* Start of the shared memory buffer	*/
				/* containing process_image and		*/
				/* process_header.			*/
int process_shid;		/* Shared memory ID of process buffer,	*/
				/* either shid or bkNormShid.		*/

int shid3;			/* FITS header shared memory		*/
char *mem_fits;			/* Address of FITS shared memory	*/
int shid4;			/* Shared memory ID for header data	*/
char *head_data;		/* Shared memory address of header data	*/
int id,r,c;
short *header;			/* Start of FITS header for descrambled	*/
				/* image.				*/
int maxheadcards;		/* Maximum number of 80-character cards */
				/* possible in the header buffer.	*/
int total_err;
int msid = -1;
int record_idnum = 0;
int record_count = 0;
char object_name[256];
char file_name[256];
int display_interval = 0;	/* If non-0 then a new DISPLAY_NEWIM	*/
				/* message is send every		*/
				/* display_interval milliseconds during	*/
				/* a readout.				*/

int camera_id = -1;		/* Will be set to the camera's ID.	*/
				/* Numerous FITS cards have values	*/
				/* indexed in the database by camera ID.*/

extern int merrno;              /* msgio error codes            */
extern char msg_err_txt[];      /* Error text corresponding to  */
                                /* value of merrno.             */
extern char tgofork_err[];      /* Error messages from tgofork  */
int traffic_fd = -1;            /* fd to traffic controller     */
int traffic_ok = 0;
char our_name[40];              /* Name t.c. knows us by.       */
int our_addr;                   /* Our traffic address.         */
int notifier_port = 2002;	/* Notifier port number. Can be	*/
				/* modified by -n command argument.	*/

int tremoteopen(char *ourname,char *remotehost,int *trafficfd);
void connect_to_traffic(char *proc);
void flogerror(char *f, ...);
char *loglasterror(void);
static void set_dbuid(void);
int ConnectGcam(char **devicename);
void wait_for_image_to_descramble(int msid);
void wait_for_image_to_display(int msid);
void wait_for_image_to_record(int msid);
void wait_for_image_to_guide(int msid);
void wait_for_fits(void);
int read_image(int sem_read);
int send_newim_msgs(int msid);
int send_msgs_list(int msid, int m[]);
int send_display_newim_msg(int msid);
void swab(const void *from, void *to, ssize_t n);
static void send_progress_report(int total_pixels,int current_pixels);
static void process_record_msg(char *msg);
static void process_record_newim_msg(char *msg);
char *strdup_envar(char *item);
void fits_card_image(char p,char *c,void *param,char *keywrd,char *comm);

int report_progress = 0;
int report_progress_msg = 600;

int icnt= 0;
int double_buf = 1;		/* Set if we are doing double	*/
				/* buffered image reads.	*/
int fastfiber = 0;		/* Set if we should read from	*/
				/* the fast fiber.		*/
char fiberport[64];		/* Name of the fiber port. This	*/
				/* name is retrieved from the	*/
				/* database and depends on	*/
				/* the version of the guider.	*/
char version[32];		/* Our version name		*/
int telpos_port = -1;		/* Telescope position port	*/
				/* number, if >= 0.		*/
int simulation = 0;		/* Specifies how to simulate	*/
				/* the camera data (0 == none).	*/
int interface_flag = 0;         /* Tells us how to configure and*/
                                /* use the fiber interface. This*/
                                /* is useful only with the dual */
                                /* channel USB interface.       */
int guider_mode = 1;		/* If set then operate in guider*/
				/* mode. If not set (by -i argv)*/
				/* then operate in instrument	*/
				/* mode.			*/
int record_unrecorded = 0;	/* Set to cause unrecorded	*/
				/* images to actually be record-*/
				/* ed in file unrecorded_file.	*/
char unrecorded_file[256];	/* Destination for unrecorded	*/
				/* images.			*/

int send_guider_message(int sock,int r, int c);
char *check_msg(int msid,int msg_type);
char *check_msg_wait(int msid,int msg_type);
static char *interface_meaning(int val);
static void interface_setup(int val);
static char *getlogfile(char *process_name);
static char *next_token(char *token_string);
static char *local_strndup(char *string,int nc);
char *nextelement(char *cp);
char *lindex(char *list,int item);
static void RetrieveBinning(int id,int *Cdelt,int *Rdelt);
int find_fits_card(char *fitsbuf,int ncards,char *keyword,char **card);
void update_fits_card(int imageid,char *keyword,char ptype,
	void *param,char *comment);

#define RECORD_MSG 332
#define GUIDER_MSG 322
#define GUIDER_NEWIM 312
#define RECORD_NEWIM 302
#define DISPLAY_NEWIM 342

#define HEADER_SIZE 52		/* Number of bytes transmitted in	*/
				/* normal camera header.		*/
int shm_nbytes;			/* This is the size of the shared	*/
				/* memory allocated for the descrambled	*/
				/* image.				*/
int nbytes;			/* This is the size of the shared	*/
				/* memory allocated for reading raw	*/
				/* image data.				*/
//int maxrc = (4608+512);
int maxrc = (8196+512);

struct window {
	int r;		/* number of rows		*/
	int c;		/* number of columns		*/
	int rorg;	/* row origin			*/
	int corg;	/* column origin		*/
};

struct header_data {
	int type;	/* Image geometry code		*/
	int id;		/* Unique image ID		*/
	int checksum_flag;	/* If set the the camera has	*/
				/* supplied a checksum.		*/
	unsigned short cam_checksum; /* Camera computed checksum */
	unsigned short im_checksum; /* Our computed checksum */
	int size;	/* Number of bytes in the header*/
	int exposure;	/* Exposure time (10ms units)	*/
	int shutter;	/* Shutter state code		*/
	int amps_per_row; /* Number of amplifiers supplying	*/
			  /* data for each row.			*/
	int amps_per_column; /* Number of amplifiers supplying	*/
			     /* data for each column.		*/
	struct window per_amp;	/* Per amplifier values	*/
				/* from header.		*/
	struct window image;	/* Overall image size	*/
	struct window window;	/* Desired window within*/
				/* the overall image.	*/
	struct window over;	/* Overscan data	*/
	struct window final;	/* Final image size combining	*/
				/* struct window window and	*/
				/* struct window over.		*/
/*	Since version 3.00 of the camera software the binning	*/
/*	data is passed in the image header.			*/
	int row_bin;		/* Image binning in rows.	*/
	int col_bin;		/* Image binning in columns.	*/
};

#ifndef __IMAGE_FLIPS
/*	Flags to invoke additional flips in rows or columns when	*/
/*	the image is descrambled.					*/
struct image_flips {
	int row_flip;	/* non-0 means a row flip should happen		*/
	int col_flip;	/* non-0 means a column flip should happen	*/
	int row;	/* row_flip = row && (row1 ^ row2)		*/
	int row1;
    	int row2;
	int row3;
	int column;	/* col_flip = column && (column1 ^ column2)	*/
	int column1;
	int column2;
	int column3;
	int geometry;
};
#endif /* __IMAGE_FLIPS */

static int flip_options(struct image_flips *image_flips);
void get_flip_data_init(struct image_flips *image_flips);
int get_flip_data(void *dp);
int get_videoconfig(void *dp);
void recompute_flips(struct image_flips *image_flips);
int saveenv(char **argv);
void reset_argv(char *string);

static void decode_header(short *image,struct header_data *hdata);
void logdecoded(struct header_data *hdata);
void logbytes(void *data,int ndata);
int descramble(unsigned short *image,struct header_data *hdata);
void simulate_image(int simulation, int im_type,int headersize,int r,int c);
int fits_cards(struct header_data *,char *,struct image_flips *);
static void re_camread(pid_t children[],char *child_desc[],int n_children);
static void zswab(const void *imaddr,void *imaddr_swapped,int nshorts);
static void zflip(void *imaddr,int nshorts);
int get_cam_portnum(void);
int read_remote(char *host,char *service,char *object,int o_len);
int get_image_param(short *addr);
int record_image(unsigned short *imaddr,struct header_data *hdata,char *file);
int wait_for_data (char *ckpt,int pn);
void logrecord(char *func,int id,struct record *record);
int mstart(int to, int from, int msg);
int mput(char *item, int itemsize);
int mputl(int item);
int msend(int traffic_fd);
int makeportsocket(int portnum);
void notify_parse(struct notify_handle *nh);

static unsigned short sim_im[4200*4200];

static char *imaddr_swapped;		/* Buffer for swapped bytes for	*/
					/* FITS image.			*/
static int bytes_allocated = 0;		/* Number of bytes allocated for*/
					/* swapped image.		*/
static int header_size = (26*2);	/* Size of image header.	*/
static int row_size = -1;		/* Number of rows in the sensor	*/
					/* as read from the database.	*/
					/* -1 means it is undefined.	*/
static int col_size = -1;		/* Number of columns in the	*/
					/* sensor as read from the	*/
					/* database. -1 means it is	*/
					/* undefined.			*/

struct keydef *all_keywords = NULL;
/*	Set this by command line arg if recording of checksum error	*/
/*	images is needed.						*/
static int record_checksum_error_image = 0;

/*  A single FITS card in a list of cards.				*/
struct fitscard {
    char *card;		/* Pointer to 80-coumn card image		*/
    struct fitscard *fnext; /* Pointer to next fits card in a list	*/
};

/*	List of FITS cards.						*/
struct fitslist {
    int ncards;		/* Number of cards in the list			*/
    struct fitscard *flist; /* Pointer to head of list of cards.	*/
};

/*	Background subtraction data structures and buffers.		*/
static int last_was_background = 0;	/* Set non-zero if last image	*/
					/* was background.		*/
static char *bkAccShad;			/* Shared memory buffer address	*/
					/* for bkAccImg and bkAccHeader.*/
static char *bkAccImg;			/* Pointer to accumulated	*/
					/* background image.		*/
static char *bkAccHeader;		/* Pointer to raw header for	*/
					/* accumulated image.		*/
static char *bkNormShad;                /* Shared memory buffer address */
                                        /* for bnNormImg and            */
                                        /* bkNormHeader                 */
static char *bkNormImg;			/* Pointer to normalized back-	*/
					/* ground image.		*/
static char *bkNormHeader;		/* Pointer to raw header for	*/
					/* normalized image.		*/
static int bkCnt=0;			/* Number of accumulated	*/
					/* background images.		*/
static int bkAccShid;			/* bkAccImg shared memory ID.	*/
static int bkNormShid;			/* bkNormShad shared memory ID.	*/
static int subtractBackground = 0;	/* Set to 1 if doing background	*/
					/* subtraction.			*/
static int image_aborted = 0;		/* Set to 1 if a exposure is	*/
					/* aborted.			*/

/*	Background processing functions.				*/
int image_is_background(struct header_data *hdata);
void save_background(struct header_data *hdata);
void get_background_data_init(void);
int get_backgroundconfig(void *dp);
static void subtract_background(char *mem_image,struct header_data *hdata);
float headers_match(struct header_data *hdata,struct header_data *bk);
int check_for_abort(void *dp);

/*	Database handling functions:					*/
int get_record_dir(void *dp);
int get_record_root(void *dp);
int get_record_suffix(void *dp);
int get_record(void *dp);
int get_record_number(void *dp);
int get_obstype(struct notify_handle *nh);
static void get_keyword_list(void);
static void free_key(struct keydef *key);
static void add_key(struct keydef *key);
static int version_match(char *test_version);
struct keydef *find_keydef(char *keyname);
struct keydef *find_keyword(char *keyword);
static int handle_notify_module(void *dp);

/*	sql retrieval functions						*/
void exit_sql(void);
int dbget(char *format,char *dbitem,void *item);
int dbshow(char *format,char *dbitem,void *item);
void save_error_status(char * status);
void init_version_data(void);

/* These semaphores shared between camread and wait_for_image_to_descramble */
int sem_read;

/*	For saving the command arguments.				*/

#define MAX_SAVED_ARG 10
static char *saved_arg[MAX_SAVED_ARG+1];

/***********************************************************************/

int main(int argc,char *argv[])
{
	char *get_cam_hostname(char *host);
	char *devicename;
	pid_t descram_child,record_child,display_child,guider_child;
	pid_t fits_child;
	pid_t children[10];
	char *child_desc[10];
	int n_children=0;
	int ret;
	void sigcatch(int sig);
	struct shmid_ds shbuf;
	int timeout = 100;		/* 100 millisecond timeout	*/
	char *optstring = "i::r::2::d::t::m:n:u:";
	int c;
	int memsize;
	int umemsize=0;
        extern char *optarg;
        extern int optind, opterr, optopt;
	void *get_bigger;
	void disconnect();
	int sem_ret;
#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
        /* union semun is defined by including <sys/sem.h> */
#else
        /* according to X/OPEN we have to define it ourselves */
        union semun {
         int val;                  /* value for SETVAL */
         struct semid_ds *buf;     /* buffer for IPC_STAT, IPC_SET */
         unsigned short *array;    /* array for GETALL, SETALL */
                                   /* Linux specific part: */
         struct seminfo *__buf;    /* buffer for IPC_INFO */
        };
#endif
	union semun semun; 

/*	Disconnect from controlling terminal /dev/tty.			*/
	disconnect();

/*	Save command line arguments					*/
	for(c=0; c<MAX_SAVED_ARG; c++) {
	    if(c == argc) break;
	    saved_arg[c] = strdup(argv[c]);
	}

/*	Define routine to catch SIGHUP signal			*/
	signal(SIGHUP,sigcatch);

/*	Make ourself the head of a new process group.		*/
	setpgrp();

/*	First get the version.						*/
	ret = dbshow("%s","telefocus.variables.version.value",version);
	if (ret == -1)
	    strcpy(version,"cat");	/* Pick a default		*/

/*	Define our log filename.					*/
	log_filename = getlogfile("camread");

	while((c=getopt(argc,argv,optstring)) >= 0) {
	    switch(c) {
	    case 'i':
		guider_mode = 0;	/* Instrument mode.		*/
		break;
	    case 'r':
		report_progress = 1;	/* Make readout progress reports*/
		break;
            case '2':
		if(optarg && (strcmp(optarg,"00") == 0)) {
		    fastfiber = 1;
		}
                break;
            case 'd':
		if(optarg && (strcmp(optarg,"ouble")) == 0) {
		    double_buf = 2;
		}
                break;
            case 't':
		if(optarg && (strcmp(optarg,"riple")) == 0) {
		    double_buf = 3;
		}
                break;
	    case 'c':		/* Record images with checksum errors	*/
		record_checksum_error_image = 1;
		break;
	    case 'n':
		if(optarg) {
		    int p;
		    if(sscanf(optarg,"%d",&p) == 1) {
			notifier_port = p;
		    }
		}
		break;
	    case 'm':
		if(optarg) {
		    int nc,factor;
		    nc = strlen(optarg)-1;
		    switch(*(optarg+nc)) {
		    case 'M':
			factor = 1024*1024;
			*(optarg+nc) = '\0';
			break;
		    case 'K':
			factor = 1024;
			*(optarg+nc) = '\0';
			break;
		    default:
			factor = 1;
		    }
		    if(sscanf(optarg,"%d",&memsize) == 1) {
			umemsize = factor*memsize;
		    } 
		}
		break;
	    case 'u':
		if(optarg) {
		    if(sscanf(optarg,"%d",&display_interval) != 1) {
			display_interval = 0;
			flogerror("Bad (non-integer) display interval: %s",
				optarg);
		    }
		    if(display_interval < 0) {
			display_interval = 0;
			flogerror("Bad (negative) display interval: %s",
				optarg);
		    }
		} else {
		    flogerror("Missing display interval for -u option");
		}
		break;
            case '?':
                break;
            }
        }

/*	Increase our priority to the max.			*/
	ret = setpriority(PRIO_PROCESS,0,-20);
	if (ret) {
	    flogerror("Could not set high priority: %s",strerror(errno));
	}

/*	Having given ourselves high priority we now set our UID to	*/
/*	something known to the database.				*/

	set_dbuid();

/*	Read from the database all of the version-dependent information	*/
/*	we need.							*/
	init_version_data();

#ifdef FIBER_BY_ARG
	if(argv[optind]) {
		strcpy(fiberport,argv[optind]);
	} else {
            strcpy(fiberport,"guider_fiberport");
        }
#endif

/*	Increase our process size by allocating a big block of	*/
/*	memory and then freeing it again. On some older linux	*/
/*	systems this keeps the dynamic library loader from doing*/
/*	a segv fault. I don't know why this is the case, but	*/
/*	this hack seems to work.  rjs				*/

	get_bigger = malloc(20*1024*1024);
	free(get_bigger);

	printf("fastfiber=%d double_buf=%d fiberport=%s\n",
		fastfiber,double_buf,fiberport);
	flogerror("camread pid %d",getpid());

/*	Save the environment so we can use the space to change	*/
/*	what is seen with a ps command.				*/
	(void)saveenv(argv);

/*	Get a big block of shared memory. The maximum image	*/
/*	size we expect to see (rows or columns) is given by	*/
/*	maxrc.							*/

/*	The block of memory is for the descrambled image.	*/
	nbytes = sizeof(short)*maxrc*maxrc;
	if(umemsize > 0) {
	    nbytes = umemsize;
	    flogerror("Initial command-line request for memory = %d bytes.",
		nbytes);
	} else {
	    flogerror("Initial default request for memory = %d bytes.",
		nbytes);
	}

	for(shid = -1; shid == -1; ) {
	    shid = shmget(2,nbytes,IPC_CREAT|0777);
	    if(shid == -1) {
		nbytes -= 1024;
		if(nbytes < 1024*1024)
			break;
	    } else
		break;
	}
	if(shid == -1) {
		flogerror("Can't get %d bytes of descramble shared memory (%s)",
			nbytes,strerror(errno));
		exit(0);
	} else
		flogerror("Got %d bytes of shared memory",nbytes);

	shm_nbytes = nbytes;

#ifdef SMH_LOCK
	shmctl(shid,SHM_LOCK,&shbuf);
#endif
	shad = shmat(shid,(char *)0,0);
	if(shad == (char *)-1) {
	   flogerror("shmat(%d,0,0) failed:%s",shid,strerror(errno));
	   exit(0);
    	}
	shmctl(shid,IPC_RMID,&shbuf);

/*	Get a big block of memory for the raw image data.	*/
	if(umemsize > 0) {
	    nbytes = double_buf*umemsize;
	} else {
	    nbytes = double_buf*sizeof(short)*maxrc*maxrc;
	}

	for(shid2 = -1; shid2 == -1; ) {
	    shid2 = shmget(3,nbytes,IPC_CREAT|0777);
	    if(shid2 == -1) {
		nbytes -= 1024;
		if(nbytes < shm_nbytes)
			break;
	    } else
		break;
	}
	if(shid2 == -1) {
	    flogerror("Can't get %d bytes of 1st raw image shared memory (%s)",
		nbytes,strerror(errno));
	    exit(0);
	}
#ifdef SMH_LOCK
	shmctl(shid2,SHM_LOCK,&shbuf);
#endif
	shad2 = shmat(shid2,(char *)0,0);
	shmctl(shid2,IPC_RMID,&shbuf);

/*	Get a another big block of memory for building the FITS headers	*/
/*	and the recording structures.					*/
/*	We request enough memory to create 100 80-column card images	*/
/*	for general data, 100 cards for start of exposure data and	*/
/*	100 cards of end of exposure data for 256 different image IDs.	*/
/*	Round everything up to multiples of 1024 bytes.			*/
#ifndef NID
#define NID 256
#define NOTHER 100
#define NST 100
#define NEN 100
#define NCTOTAL (NID*NOTHER*80)
#define NCSTART (NID*NST*80)
#define NCEND   (NID*NEN*80)
#define NRECORD (NID*sizeof(struct record))
#define NTIMEV  (NID*sizeof(struct timeval))
#define NFLIP   (sizeof(struct image_flips))
#define FITS_MEM \
   (((1023+(NCTOTAL+NCSTART+NCEND+NRECORD+NFLIP+NTIMEV))/1024)*1024)
#endif  /* NID */

	shid3 = shmget(6,FITS_MEM,IPC_CREAT|0777);
	if(shid3 == -1) {
		flogerror("Can't get %d bytes of FITS records shared memory (%s)",
		    FITS_MEM,strerror(errno));
		exit(0);
	}
#ifdef SMH_LOCK
	shmctl(shid3,SHM_LOCK,&shbuf);
#endif
	mem_fits = shmat(shid3,(char *)0,0);
	memset(mem_fits,0,FITS_MEM);

#ifndef RECORD_ADDR
#define RECORD_ADDR(id) (struct record *)(mem_fits+sizeof(struct record)*id)
#define FITS_START(id) ((char *)RECORD_ADDR(NID)+(NST*80*id))
#define FITS_END(id) (FITS_START(NID)+(NEN*80*id))
#define FITS_OTHER(id) (FITS_END(NID)+(NOTHER*80*id))
#define FITS_TIME(id) (struct timeval *)(FITS_OTHER(NID)+(sizeof(struct timeval)*id))
#define FLIP_ADDR (struct image_flips *)(FITS_TIME(NID))
#endif

/*	Allow this shared memory segment to be destroyed when the	*/
/*	last process detaches from the memory. This happen on exit.	*/

	shmctl(shid3,IPC_RMID,&shbuf);

/*	Get another block of memory to save the image header.		*/
#define HEAD_MEM (1024)
	shid4 = shmget(7,HEAD_MEM,IPC_CREAT|0777);
	if(shid4 == -1) {
		flogerror("Can't get %d bytes of header shared memory (%s)",
		    HEAD_MEM,strerror(errno));
		exit(0);
	}
#ifdef SMH_LOCK
	shmctl(shid4,SHM_LOCK,&shbuf);
#endif
	head_data = shmat(shid4,(char *)0,0);

/*	Allow this shared memory segment to be destroyed when the	*/
/*	last process detaches from the memory. This happens on exit.	*/

	shmctl(shid4,IPC_RMID,&shbuf);


/*	In guider mode we may need to subtract a background image.	*/
/*	So we must allocate two additional buffers for this.		*/

	if(guider_mode) {

/*	    One buffer for accumulated background image.		*/
/*	    This is a 32-bit integer buffer, so we double the		*/
/*	    usual number of bytes.					*/

	    bkAccShid = shmget(8,2*shm_nbytes,IPC_CREAT|0777);
	    if(bkAccShid == -1) {
	        flogerror("Can't get %d bytes of accumulated background"
		    " image shared memory (%s)",
		    shm_nbytes,strerror(errno));
	        exit(0);
	    }
#ifdef SMH_LOCK
	    shmctl(bkAccShid,SHM_LOCK,&shbuf);
#endif
	    bkAccShad = shmat(bkAccShid,(char *)0,0);
	    shmctl(bkAccShid,IPC_RMID,&shbuf);

/*	    One buffer for the normalized version of the accumulated	*/
/*	    background image.						*/

	    bkNormShid = shmget(9,shm_nbytes,IPC_CREAT|0777);
	    if(bkNormShid == -1) {
	        flogerror("Can't get %d bytes of normalized background"
		    " image shared memory (%s)",
		    shm_nbytes,strerror(errno));
	        exit(0);
	    }
#ifdef SMH_LOCK
	    shmctl(bkNormShid,SHM_LOCK,&shbuf);
#endif
	    bkNormShad = shmat(bkNormShid,(char *)0,0);
	    shmctl(bkNormShid,IPC_RMID,&shbuf);
	}

/*	'image' is the array we read raw image data into.		*/
	image = (unsigned short *)shad2;

/*	Put the in-memory FITS header at the front of the descrambled	*/
/*	image. Make room for up to 5 FITS records worth of header.	*/
/*	This is 180 card images.					*/
#define F_REC_NUM 5

	header = (short *)shad;
	maxheadcards = F_REC_NUM*2880/80;

	memset((void *)header,' ',2880*F_REC_NUM);

/*	Make sure the last card image is an END card			*/

	strncpy((char *)header+2880*F_REC_NUM-80,"END",3);

/*	'mem_image' is the array for the descrambled image.		*/
	mem_image = shad+ 2880*F_REC_NUM;
	flogerror("mem_image start=%x end=%x",
	    (unsigned int)mem_image,(unsigned int)(shad+shm_nbytes-2));

/*	Establish a message que for sending messages to our forked	*/
/*	copies of camread.						*/

	msid = msgget(2,IPC_CREAT|0777);
        if(msid < 0) {
                flogerror("Can't get display msg queue 2: %s: EXITING!!!",
			strerror(errno));
                exit(0);
        }

/*	Flush the que of any old messages.				*/
	flogerror("flushing queues");
	while(check_msg(msid,DISPLAY_NEWIM) != NULL) {};
	while(check_msg(msid,RECORD_NEWIM) != NULL) {};
	while(check_msg(msid,GUIDER_NEWIM) != NULL) {};
	flogerror("Queues all flushed");

/*	Make sure a connection to the camera interface is open.	*/
	im_fd = ConnectGcam(&devicename);
	flogerror("Connected to camera");

	if(im_fd < 0) {
		char status[256];
		char *emsg = strerror(errno);
		flogerror("Can't open fiber port on %s: %s",devicename,emsg);
		sprintf(status,
		    "ERROR %s: Can't open fiber port %s to read images.#n"
		    "Error message is: %s",__FILE__, devicename, emsg);
		save_error_status(status);
		sleep(5);
		re_camread(children,child_desc,n_children);
	}
	flogerror("image fd=%d on %s",im_fd,devicename);

/*	Close any open connections to the database before we start	*/
/*	forking.							*/
	exit_sql();

/*	Do any necessary setup for the camera interface.                */
	interface_setup(interface_flag);

/*	Initialize our image reading semaphore, which helps	*/
/*	synchronize the read and descrambling processes.	*/

	sem_read = semget(123,2,IPC_CREAT|S_IRWXU|S_IRWXG|S_IRWXO);
	if(sem_read == -1) {
	    flogerror("camread: Unable to acquire semaphore: %s",
		strerror(errno));
	    flogerror("Can't synchronize without semaphore!");
	}

   	semun.val = 0;
	sem_ret = semctl(sem_read,0,SETVAL,semun);
	if (sem_ret < 0) {
	    flogerror("camread: semctl initialize of semaphore 0 failed: %s",
		      strerror(errno));
	    flogerror("exiting!!");
	    exit(1);
	}

   	semun.val = 0;
	sem_ret = semctl(sem_read,1,SETVAL,semun);
	if (sem_ret < 0) {
	    flogerror("camread: semctl initialize of semaphore 1 failed: %s",
		      strerror(errno));
	    flogerror("exiting!!");
	    exit(1);
	}

/*	Fork a new version of ourself to handle image descramble.	*/

	if((descram_child = fork()) == 0) {
		reset_argv("wait_for_image_to_descramble");
		wait_for_image_to_descramble(msid);
	}
	child_desc[n_children] = "wait_for_image_to_descramble";
	children[n_children++] = descram_child;

/*	Fork a new version of ourself to handle image display.		*/

	if((display_child = fork()) == 0) {
		reset_argv("wait_for_image_to_display");
		wait_for_image_to_display(msid);
	}
	child_desc[n_children] = "wait_for_image_to_display";
	children[n_children++] = display_child;

/*	Fork a new version of ourself to handle image recording.	*/

	if((record_child = fork()) == 0) {
	    reset_argv("wait_for_image_to_record");
	    wait_for_image_to_record(msid);
	}
	child_desc[n_children] = "wait_for_image_to_record";
	children[n_children++] = record_child;

/*	Fork a new version of ourself to handle guider processing.	*/
/*	Do this only if we are in guider mode. Otherwise we don't	*/
/*	have to worry about guiding.					*/

	if(guider_mode) {
	    if((guider_child = fork()) == 0) {
	        reset_argv("wait_for_image_to_guide");
	        wait_for_image_to_guide(msid);
	    }
	    child_desc[n_children] = "wait_for_image_to_guide";
	    children[n_children++] = guider_child;
	}

/*	Fork a new version of ourself to handle FITS card processing.	*/

	if((fits_child = fork()) == 0) {
	    reset_argv("wait_for_fits          ");
	    wait_for_fits();
	}
	child_desc[n_children] = "wait_for_fits";
	children[n_children++] = fits_child;

#ifdef _POSIX_MEMLOCK
/*	Lock ourself in memory so we are never swapped out.	*/
	ret = mlockall(MCL_FUTURE);
	if (ret) {
	    flogerror("camread: mlockall failed: %s",strerror(errno));
	} else {
            flogerror("main loop locked in memory");
        }

#endif
        reset_argv("reading image           ");

/*      Decrease our priority back to normal level.		*/
        ret = setpriority(PRIO_PROCESS,0,0);
        if (ret) {
            flogerror("Could not set normal priority: %s",strerror(errno));
        }

	ioctl(im_fd,CYP_SET_MAXROW,&maxrc);
	ioctl(im_fd,CYP_SET_MAXCOL,&maxrc);
	ioctl(im_fd,CYP_SET_READ_TIMEOUT,&timeout);
	ioctl(im_fd,CYP_HEADER_SIZE,&header_size);

#if 0
/*	Initialize our image reading semaphore, which helps	*/
/*	synchronize the read and descrambling processes.	*/

	sem_read = semget(123,2,IPC_CREAT|S_IRWXU|S_IRWXG);
	if(sem_read == -1) {
	    flogerror("camread: Unable to acquire semaphore: %s",
		strerror(errno));
	    flogerror("Can't synchronize without semaphore!");
	}
#endif

/*	Read images forever.					*/
	while(im_fd > 0) {
	    if(read_image(sem_read) == 0) break;
	}

/*	Restart everything.					*/
	flogerror("restarting by calling re_camread()");
	re_camread(children,child_desc,n_children);

	return(0);

}

static void re_camread(pid_t children[],char *child_desc[],int n_children)
{
	int n;
	int ret;
	pid_t self;
	extern char **environ;

	flogerror("In %s",__FUNCTION__);

/*	Close our image connection.				*/
	close(im_fd);

/*	Kill all of our child processes.			*/

	for(n=0; n<n_children; n++) {
	    ret = kill(children[n],SIGKILL);
	    if(ret == 0) {
		flogerror("%s waiting for child %d, %s",__FUNCTION__,
		    children[n],child_desc[n]);
		waitpid(children[n],NULL,0);
	    }
	}

/*	Restart ourself after a few seconds.			*/
	self = getpid();
	exit_sql();
	if(fork() == 0) {
	    waitpid(self,NULL,0); /* collect our parent's status	*/
	    sleep(1);
	    execve(saved_arg[0],saved_arg,environ);
	}
	exit(-1);
}

/*	This function is called if we receive the HUP signal.	*/

void sigcatch(int signum)
{
	flogerror("Exiting on signal %d",signum);
	if(shad != NULL)
	    shmdt(shad);
	if(shad2 != NULL)
	    shmdt(shad2);
	if(mem_fits != NULL)
	    shmdt(mem_fits);
	if(head_data != NULL)
	    shmdt(head_data);
	if(bkNormShad != NULL)
	    shmdt(bkNormShad);
	if(bkAccShad != NULL)
	    shmdt(bkAccShad);

	exit_sql();
	flogerror("sigcatch calling exit(0)");

	exit(0);
}

/*	We wait for an image readout to begin and then we descramble	*/
/*	the image as it is read out.					*/

void wait_for_image_to_descramble(int msid)
{
	int ret;
	int nr;
	struct header_data hdata;
	pthread_t thread;
	void *wait_for_image_flip_data(void *);
	struct timeval tv0;
	struct timeval tv,*tvp;
	int dt;
	struct sembuf sops;
	int sem_ret;

/*	Get the name of our log file from the database.			*/

	log_filename = getlogfile("wait_for_image_to_descramble");

	flogerror("%s pid %d",__FUNCTION__,getpid());

/*	Increase our priority to slightly less than the max.		*/
	ret = setpriority(PRIO_PROCESS,0,-18);
	if (ret) {
	    flogerror("wait_for_image_to_descramble: "
		"Could not set high priority: %s",strerror(errno));
	}

/*      First create a thread to handle image flip information.		*/
	if(pthread_create(&thread,NULL,wait_for_image_flip_data,
	    (void *)FLIP_ADDR)) {
		flogerror(
	"wait_for_image_flip_data thread creation problem: %s",
                strerror(errno));
                exit(0);
        }

/*	Make a connection to the traffic controller if we are sending	*/
/*	messages about image readout progress.				*/

	if(report_progress) {
	    connect_to_traffic("CAMREAD_PROGRESS");
	    ret = dbget("%d","ctalk.msgs.REPORT_PROGRESS.value",
			&report_progress_msg);
	    if(ret == -1) {
		report_progress_msg = 600;
		flogerror("Failed to get ctalk.msgs.REPORT_PROGRESS.value");
		flogerror("    Using default value %d",
			report_progress_msg);
	    } else {
		flogerror("Using ctalk.msgs.REPORT_PROGRESS.value = %d",
			report_progress_msg);
            }
	}

#if 0
/*	Get the semaphore id for our image-reading synchronization.	*/

	sem_read = semget(123,2,IPC_CREAT|S_IRWXU|S_IRWXG);
	if(sem_read == -1) {
	    flogerror("%s: Unable to acquire semaphore: %s",
		__FUNCTION__,strerror(errno));
	    flogerror("Can't synchronize without semaphore!");
	}
#endif

/*	Wait for an image header to arrive.				*/
	while (1) {
/*
	    flogerror("waiting for image");
*/
/*	    Capture the current time.					*/
	    ret = gettimeofday(&tv0,NULL);
	    if(ret < 0) {
		flogerror("Failed gettimeofday; %s",strerror(errno));
	    }

/*	    Increment the value of our semaphores.			*/
/*	    The first semaphore should cause the image reading process	*/
/*	    to wait until we have acquired the information about the	*/
/*	    new image.	The second will cause this process to wait	*/
/*	    until the reading process has acknowledged its release from	*/
/*	    the first semaphore.					*/
	    if(sem_read >= 0) {
	        sops.sem_num = 0;
	        sops.sem_op = 1;
	        sops.sem_flg = 0;
	        sem_ret = semop(sem_read,&sops,1);
	        if (sem_ret < 0) {
		    flogerror("%s: semop synchonization semaphore 0 increment failed: %s",
		        __FUNCTION__,strerror(errno));
	        }
	        sops.sem_num = 1;
	        sops.sem_op = 1;
	        sops.sem_flg = 0;
	        sem_ret = semop(sem_read,&sops,1);
	        if (sem_ret < 0) {
		    flogerror("%s: semop synchonization semaphore 1 increment failed: %s",
		        __FUNCTION__,strerror(errno));
	        }
	    }

/*	    Now wait for the image to arrive.				*/
	    ret = ioctl(im_fd,CYP_WAIT_READ,&nr);
	    if (ret) {
		flogerror("CYP_WAIT_READ ioctl error:");
		(void)loglasterror();
		close(im_fd);
		pause();    /* Wait for parent process to restart us.	*/
	    }

/*	    Capture the current time again.				*/
	    ret = gettimeofday(&tv,NULL);
	    if(ret < 0) {
		flogerror("Failed gettimeofday; %s",strerror(errno));
	    }

/*	    Compute the time elapsed during the CYP_WAIT_READ call.	*/
/*	    Units of dt are milliseconds.				*/

	    dt = (tv.tv_sec - tv0.tv_sec)*1000;
	    dt += (tv.tv_usec - tv0.tv_usec)/1000;

/*	    The address of the image is passed back to us in the	*/
/*	    CYP_WAIT_READ ioctl call.					*/

	    rimage = (unsigned short *)nr;

/*	    Log the elapsed time unless it is short. We assume that	*/
/*	    if it is short then we are interested in very fast readouts	*/
/*	    and we don't want to cause any delays with excessive I/O.	*/

	    if(dt>=100)
	        flogerror("image addr %x after %d ms",
			(unsigned int)nr,dt);

/*	    Extract the information from the header.			*/
	    decode_header((short *)rimage,&hdata);

/*	    Copy the header data from the raw image buffer to the	*/
/*	    head_data buffer.						*/
	    memcpy(head_data,rimage,hdata.size);

/*	    Save the time of image readout.				*/
	    tvp = FITS_TIME(hdata.id);
	    tvp->tv_sec = tv.tv_sec;
	    tvp->tv_usec = tv.tv_usec;

/*	    Decrement the value of semaphore 0.				*/
/*	    This should cause the image reading process to wake up	*/
/*	    if it was asleep, waiting for the semaphore to return to	*/
/*	    a value of 0.						*/
	    if(sem_read >= 0) {
	        sops.sem_num = 0;
	        sops.sem_op = -1;
	        sops.sem_flg = 0;
	        sem_ret = semop(sem_read,&sops,1);
	        if (sem_ret < 0) {
		    flogerror("%s: semop synchonization semaphore 0 decrement failed: %s",
		        __FUNCTION__,strerror(errno));
		    flogerror("exiting!!");
		    exit(1);
	        }
	    }

/*	    Clear the flag that says an image readout is aborted.	*/
	    image_aborted = 0;

/*	    Descramble image and send a message to our friends if	*/
/*	    we descramble it successfully.				*/

	    process_image = mem_image;
	    process_header = (char *)header;
	    process_buffer = shad;
	    process_shid = shid;
	    if(descramble(rimage,&hdata)) {
		if(image_is_background(&hdata)) {
		    save_background(&hdata);
		} else {
		    subtract_background(mem_image,&hdata);
		    process_image = mem_image;
		    process_header = (char *)header;
		    process_buffer = shad;
		    process_shid = shid;
		    last_was_background = 0;
		}
		if(send_newim_msgs(msid) < 0) {
		    flogerror("Unable to send NEWIM message: %s",
			strerror(errno));
		}
	    } else {
		flogerror("descramble() failed");
	    }

/*	    Wait on semaphore 1 if it is not now 0. This semaphore is	*/
/*	    set to 0 by the image reading process when that process has	*/
/*	    awakened from waiting on semaphore 0.			*/

	    sops.sem_num = 1;
	    sops.sem_op = 0;
	    sops.sem_flg = 0;
	    sem_ret = semop(sem_read,&sops,1);
	    if (sem_ret < 0) {
		flogerror("%s: semop synchonization wait on semaphore 1 failed: %s",
		    __FUNCTION__,strerror(errno));
		flogerror("exiting!!");
		exit(1);
	    }

	}
}

/*	We monitor the database for the flags that tell us if image	*/
/*	flipping in rows or columns is supposed to occur.		*/
/*	Also monitor the background subtraction flag.			*/
/*	Also monitor the database element that will tell us that a	*/
/*	readout has been aborted.					*/

void *wait_for_image_flip_data(void *vp)
{
	struct image_flips *image_flips = (struct image_flips *)vp;
	struct notify_handle nh;
	char *msg;
	char datum[64];

	memset(&nh,0,sizeof(nh));
	flogerror("%s pid %d",__FUNCTION__,getpid());

top:
	connect_to_notifier("CREAD_FLIP",notifier_port,&nh);
	nh.notify_data = (void *)image_flips;

/*	Tell notifier we want to learn about database items of interest	*/
/*	to us.								*/

	sprintf(datum,"telefocus.imageflips..%s",version);
	notify_desired(&nh,datum,get_flip_data);

/*	Get initial values for all flip variables			*/
	get_flip_data_init(image_flips);

/*	Tell notifier we also want to know about videoboard		*/
/*	configuration.							*/
	sprintf(datum,"telefocus.configs.iflag.%s",version);
	notify_desired(&nh,datum,get_videoconfig);

/*	Tell notifier we also want to know about background subtraction	*/
/*	control flag.							*/
	sprintf(datum,"telefocus.background.value.%s",version);
	notify_desired(&nh,datum,get_backgroundconfig);

/*	Get initial value for background subtraction control flag.	*/
	get_background_data_init();

/*	Tell notifier we also want to know about events. We will learn	*/
/*	if an exposure is aborted.					*/
	sprintf(datum,"telefocus.variables.value.event");
	notify_desired(&nh,datum,check_for_abort);

/*      Wait for messages from the database notifier.                   */
        while((msg=notify_read(&nh)) != NULL) {

/*              Parse the notification message                          */
                notify_parse(&nh);

/*              Call the callback function                              */
                notify_callcb(&nh);
        }
        sleep(10);
        goto top;

}

/*	Retrieve all of the current image flipping values from	*/
/*	the database and save them in our image_flips structure.*/

void get_flip_data_init(struct image_flips *image_flips)
{
	int val;
	char datum[64];

	sprintf(datum,"telefocus.imageflips.%s.row0",version);
	if(dbget("%d",datum,&val) < 0)
		val = 0;
	image_flips->row = val;

	sprintf(datum,"telefocus.imageflips.%s.row1",version);
	if(dbget("%d",datum,&val) < 0)
		val = 0;
	image_flips->row1 = val;

	sprintf(datum,"telefocus.imageflips.%s.row2",version);
	if(dbget("%d",datum,&val) < 0)
		val = 0;
	image_flips->row2 = val;

	sprintf(datum,"telefocus.imageflips.%s.row3",version);
	if(dbget("%d",datum,&val) < 0)
		val = 0;
	//image_flips->row3 = val;

	sprintf(datum,"telefocus.imageflips.%s.column0",version);
	if(dbget("%d",datum,&val) < 0)
		val = 0;
	image_flips->column = val;

	sprintf(datum,"telefocus.imageflips.%s.column1",version);
	if(dbget("%d",datum,&val) < 0)
		val = 0;
	image_flips->column1 = val;

	sprintf(datum,"telefocus.imageflips.%s.column2",version);
	if(dbget("%d",datum,&val) < 0)
		val = 0;
	image_flips->column2 = val;

	sprintf(datum,"telefocus.imageflips.%s.column3",version);
	if(dbget("%d",datum,&val) < 0)
		val = 0;
	//image_flips->column3 = val;

/*	Update the overall image row and column flipping flags.		*/
	recompute_flips(image_flips);

}

/*	Retrieve all of the current image flipping values from	*/
/*	the database and save them in our image_flips structure.*/

void get_background_data_init(void)
{
	int val;
	char datum[64];

	sprintf(datum,"telefocus.background.%s.value",version);
	if(dbget("%d",datum,&val) < 0)
		val = 0;
	subtractBackground = val;
	if(val)
	    flogerror("Initially found background subtraction enabled");
	else
	    flogerror("Initially found background subtraction disabled");
}

/*	Recompute the image row and column flipping flags based	*/
/*	on the current formula for combining the flip variables.*/
void recompute_flips(struct image_flips *image_flips)
{

/*	image_flips->row can be used to turn flipping on or off.	*/
/*	The bitwise exclusive OR of row1 and row2 allows two sources	*/
/*	for row flipping, which cancel each other if they are both on.	*/

	image_flips->row_flip = image_flips->row &&
		(image_flips->row1 ^ image_flips->row2);

/*	Column flipping uses the same line of logic as row flipping.	*/

	image_flips->col_flip = image_flips->column &&
		(image_flips->column1 ^ image_flips->column2);

	flogerror("row_flip %d   row %d row1 %d row2 %d",
	    image_flips->row_flip,image_flips->row,image_flips->row1,
	    image_flips->row2);
	flogerror("col_flip %d   col %d col1 %d col2 %d",
	    image_flips->col_flip,image_flips->column,image_flips->column1,
	    image_flips->column2);

}

/*	This callback is called when any imageflips db table values	*/
/*	are changed for the current version.				*/
int get_flip_data(void *dp)
{
	struct notify_handle *nh = dp;
	int val;
	char *item;
	struct image_flips *image_flips =
		(struct image_flips *)nh->notify_data;

/*	Convert the value to numeric.					*/
	sscanf(nh->msg.value,"%d",&val);
	item = nh->msg.column;
	flogerror("in get_flip_data flip item %s",item);

/*	Set the appropriate value.					*/
	if(*item == 'r') {
	    if(strcmp(item,"row0") == 0)
		image_flips->row = val;
	    else if(strcmp(item,"row1") == 0)
                image_flips->row1 = val;
	    else if(strcmp(item,"row2") == 0)
                image_flips->row2 = val;
	   // else if(strcmp(item,"row3") == 0)
                //image_flips->row3 = val;
	} else if (*item == 'c') {
	    if(strcmp(item,"column0") == 0) {
		image_flips->column = val;
	    } else if(strcmp(item,"column1") == 0)
                image_flips->column1 = val;
	    else if(strcmp(item,"column2") == 0)
                image_flips->column2 = val;
	   // else if(strcmp(item,"column3") == 0)
               // image_flips->column3 = val;
	}

/*	Update the overall image row and column flipping flags.		*/
	recompute_flips(image_flips);

	return 0;
}

/*	Get the video board configuration.				*/
int get_videoconfig(void *dp)
{
	struct notify_handle *nh = dp;
	int val;

/*	Convert the value to numeric.					*/
	sscanf(nh->msg.value,"%d",&val);
	interface_flag = val;

        flogerror("Now using interface flag %d: %s",interface_flag,
		interface_meaning(interface_flag));

/*	Do any necessary setup for the new value.                       */
	interface_setup(val);

	return(0);
}

/*	Get the background subtraction control flag.			*/
int get_backgroundconfig(void *dp)
{
	struct notify_handle *nh = dp;
	int val;

/*	Convert the value to numeric.					*/
	sscanf(nh->msg.value,"%d",&val);
	subtractBackground = val;

        flogerror("Now using subtractBackground %d",subtractBackground);
	return(0);
}

/*	Check to see if we have an exposure-abort event.		*/
int check_for_abort(void *dp)
{
	struct notify_handle *nh = dp;
	char *r = nh->msg.message;
	while (*r != '\0') {
	    if((*r == 'R') && (strcmp(r,"ReadoutAbort") == 0)) {
	        image_aborted = 1;
		flogerror("wait_for_image_to_descramble: Readout aborted");
		return(1);
	    }
	    r++;
	}
	return(0);
}

/*	We wait for a DISPLAY_NEWIM message to be sent			*/
/*	from the image reading program.					*/
/*	We send a message (via traffic) to the display program with	*/
/*	the shared memory information.					*/

void wait_for_image_to_display(int msid)
{
	char *msg;
	int ret;
	char msg_word[16];

/*	Get the name of our log file from the database.			*/

	log_filename = getlogfile("wait_for_image_to_display");
	flogerror("%s pid %d",__FUNCTION__,getpid());

/*	Close connection to database.					*/
	exit_sql();

/*	Increase our priority to slightly less than the max.		*/
	ret = setpriority(PRIO_PROCESS,0,-18);
	if (ret) {
	    flogerror("wait_for_image_to_display: "
		"Could not set high priority: %s",strerror(errno));
	}

	if(im_fd >= 0) close(im_fd);	/* Close image reading channel	*/

#ifdef _POSIX_MEMLOCK
	ret = mlockall(MCL_FUTURE);
	if (ret) {
	    flogerror("%s: mlockall failed: %s",__FUNCTION__,
		strerror(errno));
	} else {
	    flogerror("%s locked in memory",__FUNCTION__);
	}
#endif

/*	Make a connection to the traffic controller. We'll be sending	*/
/*	messages to the display program via traffic.			*/

	connect_to_traffic("GREADXV_DISPLAY");

/*	Now we enter a loop waiting for images to display.		*/
	while(1) {
	    void send_display_message(void);
//	flogerror("calling check_msg_wait DISPLAY_NEWIM");
	    msg = check_msg_wait(msid,DISPLAY_NEWIM);
	    sscanf(msg,"%s %d %x %x %x",msg_word,&process_shid,
		(unsigned int *)&process_buffer,
		(unsigned int *)&process_image,
		(unsigned int *)&process_header);

/*	    Send a message to the display program.			*/
	    send_display_message();
	}
}

/*	We wait for a GUIDER_NEWIM message to be sent			*/
/*	from the image reading program.					*/
/*	When we receive one we check to see if we have a request	*/
/*	from the guider program. If we do we send back the shared memory*/
/*	location of the image.						*/

void wait_for_image_to_guide(int msid)
{
	char *msg;
	int guider_flag = 0;
	int guider_msid = -1;
	int guider_sock = -1;
	int ret;
	char msg_word[16];
	struct header_data hdata;
	pthread_t thread;
	void wait_for_guider_sock(void *i);

/*	Get the name of our log file from the database.			*/

	log_filename = getlogfile("wait_for_image_to_guide");
	flogerror("%s pid %d",__FUNCTION__,getpid());

/*	Close connection to database.					*/
	exit_sql();

/*	First create a thread to handle guider socket connections.	*/
	if(pthread_create(&thread,NULL,(void *)wait_for_guider_sock,
	  (void *)&guider_sock)) {
	    flogerror("guider thread creation problem: %s",
                strerror(errno));
		exit(0);
	}

	sleep(1);

/*	Increase our priority to slightly less than the max.		*/
	ret = setpriority(PRIO_PROCESS,0,-18);
	if (ret) {
	    flogerror("wait_for_image_to_guide: "
		"Could not set high priority: %s",strerror(errno));
	}

        guider_msid = msgget(3,IPC_CREAT|0777);
        if(guider_msid < 0) {
                flogerror("Can't get guider msg queue 3: %s: EXITING!!!",
                        strerror(errno));
                exit(0);
        }

#ifdef _POSIX_MEMLOCK
	ret = mlockall(MCL_FUTURE);
	if (ret) {
	    flogerror("%s: mlockall failed: %s",__FUNCTION__,
		strerror(errno));
	} else {
	    flogerror("%s locked in memory",__FUNCTION__);
	}
#endif

/*	Now we enter a loop waiting for images to guide on.		*/
	while(1) {
/*	   First wait for a new image to arrive				*/
	    msg = check_msg_wait(msid,GUIDER_NEWIM);
	    sscanf(msg,"%s %d %x %x %x",msg_word,&process_shid,
		(unsigned int *)&process_buffer,
		(unsigned int *)&process_image,
		(unsigned int *)&process_header);

/*	    Then check to see if the guide program has told us it is	*/
/*	    ready to process a new image.				*/
	    while((msg = check_msg(guider_msid,GUIDER_MSG)) != (char *)0) {
		int process_guider_msg(char *msg);
		guider_flag = process_guider_msg(msg);
	    }

/*	    Send message to guider program if requested.		*/
	    if(guider_flag) {
/*              Extract the information from the header.                */

                decode_header((short *)head_data,&hdata);

/*		Send the guider message only if this is not a		*/
/*		background image.					*/

		if(image_is_background(&hdata) == 0) {
		
/*		    Send the message.					*/

                    send_guider_message(guider_sock,hdata.final.r,
			hdata.final.c);

/*		    Send no more messages until requested again.	*/

		    guider_flag = 0;
		}
	    }

	}
}

/*	Thread that waits for a socket connection from a guider		*/
/*	program.							*/
void wait_for_guider_sock(void *vmsid)
{
	int *msid = (int *)vmsid;
	int guider_port;
	socklen_t salen;
	struct sockaddr sa;
	fd_set Readmask;
	int sock;
	int scnt;

	flogerror("%s pid %d",__FUNCTION__,getpid());

	for(scnt=0; scnt < 20; scnt++) {
	    guider_port = makeportsocket(6600);
            if(guider_port < 0) {
                flogerror("%d Can't get guider msg socket: %s: wait 2 seconds",
                        scnt,strerror(errno));
		sleep(2);
	    } else {
		break;
	    }
        }

        if(guider_port < 0) {
		flogerror("wait_for_guider_sock EXITING!!!");
                exit(0);
	} else {
		flogerror("wait_for_guider_sock set up socket %d",guider_port);
	}

	if (fcntl(guider_port,F_SETFL,O_NONBLOCK) < 0) {
	    flogerror("Could not set guider_port to non-blocking: %s",
			strerror(errno));
	    exit(0);
	}
	flogerror("guider socket %d",guider_port);

	salen = sizeof(sa);
	while(1) {
	    FD_ZERO(&Readmask);
	    FD_SET(guider_port,&Readmask);
	    select(guider_port+1,&Readmask, (fd_set *)0, (fd_set *)0,
                (struct timeval *)0);
	    *msid = sock = accept(guider_port,&sa,&salen);
	    flogerror("guider connection %d",sock);
	    FD_ZERO(&Readmask);
	    FD_SET(sock,&Readmask);
	    select(sock+1,&Readmask,(fd_set *)0, (fd_set *)0,
		(struct timeval *)0);
	    *msid = -1;
	    flogerror("closing guider connection %d",sock);
	    close(sock);
	}
}

/*      See if we need to swap bytes because we're running on a PC      */
/*      or similar machine.                                             */
/*      To use this function, store a 1 in a short (2-byte) integer     */
/*      variable and then pass the address of the variable [cast        */
/*      the address as a (char *)].                                     */
int needswab(char *test)
{
        if(*test)
                return(1);              /* Need to swap                 */
        else
                return(0);              /* Do not need to swap          */
}


/*	We wait for a RECORD_NEWIM message to be sent			*/
/*	from the image reading program.					*/
/*	When we receive one we first check to see if we have one or	*/
/*	more requests to change our recording status.			*/
/*	If we do we read all such messages and use the last one to	*/
/*	control our recording.						*/
/*	If, after reading the recording status message, we are still	*/
/*	recording then we record the image according to the last	*/
/*	message.							*/

void wait_for_image_to_record(int msid)
{
	char *msg;
	int ret;
	struct header_data hdata;
	struct record *record;
	struct itimerval brtimer,old_brtimer;
	void sendbcast(int sigval);
	static short swabtest = 1;

/*	Get the name of our log file from the database.			*/

	log_filename = getlogfile("wait_for_image_to_record");
	flogerror("%s pid %d",__FUNCTION__,getpid());

/*	Close connection to database.					*/
	exit_sql();

/*	Increase our priority to slightly less than the max.		*/

	ret = setpriority(PRIO_PROCESS,0,-18);
	if (ret) {
	    flogerror("wait_for_image_to_record: "
		"Could not set high priority: %s",strerror(errno));
	}

/*      Allocate an image swapping area if needed.                      */
        if(imaddr_swapped == NULL) {
            if(needswab((char *)&swabtest)) {
                imaddr_swapped = malloc(shm_nbytes);
                if(imaddr_swapped == NULL) {
                    flogerror(
                      "CAN'T ALLOCATE MEMORY (%d bytes) FOR SWAPPED IMAGE"
                      "\n    Reason: %s",shm_nbytes,strerror(errno));
                    imaddr_swapped = (char *)mem_image;
                }
            } else {
                imaddr_swapped = (char *)mem_image;
            }
            bytes_allocated = shm_nbytes;
        }

#ifdef _POSIX_MEMLOCK
	ret = mlockall(MCL_FUTURE);
	if (ret) {
	    flogerror("%s: mlockall failed: %s",__FUNCTION__,
		strerror(errno));
	} else {
            flogerror("%s locked in memory",__FUNCTION__);
        }

#endif
	if(im_fd >= 0) close(im_fd);	/* Close image reading channel	*/

	connect_to_traffic("CREAD_RECORD");

	while(1) {
	    msg = check_msg_wait(msid,RECORD_NEWIM);
	    if(msg == NULL) {
		flogerror("Why did check_msg_wait return with an error: %s",
			strerror(errno));
		usleep(100000);
		continue;
	    } else {
	        process_record_newim_msg(msg);
	    }
	    while((msg = check_msg(msid,RECORD_MSG)) != (char *)0)
		process_record_msg(msg);

/*	    Extract the information from the header.			*/

	    decode_header((short *)head_data,&hdata);

/*	    If our record count is non-zero we are to record.		*/
	    record = RECORD_ADDR(hdata.id);

/*
	flogerror("check image record %x id %d count=%d",
		record,hdata.id,record->count);
*/

	    if(record->count != 0) {

		if(record_image((unsigned short *)process_image,
		  &hdata,NULL)) {
		    int nc;

/*		    To avoid flooding the user GUI with too many	*/
/*		    messages we will send a broacast not more than	*/
/*		    five times per second. We do this by using a timer	*/
/*		    that triggers after 200000 microseconds.		*/

/*		    Cancel previous timer so it doesn't happen		*/
/*		    while we are updating the music message.		*/
/*		    While doing this we get the remaining interval	*/
/*		    from any timer already running.			*/

		    brtimer.it_interval.tv_usec = 0;
		    brtimer.it_interval.tv_sec = 0;
		    brtimer.it_value.tv_usec = 0;
		    brtimer.it_value.tv_sec = 0;
		    old_brtimer.it_interval.tv_usec = 0;
		    old_brtimer.it_interval.tv_sec = 0;
		    old_brtimer.it_value.tv_usec = 0;
		    old_brtimer.it_value.tv_sec = 0;
		    setitimer(ITIMER_REAL,&brtimer,&old_brtimer);

/*		    Establish our signal handler.		*/
		    signal(SIGALRM,sendbcast);

/*		    Set up the message to broadcast		*/
		    mstart(BROADCAST,our_addr,RECORDED_IMAGE);
		    nc = strlen(record->recorded_file)+1;
		    mputl(nc);
		    mput(record->recorded_file,nc);
		    mputl(record->record_id);

/*		    Schedule it to be delivered on the old timer	*/
/*		    interval, or on a new one if there was none pending.*/

		    brtimer.it_interval.tv_usec = 0;
		    brtimer.it_interval.tv_sec = 0;

		    if((old_brtimer.it_value.tv_sec != 0) && (
		       (old_brtimer.it_value.tv_usec) != 0)) {
		        brtimer.it_value.tv_usec = old_brtimer.it_value.tv_usec;
		        brtimer.it_value.tv_sec = old_brtimer.it_value.tv_sec;
		    } else {
		        brtimer.it_value.tv_usec = 200000;
		        brtimer.it_value.tv_sec = 0;
		    }

/*		    Restart the timer.					*/
		    if(setitimer(ITIMER_REAL,&brtimer,NULL)) {
			flogerror("setitimer error: %s",strerror(errno));
		    }

/*		    If our count is > 0 then decrement it. This logic	*/
/*		    gives us one of two behaviors. If count was positive*/
/*		    we'll continue recording until the count reaches 0.	*/
/*		    But if the count was < 0 we'll continue to record	*/
/*		    until the count is modified by another RECORD_MSG	*/
/*		    or the id changes.					*/

		    if(record->count > 0)
		        record->count--;
		} else {
		    flogerror("Unable to record image in file %s",
			record->recorded_file);
		}
	    } else if (record_unrecorded) {
flogerror("Recording unrecorded image");
		if(record_image((unsigned short *)process_image,
		  &hdata,unrecorded_file) == 0) {
		    flogerror("Unable to record 'unrecorded' image in file %s",
			unrecorded_file);
		}
flogerror("Successfully recorded unrecorded image");
	    }
	}
}

/*	In response to a SIGALRM we try to do an msend() to send the	*/
/*	RECORDED_IMAGE message.						*/

void sendbcast(int sigval)
{
    if(msend(traffic_fd) < 0) {
	flogerror("sendbcast unable to send broadcast of RECORDED_IMAGE");
    }
}

/*	Process that waits for FITS header data and builds FITS		*/
/*	cards in shared memory.						*/

void wait_for_fits(void)
{
	char *msg;
	struct notify_handle nh;
	struct keydef *keys;

	memset(&nh,0,sizeof(nh));

/*	Get the name of our log file from the database.			*/

	log_filename = getlogfile("wait_for_fits");
	flogerror("%s pid %d",__FUNCTION__,getpid());

	if(im_fd >= 0) close(im_fd);	/* Close image reading channel	*/

	get_keyword_list();

/*	Close connection to database.					*/
	exit_sql();
top:
	connect_to_notifier("CREAD_FITS",notifier_port,&nh);

/*	Tell notifier we want to learn about database items of interest	*/
/*	to us.								*/


	for(keys=all_keywords; keys != NULL; keys = keys->key_next) {
	    if(strcmp(lindex(keys->acquire,0),"from_notify") == 0) {
	        flogerror("setting up notify for %s",keys->source);
		keys->data = (void *)&nh;
		notify_desired(&nh,keys->source,handle_notify_module);
	    }
	}

/*	The directory in which to record images.			*/
	notify_desired(&nh,"imageid.recording.dir.",get_record_dir);

/*	The "root" part of the file name.				*/
	notify_desired(&nh,"imageid.recording.root.",get_record_root);

/*	The file suffix to use (like .fits).				*/
	notify_desired(&nh,"imageid.recording.suffix.",get_record_suffix);

/*	Number of images to record for a given image ID.		*/
	notify_desired(&nh,"imageid.recording.value.",get_record);

/*	The file number (observation number) to be used to construct	*/
/*	the image file name.						*/
	notify_desired(&nh,"imageid.recording.number.",get_record_number);

/*	Wait for messages from the database notifier.			*/
	while((msg=notify_read(&nh)) != NULL) {

/*		Parse the notification message				*/
		notify_parse(&nh);
		flogerror(
    "client %s\n callback %x\n db %s\n table %s\n row %s\n col %s\n value %s",
    		nh.msg.client_t,nh.msg.callback,nh.msg.database,nh.msg.table,
		nh.msg.row,nh.msg.column,nh.msg.value);

/*		Call the callback function				*/
	    	notify_callcb(&nh);
        }
	sleep(10);
	goto top;
}

/*	Read the database to get a list of all of the FITS	*/
/*	keywords that may appear in the image FITS headers.	*/
static void get_keyword_list(void)
{
	struct tuple *keynames;
	struct keydef *newkey;
	void *handle;
	int ret;
	int nkey;
	struct tuple value[2] = {{0,0,(char *)0,(char *)0},
                                 {0,0,(char *)0,(char *)0}};
	struct tuple *tuplep;

/*	Get the number of FITS keywords in the database.		*/

	ret = sqlget("nitems","imageid","keywords","","","__STD",value,2);
        if(ret < 0) {
	    flogerror(
      "SERIOUS: Could not get number of FITS keywords from imageid.keywords");
	    flogerror("     Database error: %s",value[0].value);
            freetuple(&value[0]);
	    return;
	}
	sscanf(value[0].value,"%d",&nkey);
        freetuple(&value[0]);
	flogerror("There were %d FITS keywords found in the database.",
		nkey);
        nkey++;
	keynames = calloc(nkey,sizeof(struct tuple));
	if(keynames == NULL) {
		flogerror("SERIOUS: "
	"Could not get %d bytes memory to store FITS keywords"
	"\n    %s",nkey*sizeof(struct tuple),strerror(errno));
	}

/*	Get the list of names of all FITS keywords.		*/

	ret = sqlget("items","imageid","keywords","","","__STD",keynames,nkey);
	if(ret < 0) {
	    flogerror(
      "SERIOUS: Could not get list of FITS keywords from imageid.keywords");
	    flogerror("     Database error: %s",keynames->value);
	    freetuple(keynames);
	    return;
	}

/*	Process each keyword.						*/

	for(tuplep = keynames; tuplep->value != (char *)0; tuplep++) {
	    int nc;

	    flogerror("process keyname %s",tuplep->value);

/*	    Get the list of versions to which this keyword applies.	*/
	    ret = sqlget("select","imageid","keywords",tuplep->value,"versions",                "__STD",&value[0],2);
	    if(ret < 0) {
		flogerror(
		    "SERIOUS: Could not get FITS keyword db item %s.%s.%s.%s",
			"imageid","keywords",tuplep->value,"versions");
	        flogerror("     Database error: %s",value[0].value);
	        freetuple(&value[0]);
	  	continue;
	    }

/*	    If the current version is not in the list, skip this	*/
/*	    keyword.							*/
	    if (version_match(value[0].value) == 0) {
	        freetuple(&value[0]);
		continue;
	    }
	    freetuple(&value[0]);


	    newkey = calloc(1,sizeof(struct keydef));
	    if(newkey == NULL) {
		flogerror("SERIOUS: "
	"Could not get %d bytes memory to store FITS keyword data for %s"
	"\n    %s",sizeof(struct keydef),tuplep->value,strerror(errno));
		continue;
	    }
	    newkey->keyname = strdup(tuplep->value);

/*	    Get the FITS keyword.					*/

	    ret = sqlget("select","imageid","keywords",tuplep->value,
                "keyword","__STD",&value[0],2);
	    if(ret < 0) {
		flogerror(
		    "SERIOUS: Could not get FITS keyword db item %s.%s.%s.%s",
			"imageid","keywords",tuplep->value,"keyword");
	        flogerror("     Database error: %s",value[0].value);
		freetuple(&value[0]);
		free_key(newkey);
		continue;
	    }
	    newkey->keyword = strdup(value[0].value);

/*	    Make sure the keyword is no more than 8 characters in length. */
	    nc = strlen(newkey->keyword);
	    if (nc > 8) {
		flogerror(
		  "SERIOUS: found %d-byte keyword %s - truncating to 8 bytes",
			nc,newkey->keyword);
		nc = 8;
		*(newkey->keyword+nc) = '\0';
	    } 

	    freetuple(&value[0]);

/*	    Make sure this keyword is not a duplicate.			*/
	    if (find_keyword(newkey->keyword) != NULL) {
		flogerror(
		  "SERIOUS: found duplicate keyword %s - skipping duplicate",
			newkey->keyword);
		free_key(newkey);
		continue;
	    }

/*	    Get the acquisition descriptor.				*/

	    ret = sqlget("select","imageid","keywords",tuplep->value,
                "acquire","__STD",&value[0],2);
	    if(ret < 0) {
		flogerror(
		    "SERIOUS: Could not get FITS keyword db item %s.%s.%s.%s",
			"imageid","keywords",tuplep->value,"acquire");
	        flogerror("     Database error: %s",value[0].value);
		freetuple(&value[0]);
		free_key(newkey);
		continue;
	    }
	    newkey->acquire = strdup(value[0].value);
	    freetuple(&value[0]);

/*	    Get the FITS card format.					*/

	    ret = sqlget("select","imageid","keywords",tuplep->value,
                "format","__STD",&value[0],2);
	    if(ret < 0) {
		flogerror(
		    "SERIOUS: Could not get FITS keyword db item %s.%s.%s.%s",
			"imageid","keywords",tuplep->value,"format");
	        flogerror("     Database error: %s",value[0].value);
		freetuple(&value[0]);
		free_key(newkey);
		continue;
	    }
	    newkey->format = strdup(value[0].value);
	    freetuple(&value[0]);

/*	    Get the FITS card comment.					*/

	    ret = sqlget("select","imageid","keywords",tuplep->value,
                "comment","__STD",&value[0],2);
	    if(ret < 0) {
		flogerror(
		    "SERIOUS: Could not get FITS keyword db item %s.%s.%s.%s",
			"imageid","keywords",tuplep->value,"comment");
	        flogerror("     Database error: %s",value[0].value);
		freetuple(&value[0]);
		free_key(newkey);
		continue;
	    }
	    newkey->comment = strdup(value[0].value);
	    freetuple(&value[0]);

/*	    Get the loadable object library that supplies the value.	*/

	    ret = sqlget("select","imageid","keywords",tuplep->value,
                "proc_library","__STD",&value[0],2);
	    if(ret < 0) {
		flogerror(
		    "SERIOUS: Could not get FITS keyword db item %s.%s.%s.%s",
			"imageid","keywords",tuplep->value,"proc_library");
	        flogerror("     Database error: %s",value[0].value);
		freetuple(&value[0]);
		free_key(newkey);
		continue;
	    }
	    newkey->proc_lib = strdup_envar(value[0].value);
	    freetuple(&value[0]);

/*	    Get the FITS keyword source descriptor.			*/

	    ret = sqlget("select","imageid","keywords",tuplep->value,
                "source_location","__STD",&value[0],2);
	    if(ret < 0) {
		flogerror(
		    "SERIOUS: Could not get FITS keyword db item %s.%s.%s.%s",
			"imageid","keywords",tuplep->value,"source_location");
	        flogerror("     Database error: %s",value[0].value);
		freetuple(&value[0]);
		free_key(newkey);
		continue;
	    }
	    newkey->source = strdup(value[0].value);
	    freetuple(&value[0]);

/*	    Add this keyword to our list of keywords.			*/
	    add_key(newkey);

	}
	freetuple(keynames);

/*	Now for each keyword, look up the actual format string.		*/
/*	If we don't find a match in the database then we assume we	*/
/*	already have the real format.					*/

	for(newkey=all_keywords; newkey != NULL; newkey = newkey->key_next) {
	    ret = sqlget("select","imageid","fitsformat",newkey->format,
                "format","__STD",&value[0],2);
	    if(ret < 0) {
		flogerror(
		    "SERIOUS: Could not get FITS format %s.%s.%s.%s",
			"imageid","fitsformat",newkey->format,"format");
		flogerror("    Database error: %s",value[0].value);
		freetuple(&value[0]);
		continue;
	    }
	    free(newkey->format);
	    newkey->format = strdup(value[0].value);
	    freetuple(&value[0]);
	}

/*	Now for each keyword, make sure the necessary loadable object	*/
/*	is installed.							*/
/*	Start with the camera_id keyword because other keywords depend	*/
/*	on having this one initialized first.				*/
	newkey = find_keydef("camera_id");
	handle = dlopen(newkey->proc_lib, RTLD_GLOBAL|RTLD_NOW);
        if(!handle) {
            flogerror("SERIOUS: Error loading module %s for keyword %s: %s",
                newkey->proc_lib,newkey->keyname,dlerror());
	}

	for(newkey=all_keywords; newkey != NULL; newkey = newkey->key_next) {

/*	    Load the module for this keyword if it is not already	*/
/*	    loaded. We know it is loaded if the key_getter function	*/
/*	    pointer is already defined since each module identifies	*/
/*	    the keywords it can handle and sets the key_getter pointer	*/
/*	    for all keywords it handles. Thus a module loaded earlier	*/
/*	    may have already set this keyword's key_getter pointer.	*/

	    flogerror("Loading module for %s %s",newkey->keyname,newkey->proc_lib);
	    if (newkey->key_getter == NULL) {
		handle = dlopen(newkey->proc_lib, RTLD_GLOBAL|RTLD_NOW);
		if(!handle) {
		    flogerror("SERIOUS: Error loading module %s for keyword %s: %s",
			newkey->proc_lib,newkey->keyname,dlerror());
		    continue;
		} else {
		    flogerror("Loaded module %s for keyword %s",
			newkey->proc_lib,newkey->keyname);
		}
	    } else {
		flogerror("Already have module %s loaded for keyword %s",
			newkey->proc_lib,newkey->keyname);
	    }

/*	    If this keyword has an initialization routine defined	*/
/*	    then call that routine now.					*/
	    if(newkey->key_initter != NULL) {
	        if((*newkey->key_initter)(newkey) != 1) {
		     flogerror("SERIOUS: Error initializing FITS keyword %s",
			newkey->keyname);
		} else {
		    flogerror("Initialized FITS arrays for keyword %s",
			newkey->keyname);
		}
	    }
	}
}

/*	Add the given keyword to the front of our linked list of	*/
/*	keywords.							*/
static void add_key(struct keydef *key)
{
	key->key_next = all_keywords;
	all_keywords = key;
	return;
}

/*	Free all of the memory associated with a keydef struct		*/
/*	including the memory for the struct.				*/
static void free_key(struct keydef *key)
{
	if(key == NULL) return;
	if(key->keyname != NULL) free(key->keyname);
	if(key->keyword != NULL) free(key->keyword);
	if(key->acquire != NULL) free(key->acquire);
	if(key->format != NULL) free(key->format);
	if(key->proc_lib != NULL) free(key->proc_lib);
	if(key->source != NULL) free(key->source);
	free(key);
}

/*	Find the keyword given by the passed parameter keyname.	*/
struct keydef *find_keydef(char *keyname)
{
	struct keydef *keys;

	for(keys=all_keywords; keys != NULL; keys = keys->key_next) {
	    if((*keyname == *(keys->keyname)) &&
		(strcmp(keyname,keys->keyname) == 0)) {
		return(keys);
	    }
	}
	return(NULL);
}

/*	Find the keyword entry given by the passed keyword.	*/
struct keydef *find_keyword(char *keyword)
{
	struct keydef *keys;

	for(keys=all_keywords; keys != NULL; keys = keys->key_next) {
	    if((*keyword == *(keys->keyword)) &&
		(strcmp(keyword,keys->keyword) == 0)) {
		return(keys);
	    }
	}
	return(NULL);
}

/*	Find the keyword given by the passed parameter source.	*/
struct keydef *find_keysource(char *source)
{
	struct keydef *keys;

	for(keys=all_keywords; keys != NULL; keys = keys->key_next) {
	    if((*source == *(keys->source)) &&
		(strcmp(source,keys->source) == 0)) {
		return(keys);
	    }
	}
	return(NULL);
}

/*	Return a pointer to a copy of the item'th word in the string	*/
/*	pointed to by 'list'.						*/
char *lindex(char *list,int item)
{
	static char cval[128];
	char *cp = list;
	int nc= 0;

/*	Skip over leading items.				*/
	while(item > 0) {
	    char *last;
	    last = cp;
	    cp = nextelement(cp);
	    item--;
	}

/*	If we came to the end of the string then we return the	*/
/*	null string.						*/
	if (*cp == 0) {
		cval[0] = '\0';
		return cval;
	}

/*	Now we copy all non-space characters up to the end of the	*/
/*	string or until our buffer fills.				*/
	while(isspace(*cp) == 0) {
	    if(nc < sizeof(cval)-1) {
		cval[nc] = *cp;
		nc++;
	    }
	    cp++;
	}
     	cval[nc] = '\0';
	return cval;
}

char *nextelement(char *cp)
{
/*	First skip over all non-space characters.		*/
	while(*cp != '\0') {
	    if(isspace(*cp) == 0)
		cp++;
	    else
		break;
	}

/*	If we've come to the end of the string we are done.	*/
	if(*cp == '\0') return(cp);

/*	Then skip over the following space characters.		*/
	while(*cp != '\0') {
	    if(isspace(*cp))
                cp++;
            else
                break;
        }

/*	cp now points to the next non-space or the end of the	*/
/*	string.							*/
	return(cp);
}

/*	See if the current version matches any of the versions in the	*/
/*	given list. The word "all" matches any version.			*/
static int version_match(char *test_version)
{
	char *next_version;
	int ok = 0;

	while((next_version=next_token(test_version)) != NULL) {
	    if(strcmp(next_version,"all") == 0) {
		ok = 1;
		flogerror("    Use this keyword because version list"
		" for this keyword==all");
            }
	    if(strcmp(next_version,version) == 0) {
		ok = 1;
		flogerror("   Use this keyword because version list"
		    " for this keyword matches current version: %s",
                    version);
            }
	    free(next_version);
	    if(ok) return 1;
	}
	flogerror("    Don't use this keyword because version list "
            "for this keyword does not match current version: %s", version);
	return 0;
}

/*	Find the next token in the given string. A token is	*/
/*	any string of characters that is not a comma or NULL	*/
/*	or white-space.						*/
/*	We return a pointer to a copy of the token and we	*/
/*	remove the token from the original string.		*/
/*	The memory allocated for the returned copy should be	*/
/*	freed with free().					*/
/*	We return NULL if there are no tokens.			*/

static char *next_token(char *token_string)
{
	int tstart;
	int tend;
	char *ts;
	char *token;

/*	Find the start of the next token. The start is the next	*/
/*	character which is not a comma and not white space.	*/
	ts = token_string;
	while((*ts == ',') || isspace(*ts)) {
	    ts++;
	}

/*	No token found?						*/
	if(*ts == '\0') {
//		flogerror("  Found no token.");
		return(NULL);		/* We have no token.	*/
	}

	tstart = ts-token_string;	/* Index to token start	*/

/*	Now find the end of the token.				*/
/*	ts will point to the first character following the	*/
/*	token.							*/
	while((*ts != ',') && (isspace(*ts) == 0) && (*ts!= '\0')) {
//	flogerror("  Advance over char 0x%x = '%c'",0xff&(unsigned int)*ts,*ts);
		ts++;
	}

        tend = ts-token_string;		/* Index to token end	*/

/*	Copy the token with a terminating NULL byte.		*/
	token = local_strndup(token_string+tstart,tend-tstart);

/*	Now delete the token from the original token list.	*/
	while((*ts) != '\0') {
		*token_string = *ts;
		token_string++;
		ts++;
	}
	*token_string = '\0';
	return(token);
}


/*	Search the FITS card buffer 'fitsbuf' (which holds 'ncards'
 *	80-column card images) for the keyword pointed to by 'keyword'.
 *	If found return the card number (starting at 0) and a
 *	pointer to the card through 'card.' If not found return -1
 *	and return through 'card' the address of the byte following
 *	the last card in fitsbuf.
*/

int find_fits_card(char *fitsbuf,int ncards,char *keyword,char **card)
{
	int nc;
	int cnum = ncards;

	nc = strlen(keyword);

	if(nc > 8) {
	    flogerror("%s: Invalid FITS keyword %d bytes long: %s",
		__FUNCTION__,nc,keyword);
	    flogerror("      Truncating it to 8 characters");
	    nc = 8;
	    *(keyword+nc) = '\0';
	}

/*	find the card to replace or the end of the cards	*/
	while(ncards > 0) {
/*	    Make complete test only if first character matches.	*/
	    if(*fitsbuf == *keyword) {
/*		Make sure the whole keyword matches.		*/
	        if(strncmp(fitsbuf,keyword,nc) == 0) {
		    char next;

/*		    Make sure the card contains no additional	*/
/*		    keyword characters. The next character	*/
/*		    should be a space or the equal sign.	*/

		    next = *(fitsbuf+nc);
		    if((next == ' ') || (next == '=')) {
		        *card = fitsbuf;  /* A confirmed match.	*/
			return(cnum-ncards);
		    }
		}
	    }
	    fitsbuf+=80;
	    ncards--;
	}
	*card = fitsbuf;
	return(-1);
}

/*	Update (or add if not present) a FITs card to the OTHER	*/
/*	FITS cards for the given image id.			*/
void update_fits_card(int imageid,char *keyword,char ptype,
	void *param,char *comment)
{
	char cimage[81];
	int cnum;
	struct record *record;
	int n_other;
	char *card;

	record = RECORD_ADDR(imageid);
	n_other = record->fits_other;
	card = FITS_OTHER(imageid);

/*	find the card to replace or the end of the cards	*/
	
	cnum = find_fits_card(card,n_other,keyword,&card);

/*	If we are adding a card then increment the card count	*/
	if(cnum < 0)
	    record->fits_other++;

/*	Produce the FITS card in the cimage array.		*/
	fits_card_image(ptype,cimage,param,keyword,comment);

/*	Put the card into the list of cards.			*/
/*
	flogerror("update for id %d card: %30.30s",
		imageid,cimage);
*/
	memcpy(card,cimage,80);
}

/*	Produce a FITS card in the requested format.		*/
void fits_card_image(char ptype,char *cimage,void *param,char *keyword,
    char *comment)
{
	int *iparam = (int *)param;
	float *fparam = (float *)param;
	int pd;
	char *pc;
	int nc,ncom;
	char cmarker;
	char format[64];

/*	Generate a FITS card image.				*/
	if (ptype == 'd') {
/*		The parameter is an integer. It has a fixed format	*/
		pd = *iparam;
		sprintf(cimage,"%-8.8s= %20d / %-47.47s",
		    keyword,pd,comment);
	} else if (ptype == 'f') {
/*		The parameter is a float.				*/
		sprintf(cimage,"%-8.8s= %20.6f / %-47.47s",
                    keyword,*fparam,comment);
	} else if (ptype == 'T') {
/*		The parameter is a logical (T or F in FITS card).	*/
/*		The passed parameter should be either integer 1 or 0.	*/
/*		Actually anything non-0 is taken to be T.		*/
		pd = *iparam;
		sprintf(cimage,"%-8.8s= %20.20s / %-47.47s",
		    keyword,pd ? "T" : "F",comment);
	} else {
/*		The parameter is a string. We construct the format.	*/
		pc = (char *)param;
		nc = strlen(pc);
		if(nc < 18) nc = 18;
		ncom = 65-nc;
		if(strlen(comment) == 0)
		   cmarker = ' ';
		else
		   cmarker = '/';
		sprintf(format,"%%-8.8s= '%%%d.%ds' %c %%-%d.%ds",
		    nc,nc,cmarker,ncom,ncom);
		sprintf(cimage,format,keyword,pc,comment);
	}
}


int get_record_dir(void *dp)
{
	struct notify_handle *nh = dp;
	int id;
	struct record *record;


/*	first we get the image id				*/
	sscanf(nh->msg.row,"%d",&id);
	id %= 256;
	record = RECORD_ADDR(id);

/*	Next comes the directory				*/
	strcpy(record->dir,nh->msg.value);
	flogerror("New image directory %s",record->dir);

	return(1);
}
int get_record_root(void *dp)
{
	struct notify_handle *nh = dp;
	int id;
	struct record *record;

/*	first we get the image id				*/
	sscanf(nh->msg.row,"%d",&id);
	id %= 256;
	record = RECORD_ADDR(id);

/*	Copy the root of the file name				*/
	strcpy(record->filename,nh->msg.value);
	return(1);
}

int get_record_suffix(void *dp)
{
	struct notify_handle *nh = dp;
	int id;
	struct record *record;

/*	first we get the image id				*/
	sscanf(nh->msg.row,"%d",&id);
	id %= 256;
	record = RECORD_ADDR(id);

/*	Copy the suffix						*/
	strcpy(record->suffix,nh->msg.value);

	return(1);
}

char *pparse(struct notify_handle *nh)
{
	static char pinfo[1024];
	struct notify_msg *nm = &nh->msg;
	sprintf(pinfo,
	    "callback %x %s\ndb %s\ntable %s\nrow %s\ncol %s\nval %s",
	    (unsigned int)nm->callback,nm->callback_t,nm->database,nm->table,
	    nm->row,nm->column,nm->value);
	return(pinfo);
}

/*	Received a message telling us how many images to record	*/
/*	for a given image id.					*/
int get_record(void *dp)
{
	struct notify_handle *nh = dp;
	int id,count;
	struct record *record;
//	flogerror("in get_record id %s",nh->msg.row);
	sscanf(nh->msg.row,"%d",&id);
	sscanf(nh->msg.value,"%d",&count);
	id %= 256;

	record = RECORD_ADDR(id);

	record->count = count;

	logrecord("get_record",id,record);

	return(1);
}

/*	Received a message telling us the file  number for a	*/
/*	given image id. This number is used to contruct the	*/
/*	file name when an image with the given id is recorded.	*/

int get_record_number(void *dp)
{
	struct notify_handle *nh = dp;
	int id;
	int record_id;
	struct record *record;

	sscanf(nh->msg.row,"%d",&id);
	id %= 256;
	sscanf(nh->msg.value,"%d",&record_id);
	record = RECORD_ADDR(id);
	record->record_id = record_id;
/*
	logrecord("get_record_number",record_id,record);
*/

	return(1);
}

void logrecord(char *func,int id,struct record *record)
{
	flogerror("logrecord from %s record id %d address %x",func,id,record);
	flogerror("    count %d num %d dir %s root %s",
		record->count,record->record_id,record->dir,record->filename);
}

/*	This function is called whenever we get a notifier message	*/
/*	about some change of an item in the database.			*/
static int handle_notify_module(void *dp)
{
	struct notify_handle *nh = dp;
	char keysource[128];
	char keysource2[128];
	struct keydef *key;

/*	Construct the original database source in "dot" notation.	*/
	sprintf(keysource,"%s.%s.%s.%s",
		nh->msg.database,nh->msg.table,nh->msg.column,nh->msg.row);

/*	Find the keyword entry for this source.				*/
	key = find_keysource(keysource);

/*	If we didn't find a keyword entry then try to find one without	*/
/*	the row identifier.						*/
	if (key == NULL) {
	    sprintf(keysource2,"%s.%s.%s.",
		nh->msg.database,nh->msg.table,nh->msg.column);
	    key = find_keysource(keysource2);
	}

/*	This really should not happen...				*/
	if (key == NULL) {
	    flogerror("Got notify message %s but no matching keyword",
		keysource);
	    return 0;
	}

/*	Save the pointer to the notify message data.			*/
	key->data = (void *)nh;

/*	If there is a function to call (there better be!) to handle	*/
/*	this notify message then call it.				*/

	if(key->key_getter != NULL) {
	    return((*key->key_getter)(key));
	} else
	    return -1;
}

/*	Send a list of messages on message queue specified by msid	*/
int send_newim_msgs(int msid)
{
	static int m[] = {RECORD_NEWIM,DISPLAY_NEWIM,GUIDER_NEWIM,-1};
	int ret;

	ret = send_msgs_list(msid,m);

	return(ret);
}

/*	Send the DISPLAY_NEWIM message to the queue specified by msid)	*/
int send_display_newim_msg(int msid)
{
	static int m[] = {DISPLAY_NEWIM,-1};
	int ret;

	ret = send_msgs_list(msid,m);

	return(ret);
}

/*	Send the list of messages given by m to the queue specified	*/
/*	by msid.							*/
/*	Return -1 if any of the messages can't be sent.			*/
int send_msgs_list(int msid, int m[])
{
        int rb = -1;
        struct recordmsg {
                long record_type;
                char record_msg[100];
        } recordmsg;
	int mn;

        sprintf(recordmsg.record_msg,"%s %d %x %x %x","NEWIM",
	    process_shid,(unsigned int)process_buffer,
	    (unsigned int)process_image,(unsigned int)process_header);

	for(mn=0; m[mn] != -1; mn++) {

            recordmsg.record_type = m[mn];

/*	    If we are not in guider mode then don't bother sending	*/
/*	    a message saying an image is ready for guiding.		*/
	    if((recordmsg.record_type == GUIDER_NEWIM) && (guider_mode ==0))
		continue;

/*	    This might fail if the reading program isn't pulling	*/
/*	    messages out of the queue fast enough.			*/

            rb = msgsnd(msid,&recordmsg,strlen(recordmsg.record_msg)+1,
		IPC_NOWAIT);

	    if(rb < 0) return(rb);
	}

	return(rb);
}

char *check_msg(int msid,int msg_type)
{
        int rb;
        static struct rcdmsg {
                long rcd_type;
                char rcd_msg[100];
        } rcdmsg;

again:
        if((rb=msgrcv(msid,&rcdmsg,100,msg_type,IPC_NOWAIT)) >= 0) {
                rcdmsg.rcd_msg[rb] = '\0';
		return(rcdmsg.rcd_msg);
        }

/*	If we get to here got some sort of error.			*/
/*	If the error is an interrupted system call we try again.	*/
/*	An interrupted system call will happen if a timer expires.	*/

	if(errno == EINTR) goto again;

	return((char *)0);
}

/*	Wait for a message of the specified type on the specified	*/
/*	message queue. After we receive the first message try to	*/
/*	read any more that are waiting in the queue. Return the		*/
/*	message text from the last message.				*/
char *check_msg_wait(int msid,int msg_type)
{
        int rb;
        static struct rcdmsg {
                long rcd_type;
                char rcd_msg[100];
        } rcdmsg = {0,""};
	static char mbuf[100];
	char *rp;

again:
        if((rb=msgrcv(msid,&rcdmsg,100,msg_type,0)) >= 0) {
                rcdmsg.rcd_msg[rb] = '\0';
		strcpy(mbuf,rcdmsg.rcd_msg);
		while((rp=check_msg(msid,msg_type)) != NULL) {
                    strcpy(mbuf,rp);
		}
		return(mbuf);
        }

/*	If we get to here we got some sort of error. If the error is	*/
/*	an interrupted system call we just go back and try again.	*/
/*	An interrupted system call will happen if a timer expires.	*/

	if(errno == EINTR) goto again;
	return((char *)0);
}

/*	Input:  the message text received with the RECORD_NEWIM message.*/
/*		The text is expected to have the word NEWIM, four	*/
/*		integers which are the shared memory id, the base	*/
/*		address of the image buffer, the start of the image,	*/
/*		and the start of the header.				*/
/*	Output:	The four addresses are extracted. Note that these	*/
/*		addresses are sent from one of the camread processes.	*/
/*		But we are a forked version of the sending version	*/
/*		and all of the address were dynamically established	*/
/*		before the fork(). So the address are valid for us too.	*/

static void process_record_newim_msg(char *msg)
{
	static char msg_word[8];
/*
	flogerror("process_record_newim_msg msg addr %x",msg);
	flogerror("process_record_newim_msg msg: %s",msg);
*/
        sscanf(msg,"%s %d %x %x %x",msg_word,
	    &process_shid,(unsigned int *)&process_buffer,
	    (unsigned int *)&process_image,(unsigned int *)&process_header);
/*
	flogerror("process_record_newim_msg: image %x header %x",
		process_image,process_header);
*/

}

/*	Input:  the message text received with the request to record	*/
/*		images. The text is expected to have two integers,	*/
/*		the letter 'F' followed by a filename basis, a space,	*/
/*		the letter 'O' and then the object name.		*/
/*	Output:	The fileid, record_count, filename basis and object	*/
/*		name are extracted.					*/

static void process_record_msg(char *msg)
{
	char *object;
	char *file,*fileend;
	int nc;

//	flogerror("process_record_msg msg %s",msg);
/*	First extract the two integers.					*/
        sscanf(msg,"%d %d",&record_idnum,&record_count);

/*	Then find the start of the file name				*/
	file = index(msg,'F');
	file++;

/*	Find the end of the file name (a space).			*/
	fileend = index(file,' ');

/*	Copy the file name from the message.				*/
	memset(file_name,0,sizeof(file_name));
	if((nc=(fileend-file)) >= sizeof(file_name))
		nc = sizeof(file_name)-1;
	    
	strncpy(file_name,file,nc);

/*	Finally find the object name and copy it.			*/
	object = index(fileend,'O');
	object++;
	strcpy(object_name,object);

/*	printf("Recording %d '%s' images starting at number %d\n",
		record_count,object,record_idnum);
*/
}

int process_guider_msg(char *msg)
{
	int guider_flag;
	sscanf(msg,"%d",&guider_flag);
	return(guider_flag);
}


void recheck_sums(unsigned short *final,unsigned short *raw,int pixels)
{
	unsigned short csumf,csumr;
	int p;

	csumf = csumr = 0;
	for(p=0; p<pixels; p++) {
		csumf += *final;
		csumr += *raw;
		final++;
		raw++;
	}
	flogerror("recheck of %d pixel image sums raw=%u descrambled=%u",
		pixels,csumr,csumf);

}

/*	Descramble image data as they arrive.				*/
/*	The raw image is described by the header data contained in	*/
/*	the structure hdata.						*/

/*	We assume we are saving a rectangular subarray out of the	*/
/*	total array being read. It is assumed that some other software	*/
/*	has defined the subarray and has associated that subarray with	*/
/*	the image id. So we use the image id to look up the parameters	*/
/*	of the subarray. While descrambling the image we save only the	*/
/*	subarray into the output image.					*/

int descramble(unsigned short *image,struct header_data *hdata)
{
	unsigned short *sh_im,*sh_im2;	/* Pointers for output image.	*/
	unsigned short *sh_im2i;	/* Pointers for output image.	*/
	unsigned short *sh_imfirst     ;/* Pointers for output image.	*/
	unsigned short *sh_imlast      ;/* Pointers for output image.	*/
#define MAX_AMPS_PER_ROW 32
	unsigned short *sh_quad[MAX_AMPS_PER_ROW];
	unsigned short sh_row[MAX_AMPS_PER_ROW];
	unsigned short sh_col[MAX_AMPS_PER_ROW];
	short sh_step[MAX_AMPS_PER_ROW];
	
	int pixel_num;			/* Index into raw input image.	*/
	int row,col;
	int r,c;			/* Dimensions of raw image.	*/
	int c_overscan;			/* Column number of last	*/
					/*   overscan pixel (if any)	*/
	int c_over;			/* Overscan column number	*/
	int size;		/* size of final windowed image.	*/
	int headersize;
	int nrsave,ncsave;	/* # of rows/columns in saved image not	*/
				/* counting overscan rows or columns.	*/
	int nrover,ncover;	/* # of rows/columns in overscan.	*/
	int pixels;		/* # of pixels in final windowed image.	*/
	int xmit_pixels;	/* # of pixels transmitted from camera.	*/
	int n_amps;		/* # of amplifiers.			*/
#ifdef SUMTEST
	int sum=0,nsum=0;
#endif
	int count;		/* for pixel counting.			*/
	int fits_count;		/* FITS card count.			*/
	int readout_geometry;	/* Used to select the descrambling code	*/
	static int old_geom = -1;
	struct image_flips *image_flips;
	static int old_check_flag=-10;	/* Used to record previous	*/
					/* state of checksum_flag.	*/

/*	rows_to_save returns an array of boolean ints that say whether	*/
/*	or not a particular row is to be saved.				*/
	short *rsave,*rows_to_save(struct header_data *);

/*	columns_to_save returns an array of ints that say whether	*/
/*	or not a particular column is to be saved.			*/
	short *csave,*columns_to_save(struct header_data *);

#define saveit(r,c) ((*(rsave+(r)))&(*(csave+(c))))
#define saverow(r) (*(rsave+(r)))

/*	The checksum accumulator.					*/
	unsigned short checksum = 0;
	unsigned short cam_checksum;
/*
	flogerror("r=%d c=%d id=%d itype=%d headersize=%d",
		hdata->image.r,
		hdata->image.c,
		hdata->id,
		hdata->type,
		hdata->size);
	flogerror("exposure %5.2f shutter %s",(float)hdata->exposure/100.,
		hdata->shutter ? "opened" : "closed");
*/
/*	Log the status of the camera checksum.				*/
	if(old_check_flag != hdata->checksum_flag) {
	    old_check_flag = hdata->checksum_flag;
	    flogerror("Camera images %s contain a checksum.",
		(hdata->checksum_flag ? "do": "do not"));
	}

/*	Get the array of flags that indicate which rows and columns	*/
/*	of the image are to be saved.					*/

	rsave = rows_to_save(hdata);
	csave = columns_to_save(hdata);
	nrover = hdata->over.r*hdata->amps_per_column;
	ncover = hdata->over.c*hdata->amps_per_row;

/*	Get the number of rows and columns in the saved image.		*/
	nrsave = hdata->window.r;
	ncsave = hdata->window.c;

	if ((hdata->image.r < 2) || (hdata->image.r > maxrc)) {
	   flogerror("descram bad row size %d",hdata->image.r);
	   return 0;
	}

	if ((hdata->image.c < 2) || (hdata->image.c > maxrc)) {
	   flogerror("descram bad column size %d",hdata->image.c);
	   return 0;
	}

	r = hdata->image.r;
	c = hdata->image.c;
	pixels = (nrsave + nrover) * (ncsave + ncover);
	xmit_pixels = r*c;

/*	Number of data channels to unscramble.			*/
	n_amps = hdata->amps_per_row*hdata->amps_per_column;

/*
	flogerror("final image will be %dRx%dC plus overscan %dR and %dC,"
		" total pixels %d",
		nrsave,ncsave,nrover,ncover,pixels);
*/
	if(report_progress) {
	    send_progress_report(xmit_pixels,0);
	}
	size = pixels*sizeof(short);
	if (size > shm_nbytes) {
	   flogerror("image size too big: %d > %d", size,shm_nbytes);
	   return 0;
	}

/*	Convert header size from bytes to shorts.		*/
	headersize = hdata->size/2;

/*	If we have a checksum from the camera it is stored in	*/
/*	the unsigned short immediately following the image.	*/

/*	The initial count should include the header.		*/
/*	Adjust byte available count by the header size.		*/
	count = wait_for_data("1",0);
	if (count < 0) return 0;		/* Error!	*/
	count -= headersize;

/*	Now descramble the image based on the readout geometry	*/
/*	which is defined by the image 'type'.			*/

//	flogerror("descramble type %d",hdata->type);

/*	Get the readout geometry indicator. Note that we may	*/
/*	modify this if we have requests for flipping the image.	*/

	image_flips = FLIP_ADDR;
	image_flips->geometry = hdata->type;
    	readout_geometry = image_flips->geometry;

/*	If we are simulating an image then initialize that process.	*/
/*	Otherwise check for image-flipping, which may modify the	*/
/*	effective readout geometry specifier.				*/

	if(simulation) {
	    simulate_image(simulation,readout_geometry,headersize,r,c);
	} else {
	    readout_geometry = flip_options(image_flips);
	}

	if(old_geom != readout_geometry) {
		flogerror("switching from geometry %d to %d",
			old_geom,readout_geometry);
		old_geom = readout_geometry;
	}

/*	Fill in the image FITS header				*/
	fits_count = fits_cards(hdata,(char *)header,image_flips);

//	flogerror("USING headersize=%d",headersize);
	switch(readout_geometry) {
	case 0:		/* One channel from corner (0,0)	*/
/*	    Rows and columns are saved in forward order.	*/
	    sh_im = (unsigned short *)mem_image;
	    pixel_num=headersize;
	    for(row=0; row<r; row++) {
	      for(col=0; col<c; col++) {
		if (count <= 0) {
		    if(report_progress)
	    		send_progress_report(xmit_pixels,pixel_num);
		    count = wait_for_data("2",pixel_num);
		    if (count < 0) return 0;	/* Error!	*/
		}
		checksum += image[pixel_num];
		if(saveit(row,col)) {
		    *sh_im = image[pixel_num];
		    sh_im++;
		}
		pixel_num++;
		count--;
	      }
	    }

	    break;
/*	Case 50 is for special baseline calibration image. The actual	*/
/*	descrambling algorithm doesn't matter because all of the pixels	*/
/*	are baseline pixels.						*/
	case 50:		/* One channel from corner (C,0)	*/
/*	    The pixels are saved in reverse column order.	*/
	    pixel_num = headersize;
	    c = hdata->image.c-ncover;
	    sh_im = (unsigned short *)mem_image + ncsave - 1;
	    for(row=0; row<r; row++) {
		for(col=c-1; col>=0; col--) {
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("3",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = simulation ? sim_im[pixel_num] : \
				image[pixel_num];
			    sh_im--;
			}
			pixel_num++;
			count--;
		}

/*		Only if we've saved something from this row do we	*/
/*		advance the address for the saved image.		*/

		if(saverow(row)) {
		    sh_im += ncsave + 1;
		}

/*		Transfer the overscan to the image.			*/
		for(col=c; col<c+ncover; col++) {
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("50",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
		}
			
/*		Only if we've saved something from this row do we	*/
/*		advance the address for the saved image.		*/

		if(saverow(row)) {
		    sh_im += ncsave-1;
		}

	    }
	    break;


      	case 13: /* Two channel read out from corner(0,0) and (0,R)  (add by Roy) */
	    sh_im = (unsigned short *)mem_image;
	    sh_im2 =(unsigned short *)mem_image+pixels-
				(nrover*(ncsave+ncover))-(ncsave+ncover);
	    pixel_num=headersize;
	    
	    for(row=0; row<r/2; row++) {
	      for(col=0; col<c; col++) {
		if (count <= 0) {
		    if(report_progress)
	    		send_progress_report(xmit_pixels,pixel_num);
		    count = wait_for_data("23",pixel_num);
		    if (count < 0) return 0;	/* Error!	*/
		}

	       /*Read from up-left coner. */
		checksum += image[pixel_num];
		if(saveit(row,col)) {
		    *sh_im = image[pixel_num];
		    sh_im++;
		}
		pixel_num++;
		count--;
	
	      /*Read from down-left coner. */
              if (count <= 0) {
		    if(report_progress)
	    		send_progress_report(xmit_pixels,pixel_num);
		    count = wait_for_data("24",pixel_num);
		    if (count < 0) return 0;	/* Error!*/	
	      } 
           
		checksum += image[pixel_num];
		if(saveit(r-1-row,col)) {
		    *sh_im2 = image[pixel_num];
		    sh_im2++;
		}
		pixel_num++;
		count--;
	      
	    }

	   if(saverow(r-1-row))	{
	   	sh_im2-=2*(ncsave+ncover);
	   }

          }
	    break;



	case 14:		/* Two channels from corner (C,0) and (C,R)	*/
/*	    The pixels are saved in reverse column order.	*/
	    pixel_num = headersize;
	    c = hdata->image.c-ncover;
	    sh_im = (unsigned short *)mem_image + ncsave - 1;
            sh_im2 =(unsigned short *)mem_image + pixels - 
                                (nrover*(ncsave+ncover))-ncover-1;
	    for(row=0; row<r/2; row++) {
		        
                        printf("| %d | ",row);
		for(col=c-1; col>=0; col--) {
                        /*First pixel from upper right*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("33",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
                       // printf("| %d | ",col);
			checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im--;
			}
			pixel_num++;
			count--;
                        /*Second pixel from down right*/
                        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("34",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
                       checksum += image[pixel_num];
                       if(saveit(r-row-1,col)) {
                            *sh_im2 = image[pixel_num];
                             sh_im2--;
	               }
			pixel_num++;
			count--;
                }
/*		Only if we've saved something from this row do we	*/
/*		advance the address for the saved image.		*/
               
		if(saverow(row)) {
		    sh_im += ncsave + 1;
		}

                if(saverow(r-row-1)) {
                    sh_im2 += ncsave + 1;
                }
                
/*		Transfer the overscan to the image.			*/
		for(col=c; col<c+ncover; col++) {
                        /*Transfer upper right over sacn pixel first*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("35",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
                        /*Then the down right over scan pixel */
                        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("36",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
                        checksum += image[pixel_num];
                        if(saveit(r-row-1,col)) {
                            *sh_im2 = image[pixel_num];
                            sh_im2++;
                        }
                        pixel_num++;
                        count--;
		}
			
/*		Only if we've saved something from this row do we	*/
/*		advance the address for the saved image.		*/

		if(saverow(row)) {
		    sh_im += ncsave-1;
		}
                if(saverow(r-row-1)) {
                    sh_im2 -=(ncsave+2*ncover+1);
                }
            
            
      }
	    break;


	case 11:     /* Two channel from corners (0,0) and (C,R)*/
/*	    The saved pixels are stored in forward-row order.	*/
/*	    The even numbers pixels are saved in forward	*/
/*	    column order.					*/
/*	    The odd numbers pixels are saved in reverse		*/
/*	    column order.					*/

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    We are going to separate out any overscan pixels	*/
/*	    and place them at the right side of the main array.	*/

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - ncover;

/*	    Address of first saved image pixel.			*/
	    sh_im = (unsigned short *)mem_image;
            sh_im2 =(unsigned short *)mem_image+pixels-
                           nrover*(ncsave+ncover)-ncover-1;
	    for(row=0; row<r/2; row++) {
/*		Copy main part of the image.				*/
		for(col=0; col<c; col++) {

/*			First pixel is right down amp.			*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("11a",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(r-1-row,c-col-1)) {
			    *sh_im2 = image[pixel_num];
			    sh_im2--;
			}
			pixel_num++;
			count--;

/*			Next pixel is left up amp.				*/

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("11b",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
		}
		if(saverow(row)) {
		  //  sh_im = sh_im + 1;
		    sh_im2 = sh_im2 + ncsave + ncover;
		}

/*		Copy the overscan portion of the image.			*/
		c_over = c_overscan;

		for(col=0; col<ncover; col++) {

/*			First is right amp pixel			*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("11c",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(r-1-row,c+ncover-1-col)) {
			    *sh_im2 = image[pixel_num];
			    sh_im2--;
			}
			c_over--;
			pixel_num++;
			count--;

/*			Next is left amp pixel.				*/

			if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
			    count = wait_for_data("11d",pixel_num);
			    if (count < 0) return 0;    /* Error!	*/
			}
		    	checksum += image[pixel_num];
			if(saveit(row,c+col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
		}
		if(saverow(row)) {
		//    sh_im = sh_im + 1;
                    sh_im2 = sh_im2 - (ncover + ncsave); 
		}
	    }
	    break;

     case 12:     /* Two channel from corners (C,0) and (0,R)	*/
/*	    The saved pixels are stored in forward-row order.	*/
/*	    The even numbers pixels are saved in forward	*/
/*	    column order.					*/
/*	    The odd numbers pixels are saved in reverse		*/
/*	    column order.					*/

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;
//	memset(mem_image,0xff,r*c*2);

/*	    We are going to separate out any overscan pixels	*/
/*	    and place them at the right side of the main array.	*/

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - ncover;

/*	    Address of first saved image pixel.			*/
	    sh_im = (unsigned short *)mem_image + pixels - 
                           nrover*(ncsave+ncover)-(ncsave+ncover);
            sh_im2 =(unsigned short *)mem_image + ncsave - 1;
	    for(row=0; row<r/2; row++) {
/*		Copy main part of the image.				*/
		for(col=0; col<c; col++) {

/*			First pixel is right up amp.			*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("12a",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(row,c-col-1)) {
			    *sh_im2 = image[pixel_num];
			    sh_im2--;
			}

			pixel_num++;
			count--;

/*			Next pixel is lower left amp.			*/

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("12b",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(r-1-row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
		}
		if(saverow(row)) {
		   // sh_im = sh_im + 2*hdata->over.c-1;
		    sh_im2 = sh_im2 + ncsave + 1;
		}

/*		Copy the overscan portion of the image.			*/
		c_over = c_overscan;

		for(col=c; col<c+ncover; col++) {

/*			First is up  amp pixel			*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("12c",pixel_num);
		            if (count < 0) return 0;	
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im2 = image[pixel_num];
			    sh_im2++;
			}
			c_over--;
			pixel_num++;
			count--;

/*			Next is down amp pixel.				*/

			if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
			    count = wait_for_data("12d",pixel_num);
			    if (count < 0) return 0;    /* Error!	*/
			}
		    	checksum += image[pixel_num];
			if(saveit(r-1-row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
		}
		if(saverow(row)) {
		    sh_im = sh_im - 2*(ncsave+ncover);
                    sh_im2 = sh_im2 + ncsave - 1; 
		}
	    }
	    break;



	case 1:		/* One channel from corner (C,0)	*/
/*	    The pixels are saved in reverse column order.	*/
	    pixel_num = headersize;
	    c = hdata->image.c-ncover;
	    sh_im = (unsigned short *)mem_image + ncsave - 1;
	    for(row=0; row<r; row++) {
		for(col=c-1; col>=0; col--) {
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("1a",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = simulation ? sim_im[pixel_num] : \
				image[pixel_num];
			    sh_im--;
			}
			pixel_num++;
			count--;
		}

/*		Only if we've saved something from this row do we	*/
/*		advance the address for the saved image.		*/

		if(saverow(row)) {
		    sh_im += ncsave + 1;
		}

/*		Transfer the overscan to the image.			*/
		for(col=c; col<c+ncover; col++) {
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("1b",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
		}
			
/*		Only if we've saved something from this row do we	*/
/*		advance the address for the saved image.		*/

		if(saverow(row)) {
		    sh_im += ncsave-1;
		}

	    }
	    break;

	case 2:		/* One channel from corner (0,R)	*/
/*	    The saved pixels are stored in reverse-row order.	*/

/*	    sh_im is the address of the destination of the first pixel.	*/

	    sh_im = (unsigned short *)mem_image + pixels -
		(nrover*(ncsave+ncover))-(ncsave+ncover);

	    pixel_num=headersize;
	    for(row=r-1; row>=0; row--) {
	      for(col=0; col<c; col++) {
		if (count <= 0) {
		    if(report_progress)
	    		send_progress_report(xmit_pixels,pixel_num);
		    count = wait_for_data("2a",pixel_num);
		    if (count < 0) return 0;	/* Error!	*/
		}
		checksum += image[pixel_num];
		if(saveit(row,col)) {
		    *sh_im = image[pixel_num];
		    sh_im++;
		}
		pixel_num++;
		count--;
	      }

/*	      Only if we've saved something from this row do we	*/
/*	      modify the address for the saved image.		*/

	      if(saverow(row)) {
		sh_im -= (ncsave+ncover)*2;
	      }
	    }
	    break;

	case 3:		/* One channel from corner (C,R)	*/
/*	    The saved pixels are stored in reverse column order.	*/
/*	    The saved pixels are stored in reverse-row order.	*/
	    pixel_num = headersize;

/*	    Exclude column overscan from column count.			*/
	    c = hdata->image.c - ncover;

/*	    sh_im is the address of the destination of the first pixel.	*/

	    sh_im = (unsigned short *)mem_image + pixels -
		(nrover*(ncsave+ncover)) - ncover -1;

	    for(row=r-1; row>=0; row--) {
		for(col=c-1; col>=0; col--) {
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("3a",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im--;
			}
			pixel_num++;
			count--;
		}

/*		Only if we've saved something from this row do	*/
/*		we modify the address for the saved image.	*/

		if(ncover > 0) {
		    if(saverow(row)) {
		        sh_im += ncsave+1;
		    }

/*		    Transfer the overscan to the image.			*/
		    for(col=c; col<c+ncover; col++) {
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("3b",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
		    }
			
/*		    Only if we've saved something from this row do we	*/
/*		    modify the address for the saved image.		*/

		    if(saverow(row)) {
		        sh_im -= (ncsave+2*ncover+1);
		    }
		}
	    }
	    break;

	case 4:     /* Two channel from corners (C,0) and (0,0)	*/
/*	    The saved pixels are stored in forward-row order.	*/
/*	    The even numbers pixels are saved in reverse	*/
/*	    column order.					*/
/*	    The odd numbers pixels are saved in forward		*/
/*	    column order.					*/

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    We are going to separate out any overscan pixels	*/
/*	    and place them at the right side of the main array.	*/

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - 2*hdata->over.c;

/*	    Address of first saved image pixel.			*/
	    sh_im = (unsigned short *)mem_image;

	    for(row=0; row<r; row++) {

	        sh_im2i = sh_im2 = sh_im + ncsave - 1;

/*		Copy main part of the image.				*/
		for(col=0; col <(c/2); col++) {

/*			First pixel is right amp.			*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("8",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(row,(c-col-1))) {
			    *sh_im2 = image[pixel_num];
			    sh_im2--;
			}

			pixel_num++;
			count--;

/*			Next pixel is left amp.				*/

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("9",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
		}
		if(saverow(row)) {
		    sh_im = sh_im2i + 1;
		    sh_im2i = sh_im2 = sh_im + hdata->over.c*2 -1;
		}

/*		Copy the overscan portion of the image.			*/
		c_over = c_overscan;

		for(col=0; col<hdata->over.c; col++) {

/*			First is right amp pixel			*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("12",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,c_over)) {
			    *sh_im2 = image[pixel_num];
			    sh_im2--;
			}
			c_over--;
			pixel_num++;
			count--;

/*			Next is left amp pixel.				*/

			if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
			    count = wait_for_data("13",pixel_num);
			    if (count < 0) return 0;    /* Error!	*/
			}
		    	checksum += image[pixel_num];
			if(saveit(row,c+col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
		}
		if(saverow(row)) {
		    sh_im = sh_im2i + 1;
		}
	    }
	    break;
	case 5:		/* Two channels from corners (0,0) and (C,0)	*/
/*	    The saved pixels are stored in forward-row order.	*/
/*	    The odd numbers pixels are saved in reverse		*/
/*	    column order.					*/
/*	    The even numbers pixels are saved in forward	*/
/*	    column order.					*/

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    We are going to separate out any overscan pixels	*/
/*	    and place them at the right side of the main array.	*/

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - 2*hdata->over.c;

/*	    Address of first saved image pixel.			*/
	    sh_im = (unsigned short *)mem_image;

	    for(row=0; row<r; row++) {

	        sh_im2i = sh_im2 = sh_im + ncsave - 1;

/*		Copy main part of the image.				*/
		for(col=0; col <(c/2); col++) {

/*			First the left pixel.				*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("20",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;

/*			Then the right pixel.				*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("19",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,(c-col-1))) {
			    *sh_im2 = image[pixel_num];
			    sh_im2--;
			}
			pixel_num++;
			count--;
		}
		if(saverow(row)) {
		    sh_im = sh_im2i + 1;
		    sh_im2i = sh_im2 = sh_im + hdata->over.c*2 -1;
		}

/*		Copy the overscan portion of the image.			*/
		c_over = c_overscan;
		for(col=0; col<hdata->over.c; col++) {

/*			First is left pixel.				*/
			if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
			    count = wait_for_data("24",pixel_num);
			    if (count < 0) return 0;    /* Error!	*/
			}
		    	checksum += image[pixel_num];
			if(saveit(row,(c+col))) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;

/*			Next is right pixel				*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("23",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,c_over)) {
			    *sh_im2 = image[pixel_num];
			    sh_im2--;
			}
			pixel_num++;
			count--;
			c_over--;

		}
		if(saverow(row)) {
		    sh_im = sh_im2i + 1;
		}
	    }
	    break;

	case 6:     /* Two channel from corners (C,R) and (0,R)	*/
/*	    The saved pixels are stored in reverse-row order.	*/
/*	    The even numbered pixels are saved in reverse	*/
/*	    column order.					*/
/*	    The odd numbered pixels are saved in forward	*/
/*	    column order.					*/

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    We are going to separate out any overscan pixels	*/
/*	    and place them at the right side of the main array.	*/

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - 2*hdata->over.c;

/*	    Address of first saved image pixel.			*/
/*	    This expression maybe have no consider on nrover (commented out By Roy)
	    sh_im = (unsigned short *)mem_image + pixels -
		ncsave - 2*hdata->over.c;			*/
  	    
	    sh_im =(unsigned short *)mem_image+pixels-
			(nrover*(ncsave+ncover))-(ncsave+ncover);

	    for(row=r-1; row>=0; row--) {

	        sh_im2i = sh_im2 = sh_im + ncsave - 1;

/*		Copy main part of the image.				*/
		for(col=0; col <(c/2); col++) {

/*			First pixel is right amp.			*/
		        if (count <= 0) {
		    	    if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("8",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,(c-col-1))) {
			    *sh_im2 = image[pixel_num];
			    sh_im2--;
			}

			pixel_num++;
			count--;

/*			Next pixel is left amp.				*/

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("9",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
		}
		if(saverow(row)) {
		    sh_im = sh_im2i + 1;
		    sh_im2i = sh_im2 = sh_im + hdata->over.c*2 -1;
		}

/*		Copy the overscan portion of the image.			*/
		c_over = c_overscan;
		for(col=0; col<hdata->over.c; col++) {

/*			First is right amp pixel			*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("12",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,c_over)) {
			    *sh_im2 = image[pixel_num];
			    sh_im2--;
			}
			c_over--;
			pixel_num++;
			count--;

/*			Next is left amp pixel.				*/
			if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
			    count = wait_for_data("13",pixel_num);
			    if (count < 0) return 0;    /* Error!	*/
			}
		    	checksum += image[pixel_num];
			if(saveit(row,c+col)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			pixel_num++;
			count--;
		}

/*		Move memory pointer back to start of next row.	*/

		if(saverow(row)) {
		    sh_im = sh_im - (2*ncsave + 3*hdata->over.c);
		}
	    }
	    break;

	case 7:     /* Two channel from corners (0,R) and (C,R)	*/
/*	    The saved pixels are stored in reverse-row order.	*/
/*	    The odd numbers pixels are saved in reverse		*/
/*	    column order.					*/
/*	    The even numbers pixels are saved in forward	*/
/*	    column order.					*/

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    We are going to separate out any overscan pixels	*/
/*	    and place them at the right side of the main array.	*/

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - 2*hdata->over.c;

/*	    Address of first saved image pixel.			*/
	    sh_im = (unsigned short *)mem_image + pixels -
		ncsave - 2*hdata->over.c;

	    for(row=r-1; row>=0; row--) {

	        sh_im2i = sh_im2 = sh_im + ncsave - 1;

/*		Copy main part of the image.				*/
		for(col=0; col <(c/2); col++) {

/*			First pixel is right amp.			*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("8",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,(c-col-1))) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}

			pixel_num++;
			count--;

/*			Next pixel is left amp.				*/

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("9",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,col)) {
			    *sh_im2 = image[pixel_num];
			    sh_im2--;
			}
			pixel_num++;
			count--;
		}
		if(saverow(row)) {
		    sh_im = sh_im2i + 1;
		    sh_im2i = sh_im2 = sh_im + hdata->over.c*2 -1;
		}

/*		Copy the overscan portion of the image.			*/
		c_over = c_overscan;
		for(col=0; col<hdata->over.c; col++) {

/*			First is right amp pixel			*/
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("12",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(row,c_over)) {
			    *sh_im = image[pixel_num];
			    sh_im++;
			}
			c_over--;
			pixel_num++;
			count--;

/*			Next is left amp pixel.				*/

			if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
			    count = wait_for_data("13",pixel_num);
			    if (count < 0) return 0;    /* Error!	*/
			}
		    	checksum += image[pixel_num];
			if(saveit(row,c+col)) {
			    *sh_im2 = image[pixel_num];
			    sh_im2--;
			}
			pixel_num++;
			count--;
		}

/*		Move memory pointer back to start of next row.	*/

		if(saverow(row)) {
		    sh_im = sh_im - (2*ncsave + 3*hdata->over.c);
		}
	    }
	    break;
	case 8:

/*	    Four channels from corners (0,0) and (C,0) (R,0) and (R,C)	*/
/*	    The data order is upper-right, upper-left, lower-right,	*/
/*	    lower-left.							*/

/*	    X-----|-----X						*/
/*	    |     |     |						*/
/*	    |  2  |  1  |						*/
/*	    |     |     |						*/
/*	    |-----|-----|						*/
/*	    |     |     |						*/
/*	    |  4  |  3  |						*/
/*	    |     |     |						*/
/*	    |-----|-----|						*/

/*	    The even numbered rows are stored in forward-row order.	*/
/*	    The odd numbered rows are stored in reverse-row order.	*/
/*	    The even numbers pixels are saved in reverse column order.	*/
/*	    The odd numbers pixels are saved in forward column order.	*/
/*	    The row overscan pixels are saved at the end of the primary	*/
/*	    image array and the column overscan pixels are saved to the	*/
/*	    right of the primary image array.				*/
/*	    IIIIIIIIIIIIIIIICCCCCCCC					*/
/*	    IIIIIIIIIIIIIIIICCCCCCCC					*/
/*	    IIIIIIIIIIIIIIIICCCCCCCC					*/
/*	    IIIIIIIIIIIIIIIICCCCCCCC					*/
/*	    IIIIIIIIIIIIIIIICCCCCCCC					*/
/*	    IIIIIIIIIIIIIIIICCCCCCCC					*/
/*	    IIIIIIIIIIIIIIIICCCCCCCC					*/
/*	    IIIIIIIIIIIIIIIICCCCCCCC					*/
/*	    RRRRRRRRRRRRRRRRCCCCCCCC					*/
/*	    RRRRRRRRRRRRRRRRCCCCCCCC					*/
/*	    RRRRRRRRRRRRRRRRCCCCCCCC					*/
/*	    RRRRRRRRRRRRRRRRCCCCCCCC					*/

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - ncover;

/*	    Change r to be the number of rows without		*/
/*	    overscan. (Unchanged if there is no overscan rows.)	*/
	    r = r - nrover;

/*	    Address of last saved image pixel (not including overscan). */
	    sh_imlast = (unsigned short *)mem_image + \
				(ncsave+ncover)*r - 1 - ncover;

/*	    Define how the column numbers advance for each quadrant.	*/
	    sh_step[0] = -1;
	    sh_step[1] = 1;
	    sh_step[2] = -1;
	    sh_step[3] = 1;

	    for(row=0; row<r/2; row++) {
		int a;

		sh_quad[1] = (unsigned short *)mem_image + row*(ncsave+ncover);
		sh_quad[0] = sh_quad[1] + ncsave - 1;
		sh_quad[2] = sh_imlast - row*(ncsave+ncover);
		sh_quad[3] = sh_quad[2] - ncsave + 1;

		sh_row[0] = sh_row[1] = row;
		sh_row[2] = sh_row[3] = r-row-1;

/*		Copy main part of the image.				*/
		for(col=0; col <(c/2); col++) {

		    sh_col[0] = c-col-1;
		    sh_col[1] = col;
		    sh_col[2] = sh_col[0];
		    sh_col[3] = sh_col[1];

/*		    Process a group of four pixels.			*/
		    for(a=0; a<4; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("8a",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }

			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a] ;
			}

			pixel_num++;
			count--;

		    }
		}
		if(hdata->over.c && saverow(sh_row[0])) {
			sh_quad[0] = (unsigned short *)mem_image +
			    (row+1)*(ncsave+ncover) -1;
			sh_quad[1] = sh_quad[0] - ncover + 1;
		}
		if(hdata->over.c && saverow(sh_row[2])) {
			sh_quad[2] = sh_imlast + ncover -
				row*(ncsave+ncover);
			sh_quad[3] = sh_quad[2] - ncover + 1;
		}

/*		Copy the overscan portion of the image.			*/

		for(col=0; col<hdata->over.c; col++) {

		    sh_col[0] = c_overscan-col;
//		    sh_col[1] = sh_col[0]-ncover+1+col;
		    sh_col[1] = c+col;
		    sh_col[2] = sh_col[0];
		    sh_col[3] = sh_col[1];

/*		    Process a group of four pixels.			*/
		    for(a=0; a<4; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("8b",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}
			pixel_num++;
			count--;

		    }
		}

	    }

/*	    Now we save the row overscan pixels, if there are any.	*/

/*	    Address of first saved row overscan pixel			*/
	    sh_imfirst = sh_imlast + ncover;

/*	    Address of last saved row overscan pixel			*/
	    sh_imlast = sh_imfirst + (ncsave+ncover)*nrover -ncover - 1;

	    for(row=0; row<hdata->over.r; row++) {
		int a;

		sh_quad[1] = sh_imfirst + row*(ncsave+ncover);
		sh_quad[0] = sh_quad[1] + ncsave - 1;
		sh_quad[2] = sh_imlast - row*(ncsave+ncover);
		sh_quad[3] = sh_quad[2] - ncsave + 1;

		sh_row[0] = sh_row[1] = r + row;
		sh_row[2] = sh_row[3] = r + nrover - row - 1;


/*		Copy main part of the row overscan.			*/
		for(col=0; col <(c/2); col++) {

		    sh_col[0] = c-col-1;
		    sh_col[1] = col;
		    sh_col[2] = sh_col[0];
		    sh_col[3] = sh_col[1];

/*		    Process a group of four pixels.			*/
		    for (a=0; a<4; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("8c",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}

			pixel_num++;
			count--;

		    }
		}
		if(hdata->over.c && saverow(sh_row[0])) {
			sh_quad[0] = sh_imfirst +
			    (row+1)*(ncsave+ncover) -1;
			sh_quad[1] = sh_quad[0] - ncover + 1;
		}
		if(hdata->over.c && saverow(sh_row[2])) {
			sh_quad[2] = sh_imlast + ncover -
				row*(ncsave+ncover);
			sh_quad[3] = sh_quad[2] - ncover + 1;
		}

/*		Copy the overscan portion of the image.			*/

		for(col=0; col<hdata->over.c; col++) {

		    sh_col[0] = c_overscan-col;
		    sh_col[1] = c+col;
		    sh_col[2] = sh_col[0];
		    sh_col[3] = sh_col[1];

/*		    Process a group of four pixels.			*/
		    for(a=0; a<4; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("8d",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}
			pixel_num++;
			count--;

		    }
		}
	    }
	    break;

	case 10:

/*	This is 4-channel readout of the Lincoln AO CCID56 or any other	*/
/*	device with multiple image segments but a single-amplifier	*/
/*	readout per sigment. The amplifier is assumed to be on the	*/
/*	left end of the serial register. This descramble option assumes	*/
/*	two segments from the top half of the array and two segments	*/
/*	from the bottom half of the array. The pixels arrive in groups	*/
/*	of four pixels, with one pixel from each segments in the order	*/
/*	show below. The first pixel from each segment is indicated by	*/
/*	the X.								*/

/*	|X-------|X-------|						*/
/*	|        |        |						*/
/*	|    2   |   1    |						*/
/*	|        |        |						*/
/*	|--------|--------|						*/
/*	|        |        |						*/
/*	|    4   |   3    |						*/
/*	|        |        |						*/
/*	|X-------|X-------|						*/

/*	This case (10) is actually written to handle any number of	*/
/*	amplifiers per row. For instance one could have three upper	*/
/*	quadrants and three lower quadrants.				*/

	    if (hdata->amps_per_row > MAX_AMPS_PER_ROW) {
		flogerror(
	"camread is configured for a maximum of %d amplifiers.",
		MAX_AMPS_PER_ROW);
		flogerror(
	"RECOMPILE WITH sh_quad[%d] increased to at least %d",
		MAX_AMPS_PER_ROW,hdata->amps_per_row);

/*		This probably won't work for descrambling the image	*/
/*		but it should prevent over-running the array.		*/

		hdata->amps_per_row = MAX_AMPS_PER_ROW;
	    }

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - ncover;

/*	    Change r to be the number of rows without		*/
/*	    overscan. (Unchanged if there is no overscan rows.)	*/
	    r = r - nrover;

/*	    Define how the column numbers advance for each quadrant.	*/
	    for(row=0;row<n_amps;row++) {
	        sh_step[row] = 1;		/* All advance forward.	*/
	    }

	    for(row=0; row<r/2; row++) {
		int a,quad_size;
		unsigned short *sh_end = NULL;

		quad_size = ncsave/hdata->amps_per_row;

/*		Define upper quadrant addresses and row number.		*/

		sh_quad[0] = (unsigned short *)mem_image +
			row*(ncsave+ncover) +
			(hdata->amps_per_row-1)*quad_size;

		for (a=1; a <hdata->amps_per_row; a++) {
		    sh_quad[a] = sh_quad[a-1] - quad_size;
		}
		for (a=0; a<hdata->amps_per_row; a++) {
		    sh_row[a] = row;
		}

/*		Define lower quadrant addresses and row number.		*/

		sh_quad[hdata->amps_per_row] = sh_quad[0] +
			(nrsave/2)*(ncsave+ncover);
		for (a=hdata->amps_per_row+1; a<n_amps; a++) {
		    sh_quad[a] = sh_quad[a-1] - quad_size;
		}
		for (a=hdata->amps_per_row; a<n_amps; a++) {
		    sh_row[a] = r-row-1;
		}

/*		Copy main part of the image.				*/

		for(col=0; col <(c/hdata->amps_per_row); col++) {

		    for(a=0; a<hdata->amps_per_row;a++) {
			sh_col[a] = c-(a+1)*quad_size+col;
			sh_col[a+hdata->amps_per_row] = sh_col[a];
		    }
			
/*		    Process group of n_amps pixels.			*/

		    for(a=0; a<n_amps; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("9a",pixel_num);
		            if (count < 0) {
				return 0;	/* Error!*/
			    }
		        }
			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}

			pixel_num++;
			count--;

		    }
		}
/*		Copy the overscan portion of the image.			*/

		if(hdata->over.c && saverow(sh_row[0])) {
		    sh_end = (unsigned short *)mem_image +
			(row+1)*(ncsave+ncover);
		    for(a=0; a<hdata->amps_per_row;a++) {
			sh_quad[a] = sh_end - (a+1)* hdata->over.c;
		    }
		}
		if(hdata->over.c && saverow(sh_row[hdata->amps_per_row])) {
		    sh_end = sh_end +
			(nrsave/2)*(ncsave+ncover);
		    for(a=0;a<hdata->amps_per_row;a++) {
			sh_quad[a+hdata->amps_per_row] =
				sh_end - (a+1)* hdata->over.c;
		    }
		}

		for(col=0; col<hdata->over.c; col++) {

		    for(a=0; a<hdata->amps_per_row;a++) {
			sh_col[a] = c-(a+1)*quad_size+col;
			sh_col[a+hdata->amps_per_row] = sh_col[a];
		    }
			
/*		    Process group of n_amps pixels.			*/
		    for(a=0; a<n_amps; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("9b",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}
			pixel_num++;
			count--;

		    }
		}

	    }

/*	    Now we save the row overscan pixels, if there are any.	*/

/*	    Address of first saved row overscan pixel			*/

	    sh_imfirst = (unsigned short *)mem_image + \
				(ncsave+ncover)*nrsave;

	    for(row=0; row<hdata->over.r; row++) {
		int a,quad_size;
		unsigned short *sh_end = NULL;

		quad_size = ncsave/hdata->amps_per_row;

/*		Define upper quadrant addresses and row number.		*/

		sh_quad[0] = (unsigned short *)sh_imfirst +
			row*(ncsave+ncover) +
			(hdata->amps_per_row-1)*quad_size;

		for (a=1; a <hdata->amps_per_row; a++) {
		    sh_quad[a] = sh_quad[a-1] - quad_size;
		}
		for (a=0; a<hdata->amps_per_row; a++) {
		    sh_row[a] = r+row;
		}

/*		Define lower quadrant addresses and row number.		*/

		sh_quad[hdata->amps_per_row] = sh_quad[0] +
			(nrover/2)*(ncsave+ncover);
		for (a=hdata->amps_per_row+1; a<n_amps; a++) {
		    sh_quad[a] = sh_quad[a-1] - quad_size;
		}
		for (a=hdata->amps_per_row; a<n_amps; a++) {
		    sh_row[a] = r+nrover-row-1;
		}

/*		Copy main part of the row overscan.			*/

		for(col=0; col <(c/hdata->amps_per_row); col++) {

		    for(a=0; a<hdata->amps_per_row;a++) {
			sh_col[a] = c-(a+1)*quad_size+col;
			sh_col[a+hdata->amps_per_row] = sh_col[a];
		    }
			
/*		    Process group of n_amps pixels.			*/
		    for(a=0; a<n_amps; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("9b",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}

			pixel_num++;
			count--;

		    }
		}

/*		Copy the column overscan part of the row overscan row.	*/

		if(hdata->over.c && saverow(sh_row[0])) {
		    sh_end = (unsigned short *)mem_image +
			(nrsave+row)*(ncsave+ncover);
		    for(a=0; a<hdata->amps_per_row;a++) {
			sh_quad[a] = sh_end - (a+1)* hdata->over.c;
		    }
		}
		if(hdata->over.c && saverow(sh_row[hdata->amps_per_row])) {
		    sh_end = sh_end +
			(nrover/2)*(ncsave+ncover);
		    for(a=0;a<hdata->amps_per_row;a++) {
			sh_quad[a+hdata->amps_per_row] =
				sh_end - (a+1)* hdata->over.c;
		    }
		}

		for(col=0; col<hdata->over.c; col++) {

		    for(a=0; a<hdata->amps_per_row;a++) {
			sh_col[a] = c-(a+1)*quad_size+col;
			sh_col[a+hdata->amps_per_row] = sh_col[a];
		    }
			
/*		    Process group of n_amps pixels.			*/
		    for(a=0; a<n_amps; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("9d",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}
			pixel_num++;
			count--;

		    }
		}
	    }
	    break;

	case 9:

/*	This is 4-channel readout of the Lincoln AO CCID56 or any other	*/
/*	device with multiple image segments but a single-amplifier	*/
/*	readout per sigment. The amplifier is assumed to be on the	*/
/*	left end of the serial register. This descramble option assumes	*/
/*	four segments from the same half of the array.			*/
/*	The pixels arrive in groups of four, one from each quadrant	*/
/*	as shown below. The first pixel in each quadrant is marked	*/
/*	with an X.							*/

/*	|X-------|X-------|X-------|X-------|				*/
/*	|        |        |        |        |				*/
/*	|    4   |    3   |    2   |    1   |				*/
/*	|        |        |        |        |				*/
/*	|--------|--------|--------|--------|				*/

/*	This case is written to handle up to 32 amplifiers as long	*/
/*	as hdata->amps_per_row is defined correctly.			*/

	    if (hdata->amps_per_row > MAX_AMPS_PER_ROW) {
		flogerror(
	"camread is configured for a maximum of %d amplifiers.",
		MAX_AMPS_PER_ROW);
		flogerror(
	"RECOMPILE WITH sh_quad[%d] increased to at least %d",
		MAX_AMPS_PER_ROW,hdata->amps_per_row);

/*		This probably won't work for descrambling the image	*/
/*		but it should prevent over-running the array.		*/

		hdata->amps_per_row = MAX_AMPS_PER_ROW;
	    }

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - ncover;

/*	    Address of last saved image pixel (not including overscan). */
	    sh_imlast = (unsigned short *)mem_image + \
				(ncsave+ncover)*nrsave - 1;

/*	    Here r is the total number of rows including overscan rows.	*/

	    for(row=0; row<r; row++) {
		int quad_size = ncsave/hdata->amps_per_row;
		int a;
		int cquad;

/*		sh_quad[0] is quadrant 0.				*/
/*		sh_quad[1] is quadrant 1.				*/
/*		sh_quad[2] is quadrant 2.				*/
/*		sh_quad[3] is quadrant 3.				*/

		sh_quad[0] = (unsigned short *)mem_image + \
		    row*(ncsave+ncover) + ncsave - quad_size;
		for(a=1; a<hdata->amps_per_row; a++)
	            sh_quad[a] = sh_quad[a-1] - quad_size;

/*		Copy the image and any row overscan.			*/
/*		Pixels come in groups and the number of such groups	*/
/*		is the same as the number of pixels per quadrant.	*/

		for(col=0; col <quad_size; col++) {

		    cquad = c - quad_size + col;

/*		    There is one pixel from each amplifier. So		*/
/*		    process them in groups of that size.		*/

		    for(a=0; a<hdata->amps_per_row; a++) {
		    
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("10a",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(row,cquad)) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a]++;
			}

			pixel_num++;
			count--;
			cquad -= quad_size;
		    }
		}
		if(hdata->over.c && saverow(row)) {
		    sh_quad[0] = (unsigned short *)mem_image + \
		    (row+1)*(ncsave+ncover) - hdata->over.c;
		    for(a=1; a<hdata->amps_per_row; a++)
	                sh_quad[a] = sh_quad[a-1] - hdata->over.c;
		}

/*		Copy the overscan portion of the image.			*/

		for(col=0; col <hdata->over.c; col++) {

		    cquad = c_overscan - hdata->over.c + col;

		    for(a=0; a<hdata->amps_per_row; a++) {
		    
		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("8",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(row,cquad)) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a]++;
			}

			pixel_num++;
			count--;
			cquad -= hdata->over.c;
		    }
		}

	    }

	    break;

/*	This case is for 8 outputs as illustrated in this diagram:	*/

/*	|X-------|-------X|X-------|-------X|				*/
/*	|        |        |        |        |				*/
/*	|    1   |    2   |    3   |    4   |				*/
/*	|        |        |        |        |				*/
/*	|--------|--------|--------|--------|				*/
/*	  ...............................				*/
/*	|--------|--------|--------|--------|				*/
/*	|        |        |        |        |				*/
/*	|    5   |    6   |    7   |    8   |				*/
/*	|        |        |        |        |				*/
/*	|X-------|-------X|X-------|-------X|				*/

	case 48:

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - ncover;

/*	    Change r to be the number of rows without		*/
/*	    overscan. (Unchanged if there is no overscan rows.)	*/
	    r = r - nrover;

/*	    Address of last saved image pixel (not including overscan). */
	    sh_imlast = (unsigned short *)mem_image + \
				(ncsave+ncover)*r - 1 - ncover;

/*	    Define how the column numbers advance for each quadrant.	*/
	    sh_step[0] = 1;
	    sh_step[1] = -1;
	    sh_step[2] = 1;
	    sh_step[3] = -1;
	    sh_step[4] = 1;
	    sh_step[5] = -1;
	    sh_step[6] = 1;
	    sh_step[7] = -1;

	    for(row=0; row<r/2; row++) {
		int a;

		sh_quad[0] = (unsigned short *)mem_image + row*(ncsave+ncover);
		sh_quad[1] = sh_quad[0]+(ncsave/2)-1;
		sh_quad[2] = sh_quad[1] + 1;
		sh_quad[3] = sh_quad[0] + ncsave - 1;
		sh_quad[7] = sh_imlast - (row*(ncsave+ncover));
		sh_quad[5] = sh_quad[7] - (ncsave/2);
		sh_quad[6] = sh_quad[5] + 1;
		sh_quad[4] = sh_quad[6] - (ncsave/2);

		sh_row[0] = sh_row[1] = sh_row[2] = sh_row[3] = row;
		sh_row[4] = sh_row[5] = sh_row[6] = sh_row[7] = r-row-1;

/*		Copy main part of the image.				*/
		for(col=0; col <(c/4); col++) {

		    sh_col[0] = sh_col[4] = col;
		    sh_col[1] = sh_col[5] = (c/2)-1-col;
		    sh_col[2] = sh_col[6] = sh_col[0]+(c/2);
		    sh_col[3] = sh_col[7] = c-1-col;

/*		    Process a group of eight pixels.			*/
		    for(a=0; a<8; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("48a",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }

			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a] ;
			}

			pixel_num++;
			count--;

		    }
		}
		if(hdata->over.c && saverow(sh_row[0])) {
			sh_quad[0] = (unsigned short *)mem_image +
				row*(ncsave+ncover) + c;
			sh_quad[1] = sh_quad[0] + hdata->over.c;
			sh_quad[2] = sh_quad[1] + hdata->over.c;
			sh_quad[3] = sh_quad[2] + hdata->over.c;
		}
		if(hdata->over.c && saverow(sh_row[4])) {
			sh_quad[4] = sh_imlast + 1 - row*(ncsave+ncover);
			sh_quad[5] = sh_quad[4] + hdata->over.c;
			sh_quad[6] = sh_quad[5] + hdata->over.c;
			sh_quad[7] = sh_quad[6] + hdata->over.c;
		}

/*		Copy the overscan portion of the image.			*/

		for(col=0; col<hdata->over.c; col++) {

		    sh_col[0] = c+col;
		    sh_col[1] = sh_col[0] + hdata->over.c;
		    sh_col[2] = sh_col[1] + hdata->over.c;
		    sh_col[3] = sh_col[2] + hdata->over.c;
		    sh_col[4] = sh_col[0];
		    sh_col[5] = sh_col[1];
		    sh_col[6] = sh_col[2];
		    sh_col[7] = sh_col[3];

/*		    Process a group of eight pixels.			*/
		    for(a=0; a<8; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("48b",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += 1;
			}
			pixel_num++;
			count--;

		    }
		}

	    }

/*	    Now we save the row overscan pixels, if there are any.	*/

/*	    Address of first saved row overscan pixel			*/
	    sh_imfirst = sh_imlast + ncover;

/*	    Address of last saved row overscan pixel			*/
	    sh_imlast = sh_imfirst + (ncsave+ncover)*nrover -ncover - 1;

	    for(row=0; row<hdata->over.r; row++) {
		int a;

		sh_quad[0] = sh_imfirst + row*(ncsave+ncover);
                sh_quad[1] = sh_quad[0]+(ncsave/2)-1;
                sh_quad[2] = sh_quad[1] + 1;
                sh_quad[3] = sh_quad[0] + ncsave - 1;
                sh_quad[7] = sh_imlast - row*(ncsave+ncover);
                sh_quad[5] = sh_quad[7] - (ncsave/2);
                sh_quad[6] = sh_quad[5] + 1;
                sh_quad[4] = sh_quad[6] - (ncsave/2);

                sh_row[0] = sh_row[1] = sh_row[2] = sh_row[3] = r + row;
                sh_row[4] = sh_row[5] = sh_row[6] = sh_row[7] =
			r + nrover - row - 1;


/*		Copy main part of the row overscan.			*/
		for(col=0; col <(c/4); col++) {

		    sh_col[0] = sh_col[4] = col;
		    sh_col[1] = sh_col[5] = (c/2)-1-col;
		    sh_col[2] = sh_col[6] = sh_col[0]+(c/2);
		    sh_col[3] = sh_col[7] = c-1-col;

/*		    Process a group of eight pixels.			*/
		    for (a=0; a<8; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("48c",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}

			pixel_num++;
			count--;

		    }
		}
		if(hdata->over.c && saverow(sh_row[0])) {
			sh_quad[0] = sh_imfirst +
			    row*(ncsave+ncover) + c;
                        sh_quad[1] = sh_quad[0] + hdata->over.c;
                        sh_quad[2] = sh_quad[1] + hdata->over.c;
                        sh_quad[3] = sh_quad[2] + hdata->over.c;
		}
		if(hdata->over.c && saverow(sh_row[2])) {
			sh_quad[4] = sh_imlast + 1 - row*(ncsave+ncover);
			sh_quad[5] = sh_quad[4] + hdata->over.c;
			sh_quad[6] = sh_quad[5] + hdata->over.c;
			sh_quad[7] = sh_quad[6] + hdata->over.c;
		}

/*		Copy the overscan portion of the image.			*/

		for(col=0; col<hdata->over.c; col++) {

                    sh_col[0] = c+col;
                    sh_col[1] = sh_col[0] + hdata->over.c;
                    sh_col[2] = sh_col[1] + hdata->over.c;
                    sh_col[3] = sh_col[2] + hdata->over.c;
                    sh_col[4] = sh_col[0];
                    sh_col[5] = sh_col[1];
                    sh_col[6] = sh_col[2];
                    sh_col[7] = sh_col[3];

/*		    Process a group of eight pixels.			*/
		    for(a=0; a<8; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("48d",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}
			pixel_num++;
			count--;

		    }
		}
	    }
	
	    break;
/*	This case is for 4 outputs as illustrated in this diagram:	*/

/*	|X-------|-------X|X-------|-------X|				*/
/*	|        |        |        |        |				*/
/*	|    1   |    2   |    3   |    4   |				*/
/*	|        |        |        |        |				*/
/*	|--------|--------|--------|--------|				*/

	case 64:

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - ncover;

/*	    Change r to be the number of rows without		*/
/*	    overscan. (Unchanged if there is no overscan rows.)	*/
	    r = r - nrover;

/*	    Address of last saved image pixel (not including overscan). */
	    sh_imlast = (unsigned short *)mem_image + \
				(ncsave+ncover)*r - 1 - ncover;

/*	    Define how the column numbers advance for each quadrant.	*/
	    sh_step[0] = 1;
	    sh_step[1] = -1;
	    sh_step[2] = 1;
	    sh_step[3] = -1;

	    for(row=0; row<r; row++) {
		int a;

		sh_quad[0] = (unsigned short *)mem_image + row*(ncsave+ncover);
		sh_quad[1] = sh_quad[0]+(ncsave/2)-1;
		sh_quad[2] = sh_quad[1] + 1;
		sh_quad[3] = sh_quad[0] + ncsave - 1;

		sh_row[0] = sh_row[1] = sh_row[2] = sh_row[3] = row;

/*		Copy main part of the image.				*/
		for(col=0; col <(c/4); col++) {

		    sh_col[0] = col;
		    sh_col[1] = (c/2)-1-col;
		    sh_col[2] = sh_col[0]+(c/2);
		    sh_col[3] = c-1-col;

/*		    Process a group of eight pixels.			*/
		    for(a=0; a<4; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("60a",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }

			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a] ;
			}

			pixel_num++;
			count--;

		    }
		}
		if(hdata->over.c && saverow(sh_row[0])) {
			sh_quad[0] = (unsigned short *)mem_image +
				row*(ncsave+ncover) + c;
			sh_quad[1] = sh_quad[0] + hdata->over.c;
			sh_quad[2] = sh_quad[1] + hdata->over.c;
			sh_quad[3] = sh_quad[2] + hdata->over.c;
		}
		

/*		Copy the overscan portion of the image.			*/

		for(col=0; col<hdata->over.c; col++) {

		    sh_col[0] = c+col;
		    sh_col[1] = sh_col[0] + hdata->over.c;
		    sh_col[2] = sh_col[1] + hdata->over.c;
		    sh_col[3] = sh_col[2] + hdata->over.c;

/*		    Process a group of eight pixels.			*/
		    for(a=0; a<4; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("48b",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += 1;
			}
			pixel_num++;
			count--;

		    }
		}

	    }

/*	    Now we save the row overscan pixels, if there are any.	*/

/*	    Address of first saved row overscan pixel			*/
	    sh_imfirst = sh_imlast + ncover;

/*	    Address of last saved row overscan pixel			*/
	    sh_imlast = sh_imfirst + (ncsave+ncover)*nrover -ncover - 1;

	    for(row=0; row<hdata->over.r; row++) {
		int a;

		sh_quad[0] = sh_imfirst + row*(ncsave+ncover);
                sh_quad[1] = sh_quad[0]+(ncsave/2)-1;
                sh_quad[2] = sh_quad[1] + 1;
                sh_quad[3] = sh_quad[0] + ncsave - 1;
                            sh_row[0] = sh_row[1] = sh_row[2] = sh_row[3] = r + row;

/*		Copy main part of the row overscan.			*/
		for(col=0; col <(c/4); col++) {

		    sh_col[0] = col;
		    sh_col[1] = (c/2)-1-col;
		    sh_col[2] = sh_col[0]+(c/2);
		    sh_col[3] = c-1-col;

/*		    Process a group of eight pixels.			*/
		    for (a=0; a<4; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("48c",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}

			pixel_num++;
			count--;

		    }
		}
		if(hdata->over.c && saverow(sh_row[0])) {
			sh_quad[0] = sh_imfirst +
			    row*(ncsave+ncover) + c;
                        sh_quad[1] = sh_quad[0] + hdata->over.c;
                        sh_quad[2] = sh_quad[1] + hdata->over.c;
                        sh_quad[3] = sh_quad[2] + hdata->over.c;
		}

/*		Copy the overscan portion of the image.			*/

		for(col=0; col<hdata->over.c; col++) {

                    sh_col[0] = c+col;
                    sh_col[1] = sh_col[0] + hdata->over.c;
                    sh_col[2] = sh_col[1] + hdata->over.c;
                    sh_col[3] = sh_col[2] + hdata->over.c;

/*		    Process a group of four pixels.			*/
		    for(a=0; a<4; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("60d",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}
			pixel_num++;
			count--;

		    }
		}
	    }
	
	    break;
/*	This case is for 2 outputs as illustrated in this diagram:	*/

/*	|X----------------|X----------------|				*/
/*	|                 |                 |				*/
/*	|        1        |        2        |				*/
/*	|                 |                 |				*/
/*	|-----------------|-----------------|				*/

	case 61:

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - ncover;

/*	    Change r to be the number of rows without		*/
/*	    overscan. (Unchanged if there is no overscan rows.)	*/
	    r = r - nrover;

/*	    Address of last saved image pixel (not including overscan). */
	    sh_imlast = (unsigned short *)mem_image + \
				(ncsave+ncover)*r - 1 - ncover;

/*	    Define how the column numbers advance for each quadrant.	*/
	    sh_step[0] = 1;
	    sh_step[1] =  1;

	    for(row=0; row<r; row++) {
		int a;

		sh_quad[0] = (unsigned short *)mem_image + row*(ncsave+ncover);
		sh_quad[1] = sh_quad[0]+(ncsave/2);

		sh_row[0] = sh_row[1] = row;

/*		Copy main part of the image.				*/
		for(col=0; col <(c/2); col++) {

		    sh_col[0] = col;
		    sh_col[1] = (c/2)+col;

/*		    Process a group of two pixels.			*/
		    for(a=0; a<2; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("61a",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }

			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a] ;
			}

			pixel_num++;
			count--;

		    }
		}
		if(hdata->over.c && saverow(sh_row[0])) {
			sh_quad[0] = (unsigned short *)mem_image +
				row*(ncsave+ncover) + c;
			sh_quad[1] = sh_quad[0] + hdata->over.c;
		}
		

/*		Copy the overscan portion of the image.			*/

		for(col=0; col<hdata->over.c; col++) {

		    sh_col[0] = c+col;
		    sh_col[1] = sh_col[0] + hdata->over.c;

/*		    Process a group of two pixels.			*/
		    for(a=0; a<2; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("48b",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += 1;
			}
			pixel_num++;
			count--;

		    }
		}

	    }

/*	    Now we save the row overscan pixels, if there are any.	*/

/*	    Address of first saved row overscan pixel			*/
	    sh_imfirst = sh_imlast + ncover;

/*	    Address of last saved row overscan pixel			*/
	    sh_imlast = sh_imfirst + (ncsave+ncover)*nrover -ncover - 1;

	    for(row=0; row<hdata->over.r; row++) {
		int a;

		sh_quad[0] = sh_imfirst + row*(ncsave+ncover);
                sh_quad[1] = sh_quad[0]+(ncsave/2)-1;
                            sh_row[0] = sh_row[1] = r + row;

/*		Copy main part of the row overscan.			*/
		for(col=0; col <(c/2); col++) {

		    sh_col[0] = col;
		    sh_col[1] = (c/2)+col;

/*		    Process a group of two pixels.			*/
		    for (a=0; a<2; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("61c",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}

			pixel_num++;
			count--;

		    }
		}
		if(hdata->over.c && saverow(sh_row[0])) {
			sh_quad[0] = sh_imfirst +
			    row*(ncsave+ncover) + c;
                        sh_quad[1] = sh_quad[0] + hdata->over.c;
		}

/*		Copy the overscan portion of the image.			*/

		for(col=0; col<hdata->over.c; col++) {

                    sh_col[0] = c+col;
                    sh_col[1] = sh_col[0] + hdata->over.c;

/*		    Process a group of four pixels.			*/
		    for(a=0; a<2; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("61d",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}
			pixel_num++;
			count--;

		    }
		}
	    }
	break;
/*	This case is for 2 outputs as illustrated in this diagram:	*/

/*	|----------------X|----------------X|				*/
/*	|                 |                 |				*/
/*	|        2        |        1        |				*/
/*	|                 |                 |				*/
/*	|-----------------|-----------------|				*/

	case 62:

/*	    pixel_num is the index into the data array. Skip	*/
/*	    over the header.					*/

	    pixel_num = headersize;

/*	    Column number of last overscan (if there is any).	*/
	    c_overscan = c - 1;

/*	    Change c to be the number of columns without	*/
/*	    overscan. (Unchanged if there is no overscan.)	*/
	    c = c - ncover;

/*	    Change r to be the number of rows without		*/
/*	    overscan. (Unchanged if there is no overscan rows.)	*/
	    r = r - nrover;

/*	    Address of last saved image pixel (not including overscan). */
	    sh_imlast = (unsigned short *)mem_image + \
				(ncsave+ncover)*r - 1 - ncover;

/*	    Define how the column numbers advance for each quadrant.	*/
	    sh_step[0] = -1;
	    sh_step[1] = -1;

	    for(row=0; row<r; row++) {
		int a;

		sh_quad[0] = (unsigned short *)mem_image + row*(ncsave+ncover)+ncsave;
		sh_quad[1] = sh_quad[0]-(ncsave/2);

		sh_row[0] = sh_row[1] = row;

/*		Copy main part of the image.				*/
		for(col=0; col <(c/2); col++) {

		    sh_col[0] = c-col;
		    sh_col[1] = (c/2)-col;

/*		    Process a group of two pixels.			*/
		    for(a=0; a<2; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("61a",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }

			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a] ;
			}

			pixel_num++;
			count--;

		    }
		}
		if(hdata->over.c && saverow(sh_row[0])) {
			sh_quad[0] = (unsigned short *)mem_image +
				row*(ncsave+ncover) + c;
			sh_quad[1] = sh_quad[0] + hdata->over.c;
		}
		

/*		Copy the overscan portion of the image.			*/

		for(col=0; col<hdata->over.c; col++) {

		    sh_col[0] = c+col;
		    sh_col[1] = sh_col[0] + hdata->over.c;

/*		    Process a group of two pixels.			*/
		    for(a=0; a<2; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("48b",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += 1;
			}
			pixel_num++;
			count--;

		    }
		}

	    }

/*	    Now we save the row overscan pixels, if there are any.	*/

/*	    Address of first saved row overscan pixel			*/
	    sh_imfirst = sh_imlast + ncover;

/*	    Address of last saved row overscan pixel			*/
	    sh_imlast = sh_imfirst + (ncsave+ncover)*nrover -ncover - 1;

	    for(row=0; row<hdata->over.r; row++) {
		int a;

		sh_quad[0] = sh_imfirst + row*(ncsave+ncover);
                sh_quad[1] = sh_quad[0]+(ncsave/2)-1;
                            sh_row[0] = sh_row[1] = r + row;

/*		Copy main part of the row overscan.			*/
		for(col=0; col <(c/2); col++) {

		    sh_col[0] = c-col;
		    sh_col[1] = (c/2)-col;

/*		    Process a group of two pixels.			*/
		    for (a=0; a<2; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("61c",pixel_num);
		            if (count < 0) return 0;	/* Error!*/
		        }
			checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}

			pixel_num++;
			count--;

		    }
		}
		if(hdata->over.c && saverow(sh_row[0])) {
			sh_quad[0] = sh_imfirst +
			    row*(ncsave+ncover) + c;
                        sh_quad[1] = sh_quad[0] + hdata->over.c;
		}

/*		Copy the overscan portion of the image.			*/

		for(col=0; col<hdata->over.c; col++) {

                    sh_col[0] = c+col;
                    sh_col[1] = sh_col[0] + hdata->over.c;

/*		    Process a group of four pixels.			*/
		    for(a=0; a<2; a++) {

		        if (count <= 0) {
		            if(report_progress)
	    		        send_progress_report(xmit_pixels,pixel_num);
		            count = wait_for_data("61d",pixel_num);
		            if (count < 0) return 0;	/* Error!	*/
		        }
		    	checksum += image[pixel_num];
			if(saveit(sh_row[a],sh_col[a])) {
			    *sh_quad[a] = image[pixel_num];
			    sh_quad[a] += sh_step[a];
			}
			pixel_num++;
			count--;

		    }
		}
	    }
	break;

	default:
		flogerror("Unknown image type %d", hdata->type);
		return 0;
	}

	if(report_progress)
	    send_progress_report(xmit_pixels,xmit_pixels);

	if(hdata->checksum_flag) {
	
/*
	  flogerror("*** To locate checksum using pixelnum %d",
		pixel_num);
	  flogerror("*** Found camera checksum %u",
		(unsigned int)image[pixel_num]);
	  flogerror("*** Value before checksum %u",
		(unsigned int)image[pixel_num-1]);
	  flogerror("*** Value after  checksum %u",
		(unsigned int)image[pixel_num+1]);
*/
	    cam_checksum = image[pixel_num];
	    hdata->cam_checksum = cam_checksum;
	    hdata->im_checksum  = cam_checksum;
	    if(cam_checksum != checksum) {
		flogerror(
		  "Camera checksum %u DOES NOT MATCH our image checksum %u",
		    (unsigned int)cam_checksum,(unsigned int)checksum);
		hdata->im_checksum = checksum;
		if(record_checksum_error_image)
		    record_image((unsigned short *)mem_image,hdata,
			"/tmp/cksum_error.fits");
	    } else {
		flogerror(
		    "Camera checksum matches image checksum %u",
		    (unsigned int)cam_checksum);
	    }
//	    recheck_sums((unsigned short *)mem_image,&image[headersize],pixels);
	}

/*	Use the END card at the end of the header buffer, not the one	*/
/*	added by fits_cards.						*/
	strncpy((char *)header+(fits_count-1)*80,"        ",8);

	if(hdata->checksum_flag) {
	    int ok;
	    if(hdata->im_checksum == hdata->cam_checksum) {
		ok = 1;
		update_fits_card(hdata->id,"CKSUMOK",'T',&ok,
			"CHECKSUMS MATCH");
	    } else {
		ok = 0;
		update_fits_card(hdata->id,"CKSUMOK",'T',&ok,
			"CHECKSUMS DO NOT MATCH");
	    }
	    ok = (int)hdata->cam_checksum;
	    update_fits_card(hdata->id,"CAMCKSUM",'d',&ok,
		"CAMERA-COMPUTED CHECKSUM");
	    update_fits_card(hdata->id,"SFTCKSUM",'d',&ok,
		"SOFTWARE-COMPUTED CHECKSUM");
	}
		
	return 1;
}


/*	We use an array of integers to indicate which rows to save.	*/
/*	The index into the array is the row's number, starting at 0.	*/
short *rows_to_save(struct header_data *hdata)
{
	static short *rsave;
	static int nrsave = 0;
	int r,rn;
	int nrover,nr;
	int nb;			/* Number of bytes in save flag array*/

/*      Make sure we can handle the number of rows in the image.     */
        nr = hdata->image.r;
	nb = nr*sizeof(short);
	if (nrsave < nr) {
	    if(nrsave > 0)
		free(rsave);
	    rsave = (short *)malloc(nb);
	    if(rsave == NULL) {
		flogerror(
        "rows_to_save: unable to allocate %d bytes for save flags: %s",
		nb,strerror(errno));
		flogerror("rows_to_save: exiting");
		exit(2);
	    }
	    nrsave = nr;
	}

/*	Clear the array						*/
	memset(rsave,0,nb);

/*	Mark all rows in the data window			*/
	rn = hdata->window.rorg;
	nr += rn;
/*
	flogerror("first row to save is %d last is %d",
	    rn,rn+hdata->window.r-1);
*/
	for(r=0; (r < hdata->window.r) && (rn < nr); r++,rn++)
	    rsave[rn] = 1;

/*	Mark all rows of overscan				*/
	nrover = hdata->over.r*hdata->amps_per_column;

/*	If there is no overscan we are done.			*/
	if(nrover == 0) return rsave;

/*	The overscan rows are included in the overall image size	*/
/*	and are the last rows in the image. So the row number of the	*/
/*	first overscan row is determined by subtracting the number of	*/
/*	overscan rows from the overall image size.			*/

	rn = hdata->image.r-nrover;
/*
flogerror("making rows %d to %d as overscan",rn,rn+hdata->over.r-1);
*/
	for(r=0; (r < nrover) && (rn < nr); r++,rn++)
	    rsave[rn] = 1;

	return rsave;
}

/*	We use an array of integers to indicate which columns to save.	*/
/*	The index into the array is the column number, starting at 0.	*/
short *columns_to_save(struct header_data *hdata)
{
	static short *csave;
	static int ncsave = 0;
	int c,cn,ncover,nc;
	int nb;			/* Number of bytes in save flag array*/

/*	Make sure we can handle the number of columns in the image.	*/
	nc = hdata->image.c;
	nb = nc*sizeof(short);
	if (ncsave < nc) {
	    if(ncsave > 0)
		free(csave);
	    csave = (short *)malloc(nb);
	    if(csave == NULL) {
		flogerror(
        "columns_to_save: unable to allocate %d bytes for save flags: %s",
		nb,strerror(errno));
		flogerror("columns_to_save: exiting");
		exit(2);
	    }
	    ncsave = nc;
	}

	cn = hdata->window.corg;
	memset(csave,0,nb);
/*
	flogerror("first column to save is %d last is %d",
	    cn,cn+hdata->window.c-1);
*/

/*	Mark all columns in the data window			*/
	for(c=0; (c < hdata->window.c) && (cn < nc); c++,cn++)
	    csave[cn] = 1;

/*	Mark all columns of overscan				*/
	ncover = hdata->over.c*hdata->amps_per_row;

/*	If there is no overscan we are done.			*/
	if(ncover == 0) return csave;

/*	The overscan columns are included in the overall image size	*/
/*	and are the last columns in the image. So the column number of	*/
/*	the first overscan column is determined by subtracting the	*/
/*	number of overscan columns from the overall image size.		*/

	cn = hdata->image.c-ncover;
flogerror("marking columns %d to %d as overscan",cn,cn+ncover-1);
/*
*/
	for(c=0; (c < ncover) && (cn < nc); c++,cn++)
	    csave[cn] = 1;
	return csave;
}


/*	Wait for more data to become available for processing.		*/
/*	We do this with the DPP_READ_PROG ioctl call, which returns	*/
/*	the number of bytes processed since the last DPP_READ_PROG	*/
/*	ioctl.								*/
int wait_for_data(char *ckpt,int pn) {
	int count = 0;
	int xferred;
	char *err;
	static char status[256];

	while(count == 0) {
	    if(ioctl(im_fd,DPP_READ_PROG,&xferred) < 0) {

/*		It is a fatal error if this returns an error.		*/
		flogerror("%s DPP_READ_PROG returned error %d: %s",
			__FILE__,errno,strerror(errno));
		flogerror("%s wait_for_data aborting at checkpoint %s!",
		    __FILE__,ckpt);
		err = loglasterror();
		sprintf(status,
		   "ERROR Error >>%s<< while reading image data." "#n"
		   "In camread procedure wait_for_data at checkpoint %s."
		    "#n"
	    "This happened at pixel %d of the image (including header)"
		    "#n"
		    "This probably means the image did not fully read out.",
			err,ckpt,pn);
		save_error_status(status);
		return -1;
	    }
	    if(xferred > 0)
	        count = xferred/2;	/* Want number of 16-bit ints.	*/
	    else if(xferred < 0) {
		flogerror("Got negative count from READ_PROG ioctl: %d"
		    " at checkpoint %s",
			xferred,ckpt);
		err = loglasterror();
		sprintf(status,"ERROR Error >>%s<< while reading image data."
		    "#n"
		    "In camread procedure wait_for_data at checkpoint %s."
		    "#n"
		"This happened at pixel %d of the image (including header)"
		    "#n"
		    "This probably means the image did not fully read out.",
			err,ckpt,pn);
		save_error_status(status);
		return -1;
	    } else
		count = xferred;
/*
	    if(count == 0) {
		flogerror("wait_for_data got data count=0 at point %s",
			ckpt);
	    }
*/
	}
/*
	flogerror("wait_for_data returns data count=%d to checkpoint %s",
		count,ckpt);
*/
	return count;
}


/*	The four corners of the array:

	(0,0) _______________ (C,0)
             |               |
             |               |
             |               |
             |               |
             |               |
             |               |
             |               |
	(0,R)-----------------(C,R)
*/

/*	Here are the currectly defined codes for descrambling images	*/
/*
 *	0	One channel from corner (0,0)
 *	1	One channel from corner (C,0)
 *	2	One channel from corner (0,R)
 *	3	One channel from corner (C,R)
 *	4	Two channel from corners (0,0) and (C,0)
 *	5	Two channel from corners (C,0) and (0,0)
 *	6	Two channel from corners (C,R) and (0,R)
 *	7	Two channel from corners (0,R) and (C,R)
 *	9	Special CCID56 mode
 *	10	Second special CCID56 mode
 *	11	Two channels from corners (0,0) and (C,R)
 *	12	Two channels from corners (0,R) and (C,0)
 *	13	Two channels from corners (0,0) and (0,R)
 *	14	Two channels from corners (C,0) and (C,R)
*/ 

/*
	If we want to flip an image in columns make this change in the code:
	0   ->  1
	1   ->  0
	2   ->  3
	3   ->  2
	4   ->  5
	5   ->  4
	6   ->  7
	7   ->  6

	If we want to flip an image in rows make this change in the code:
	0   ->  2
	1   ->  3
	2   ->  0
	3   ->  1
	4   ->  6
	5   ->  7
	6   ->  4
	7   ->  5
*/

/*	Return an updated readout_geometry depending on how	*/
/*	the image is supposed to be flipped.			*/
static int flip_options(struct image_flips *image_flips)
{
    int readout_geometry;

/*  row_exchange and column_exchange give the mappings from	*/
/*  unflipped geometry code to flipped geometry code.		*/

    static int row_exchange[] =    {2,3,0,1,6,7,4,5};
    static int column_exchange[] = {1,0,3,2,5,4,7,6};

/*  If readout_geometry corresponds to one of the flippable	*/
/*  geometries then check for flipping.				*/

    readout_geometry = image_flips->geometry;
    if((readout_geometry >= 0) && (readout_geometry < 8)) {

/*	Flip rows if requested.					*/
        if(image_flips->row_flip)
		readout_geometry = row_exchange[readout_geometry];

/*	Flip columns if requested.				*/
        if(image_flips->col_flip)
		readout_geometry = column_exchange[readout_geometry];
    }

    return(readout_geometry);
}

/*	Send a message to the display program for the new image		*/
/*	available in shared memory. We send the shared memory id,	*/
/*	the offset of the image from the start of the shared memory,	*/
/*	the offset of the header from the start of the shared memory,	*/
/*	and the number of rows and columns in the image.		*/
void send_display_message(void)
{
	if(traffic_ok == 0) {
		flogerror("NO DISPLAY MESSAGE BECAUSE NO TRAFFIC!");
		return;
	}
/*
	flogerror("sending NEW_SHM_IMAGE shid %d offset %d header %d",
			shid,process_image-process_buffer,process_header-process_buffer);
*/
	mstart(BROADCAST,our_addr,NEW_SHM_IMAGE);
	mputl(process_shid);			/* Shared memory ID	*/
	mputl(process_image-process_buffer);	/* Image offset		*/
	mputl(process_header-process_buffer);	/* FITS header offset	*/
	msend(traffic_fd);			/* Send it		*/
}

/*	Send a message to the guider saying that a new image is		*/
/*	available in shared memory. We send the shared memory id,	*/
/*	the offset of the image from the start of the shared memory,	*/
/*	the offset of the header from the start of the shared memory,	*/
/*	and the number of rows and columns in the image.		*/
int send_guider_message(int sock,int r, int c)
{
        int rb;
	char guide_msg[100];

	if(sock < 0) return(-1);

        sprintf(guide_msg,"I %d 0x%x 0x%x %d %d",
		process_shid,process_image-process_buffer,
		process_header-process_buffer,r,c);

/*	This might fail if the guider program isn't pulling messages	*/
/*	out of the socket fast enough.					*/
        rb = write(sock,guide_msg,strlen(guide_msg)+1);

	return(rb);
}

/*	Record the image starting at address imaddr and header	*/
/*	data hdata.						*/
/*	Log an error message and return 0 if we can't record	*/
/*	the image for any reason.				*/
/*	Otherwise return 1 on success.				*/

int record_image(unsigned short *imaddr,struct header_data *hdata,
    char *passed_filename)
{
	static char file[256];	/* Constructed file name	*/
	char *filename;
	char *suffix;
	char *dir;
        FILE *fp;
	int nrec;
	int bytes,pixels;
	static char card[200*80];
	int ncards = 0;
	int nc;
	struct record *record;
	static short swabtest = 1;
	size_t nw;
	int passed_file = 0;
	int ret,m;
	char status[512];

/*	Build the file name for saving the image.		*/
/*	We do this only if the filename is not passed to us.	*/

/*	First get the directory or use a default.		*/
	record = RECORD_ADDR(hdata->id);

	if(passed_filename == NULL) {
	    logrecord("record_image",hdata->id,record);

	    dir = record->dir;
	    if (*dir == '\0')
		dir = "/data";

/*	Then get the base filename or use a default.		*/
	    filename = record->filename;
	    if (*filename == '\0')
		filename = "image";
	    else if (*filename == '/')
		filename++;		/* Skip leading /	*/

/*	Next get the filename suffix or use a default.		*/
	    suffix = record->suffix;
	    if (*suffix == '\0')
		suffix = "ccd";

/*	Combine into a final file name.				*/

	    sprintf(file,"%s/%s%d.%s",dir,filename,record->record_id,
		suffix);
	} else {
	    strncpy(file,passed_filename,sizeof(file));
	    file[sizeof(file)-1] = '\0';
	    passed_file++;
	}

/*	Save a pointer to this file name.			*/
	record->recorded_file = file;

/*	Try to open the file for writing.			*/
	fp = fopen(file,"w");
	if(fp == NULL) {
		char *emsg = strerror(errno);
		flogerror("can't open file %s (%s)",file,emsg);
		sprintf(status,"ERROR Can't open file %s to record image#n"
		    "error message is: %s", file, emsg);
		save_error_status(status);
		return 0;
	} else {
		flogerror("opened file %s for image with id %d",file,
			hdata->id);
	}

/*	Set file mode to make sure sticky bit is off		*/
	m = S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH;
	ret = chmod(file,m);
	if (ret == -1) {
	    flogerror("Unable to set mode 0%o on file %s: %s", m, file,
		 strerror(errno));
	}


/*	Generate our fits cards.				*/
//	nc = fits_cards(hdata,card,NULL);
        nc = fits_cards(hdata,card,FLIP_ADDR);

/*	If we constructed our file name then increment the file	*/
/*	id number for this image id.				*/
	if(passed_file == 0)
	    record->record_id++;

/*	Copy FITS cards to output					*/
	if(nc > 0) {
	    if((nw=fwrite(card,1,80*nc,fp)) != 80*nc) {
		char *emsg = strerror(errno);
		flogerror(
	"Able to write only %d bytes of %d-byte header then got error: %s",
		nw,80*nc,emsg);
		fclose(fp);
		sprintf(status,
		"ERROR %s: Able to write only %d bytes of %d-byte header.#n"
		"Then got error: %s",__FILE__,nw,80*nc,emsg);
		save_error_status(status);
		return 0;
	    }
	    ncards += nc;
	    ncards %= 36;
	}

/*	Pad out the last header record if needed.			*/
	if(ncards > 0) {
            for(; ncards<36; ncards++)
                fprintf(fp,"%80s"," ");
	}

        pixels = hdata->image.r*hdata->image.c;
	bytes = pixels*2;

/*	We must swap bytes because the image data are in machine (PC)	*/
/*	byte order and this is not FITS format order.			*/

/*	Allocate an image swapping area if needed.			*/
	if(imaddr_swapped == NULL) {
	    if(needswab((char *)&swabtest)) {
	        imaddr_swapped = malloc(shm_nbytes);
	        if(imaddr_swapped == NULL) {
		    flogerror(
		      "CAN'T ALLOCATE MEMORY (%d bytes) FOR SWAPPED IMAGE"
		      "\n    Reason: %s",shm_nbytes,strerror(errno));
		    imaddr_swapped = (char *)mem_image;
	        }
	    } else {
	        imaddr_swapped = (char *)mem_image;
	    }
	    bytes_allocated = shm_nbytes;
	}

/*	Swap the image bytes if needed on this machine.			*/
/*	Either way, we flip the sign bit to transform unsigned 16 bits	*/
/*	into signed 16-bits. The original data will be recovered when	*/
/*	the BSCALE and BZERO parameters from the FITS header are applied*/
/*	to the FITS file pixel data.					*/

	if(needswab((char *)&swabtest)) {
	    zswab(imaddr,imaddr_swapped,pixels);
	} else {
	    zflip(imaddr,pixels);
	}

/*	Write the image.						*/
        if((nw=fwrite(imaddr_swapped,2,pixels,fp)) != pixels) {
	    char *emsg = strerror(errno);
	    flogerror(
	"Able to write only %d pixels of %d pixel image then got error: %s",
	    nw,pixels,emsg);
	    fclose(fp);
	    sprintf(status,
	        "ERROR %s: Able to write only %d pixels of %d pixel image.#n"
		"Then got error: %s",__FILE__,nw,80*nc,emsg);
	    save_error_status(status);
	    return 0;
	}
        nrec = bytes/2880;
        nrec = nrec*2880;

/*	Fill out the image to a whole number of FITS records.		*/
        if(bytes > nrec)
            bytes = nrec + 2880 - bytes;
	else
	    bytes = 0;
	if (bytes > 0) {
            if((nw=fwrite(imaddr_swapped,2,bytes/2,fp)) != bytes/2) {
	        char *emsg = strerror(errno);
	        flogerror(
		    "Able to write %d bytes of last FITS block"
		    " but wanted to write %d bytes.",nw,bytes/2);
	        flogerror("    Then got error: %s",emsg);
	        fclose(fp);
	        sprintf(status,
	          "ERROR %s: Able to write %d bytes of last FITS block.#n"
		  "but wanted to write %d bytes.#n"
		  "Then got error: %s",__FILE__,nw,bytes/2,emsg);
	        save_error_status(status);
	        return 0;
	    }
	}

/*	Close the image file.						*/
	fclose(fp);
	return 1;
}

/*	We swap bytes for the FITS standard and compliment the	*/
/*	original sign bit to transform the value from the range	*/
/*	0 to 65535 into the range -32768 to +32767.		*/
static void zswab(const void *imaddr,void *imaddr_swapped,int nshorts)
{
	short v;

/*	First swap all of the bytes.				*/
	swab(imaddr,imaddr_swapped,nshorts*2);

/*	Then compliment the original sign bit, which is now	*/
/*	the high bit of the low-order byte.			*/

	while(nshorts > 0) {
	    v = *(short *)imaddr_swapped;    
	    v ^= 128;			/* Flip that bit.	*/
	    *(short *)imaddr_swapped = v;
	    imaddr_swapped += 2;
	    nshorts--;
	}
}

/*	We compliment the sign bit to transform the value from	*/
/*	the range 0 to 65535 into the range -32768 to +32767.	*/
static void zflip(void *imaddr,int nshorts)
{
	short v;

/*	Compliment the original sign bit.			*/

	while(nshorts > 0) {
	    v = *(short *)imaddr;    
	    v ^= (1<<15);			/* Flip that bit.	*/
	    *(short *)imaddr = v;
	    imaddr += 2;
	    nshorts--;
	}
}

char *datetime(struct timeval *tv,char *timestr)
{
	time_t t;
	struct tm *tm;

	t = tv->tv_sec;
	tm = gmtime(&t);
	sprintf(timestr,"'%4.4d-%2.2d-%2.2dT%2.2d:%2.2d:%2.2d.%2.2d'",
		tm->tm_year+1900,tm->tm_mon+1,tm->tm_mday,
		tm->tm_hour,tm->tm_min,tm->tm_sec,
		((int)tv->tv_usec*100)/1000000);
	return(timestr);
}

/*	Retrieve the pixel binning from the 'other' FITS cards	*/
static void RetrieveBinning(int id,int *Cdelt,int *Rdelt)
{
	struct record *record;
	int n_other;
	char *card;
	int cnum;

/*	Note that this is NOT the same as CDELT1 or CDELT2	*/
/*	because of the possibility of image flipping.		*/

	flogerror("%s retrieving binning for image id %d",
		__FUNCTION__,id);
	record = RECORD_ADDR(id);
	n_other = record->fits_other;
	card = FITS_OTHER(id);

/*	Look up column binning.					*/
	cnum = find_fits_card(card,n_other,"CBIN",&card);
	if(cnum < 0) {
	    *Cdelt = 1;		/* Not found in 'other' cards.	*/
	} else {
	    sscanf(card+10,"%d",Cdelt);
	}

/*	Look up row binning.					*/
	card = FITS_OTHER(id);
	cnum = find_fits_card(card,n_other,"RBIN",&card);
	if(cnum < 0) {
	    *Rdelt = 1;		/* Not found in 'other' cards.	*/
	} else {
	    sscanf(card+10,"%d",Rdelt);
	}
}
	
/*	Construct a list of basic FITS cards.				*/
/*	Place the cards in the buffer cp.				*/
/*	Return the number of 80-column cards generated.			*/
/*	It is assumed that cp is big enough to hold everything we	*/
/*	generate.							*/
int fits_cards(struct header_data *hdata,char *cp,
    struct image_flips *image_flips)
{
	int id;
	int nc;
	struct record *record;
	int ncards = 0;
	int Corg,Rorg,Rdelt,Cdelt;
	char timestr[32];
	struct timeval *tv;
	char verstr[sizeof(version)+3];
	int obsnum;

	id = hdata->id % 128;
//	flogerror("building FITS cards for image id %d",id);

	Cdelt = hdata->col_bin;
	Rdelt = hdata->row_bin;

	Corg = hdata->final.corg;
	Rorg = hdata->final.rorg;

/*      If readout_geometry corresponds to one of the flippable	*/
/*      geometries then check for flipping.			*/

	if(image_flips != NULL) {
          int readout_geometry;
	  int nrover = hdata->over.r*hdata->amps_per_column;
	  int ncover = hdata->over.c*hdata->amps_per_row;

/*
	flogerror("r: flip %d r %d r1 %d r2 %d c: flip %d c %d c1 %d c2 %d",
		image_flips->row_flip,
                image_flips->row,
                image_flips->row1,
                image_flips->row2,
                image_flips->col_flip,
                image_flips->column,
                image_flips->column1,
                image_flips->column2);
*/

          readout_geometry = image_flips->geometry;
          if((readout_geometry >= 0) && (readout_geometry < 8)) {

/*	    Flip rows if requested.				*/
            //if(image_flips->row_flip && (image_flips->row3==0)) {
            if(image_flips->row_flip) {
		Rdelt = -Rdelt;
		Rorg = hdata->row_bin*(hdata->final.r-nrover-1) + Rorg;
	    }

/*	    Flip columns if requested.				*/
            //if(image_flips->col_flip && (image_flips->column3 == 0)) {
            if(image_flips->col_flip ) {
		Cdelt = -Cdelt;
		Corg = hdata->col_bin*(hdata->final.c-ncover-1) + Corg;
	    }
	  }
        }
/*
	flogerror("Rorg %d Rdelt %d Corg %d Cdelt %d",
		Rorg,Rdelt,Corg,Cdelt);
*/

/*	Record the image recording id as the OBSNUM in the header.	*/
	record = RECORD_ADDR(id);
	obsnum = record->record_id;

        sprintf(cp,"%-80.80s",
        "SIMPLE  =                    T / NORMAL FITS IMAGE");
	ncards++;
	cp += 80;

        sprintf(cp,"%-80.80s",
        "BITPIX  =                   16 / DATA PRECISION");
	ncards++;
	cp += 80;

        sprintf(cp,"%-80.80s",
        "NAXIS   =                    2 / NUMBER OF IMAGE DIMENSIONS");
	ncards++;
	cp += 80;

        sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "NAXIS1",hdata->final.c,"NUMBER OF COLUMNS");
	ncards++;
	cp += 80;

        sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "NAXIS2",hdata->final.r,"NUMBER OF ROWS");
	ncards++;
	cp += 80;

	sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "CRVAL1",Corg,"COLUMN ORIGIN");
        ncards++;
        cp += 80;

	sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "CRVAL2",Rorg,"ROW ORIGIN");
        ncards++;
        cp += 80;

	sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "CDELT1",Cdelt,"COLUMN CHANGE PER PIXEL");
        ncards++;
        cp += 80;

	sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "CDELT2",Rdelt,"ROW CHANGE PER PIXEL");
        ncards++;
        cp += 80;

        sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "OBSNUM",obsnum,"OBSERVATION NUMBER");
	ncards++;
	cp += 80;

        sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "IDNUM",hdata->id,"IMAGE ID");
	ncards++;
	cp += 80;

        sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "IMTYPE",hdata->type,"IMAGE READOUT GEOMETRY");
	ncards++;
	cp += 80;

        sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "AMPSROW",hdata->amps_per_row,"AMPLIFIERS PER ROW");
	ncards++;
	cp += 80;

        sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "AMPSCOL",hdata->amps_per_column,"AMPLIFIERS PER COLUMN");
	ncards++;
	cp += 80;

        sprintf(cp,"%-8.8s= %20.6f / %-47.47s",
        "EXPTIME",(float)hdata->exposure/100.0,
		"Exp time (not counting shutter error)");
	ncards++;
	cp += 80;

        sprintf(cp,"%-8.8s= %20.6f / %-47.47s",
        "BSCALE",1.0,"DATA SCALE FACTOR");
	ncards++;
	cp += 80;

        sprintf(cp,"%-8.8s= %20.6f / %-47.47s",
        "BZERO",32768.0,"DATA ZERO POINT");
	ncards++;
	cp += 80;

	sprintf(cp,"%-80.80s","COMMENT = 'Real Value = FITS*BSCALE+BZERO'");
	ncards++;
	cp += 80;

        sprintf(cp,"%-8.8s= %-20.20s / %-47.47s",
        "PROGRAM","'NEWCAM'",
		"New Lick Camera");
	ncards++;
	cp += 80;

	sprintf(verstr,"'%s'",version);
        sprintf(cp,"%-8.8s= %-20.20s / %-47.47s",
        	"VERSION",verstr,"Data acquisition version");
	ncards++;
	cp += 80;

	tv = FITS_TIME(id);
	sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "TSEC",(int)tv->tv_sec,"CLOCK TICK - SECONDS");
	ncards++;
	cp += 80;

	sprintf(cp,"%-8.8s= %20d / %-47.47s",
        "TUSEC",(int)tv->tv_usec,"CLOCK TICK - MICROSECONDS");
	ncards++;
	cp += 80;

	sprintf(cp,"%-8.8s= %-24.24s / %-43.43s",
		"DATE-OBS",
		datetime(tv,timestr),"UTC DATE AND TIME OF OBSERVATION");
	ncards++;
	cp += 80;

/*	Copy fits cards captured at the start of the exposure	*/

	nc = record->fits_begin;
	if(nc) {
	    memcpy(cp, FITS_START(id),nc*80);
	    cp += nc*80;
	    ncards += nc;
	}

/*	Copy fits cards captured at the end of the exposure	*/
	nc = record->fits_end;
	if(nc) {
	    memcpy(cp, FITS_END(id),nc*80);
	    cp += nc*80;
	    ncards += nc;
	}

/*	Copy fits cards captured at other times			*/
	nc = record->fits_other;
	if(nc) {
	    memcpy(cp, FITS_OTHER(id),nc*80);
	    cp += nc*80;
	    ncards += nc;
	}

        sprintf(cp,"%-8.8s%72.72s","END"," ");
	ncards++;
	return(ncards);
}

int read_image(int sem_read)
{
	struct sembuf sops;
	int ret;
	static int oldret=-10000,retcnt=0;
	static int err = 1;
	static int rbyte1 = IFCMD_GUIDER|IFCMD_24BIT|IFCMD_XFER16|IFCMD_GRESET;
//	static int rbyte2 = IFCMD_GUIDER|IFCMD_24BIT|IFCMD_XFER16|IFCMD_RESET;
	static int cbyte = IFCMD_GUIDER|IFCMD_24BIT|IFCMD_XFER16;
	struct timespec req;
	static int bufnum = 0;
	char *nextim;
	char *endbuf;
#ifndef IFCMD_200MHZ
#define IFCMD_200MHZ IFCMD_HIBIT	/* High order bit specifies the	*/
#endif					/* fast fiber on the pci	*/
					/* interface.			*/

/*	First reset fifo if there was an error last time.	*/
/*	Since err is initialized to 1 we also reset on the first*/
/*	call to this function.					*/

	if(err) {
	    flogerror("resetting interface and driver");
	    err = 0;
	    if(fastfiber) {
		rbyte1 |= IFCMD_200MHZ;
		cbyte  |= IFCMD_200MHZ;
	    }
	    ret = ioctl(im_fd,CYP_RESET,NULL);
	    if(ret < 0) {
                flogerror("CYP_RESET ioctl error: %s",strerror(errno));
                return 0;
            }

/*	    Then switch interface into GUIDER mode - no RESET.		*/
	    ret = ioctl(im_fd,DPP_WCMD,&cbyte);
	    if(ret < 0) {
                flogerror("DPP_WCMD ioctl error: %s",strerror(errno));
                return 0;
	    }
        }


/*	Read the image (we also get the header).			*/
/*	Note that we don't have to know the actual size of the		*/
/*	image ahead of time because that information is transmitted	*/
/*	in the image header, which the driver reads for us.		*/
/*	The return value will tell us how many bytes were actually read.*/
/*	Our buffer just has to be big enough to hold any possible image.*/

/*	Force a short sleep.						*/
        req.tv_sec = 0;
        req.tv_nsec = 1000*1000;
        nanosleep(&req,NULL);

/*	Get the next buffer to use that will fit into the raw		*/
/*	image buffer.							*/
	
	endbuf = nbytes+shad2;		/* First byte past the end of	*/
					/* the raw image buffer.	*/
/*	shm_nbytes is the size of the image descramble buffer. This	*/
/*	should be the maximum size of any image we might receive.	*/
/*	So we will assume we can safely divide the raw image buffer	*/
/*	into blocks that are the size of shm_nbytes.			*/
	nextim = shm_nbytes*bufnum + shad2;

/*	If any of the chosen buffer is beyond the actual end of the	*/
/*	raw image buffer then start back over at the beginning of the	*/
/*	raw image buffer.						*/
	if((nextim > endbuf) || ((nextim+shm_nbytes) > endbuf)) {
	    nextim = shad2;
	    bufnum = 0;
	}

/*	This is the buffer address for the image we are about to read.	*/
	image = (unsigned short *)nextim;

/*	Increment our buffer number for the next image.			*/
	bufnum++;

//	memset((char *)image,0,64);	/* Clear old header.		*/

	ret = read(im_fd,(char *)image,shm_nbytes);
	if(ret != oldret) {
	    flogerror("read of image returned %d",ret);
	    oldret = ret;
	    retcnt = 0;
	} else if (++retcnt >= 30) {
	    flogerror("read of image returned %d (%d times)",ret,retcnt);
	    retcnt = 0;
	}


/*	Wait for descramble routine to wake up and decrement our	*/
/*	synchronization semaphore. Then tell the descramble routine	*/
/*	that we are awake again.					*/

	if((ret > 0) && (sem_read >= 0)) {
	    int sem_ret;
#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
            /* union semun is defined by including <sys/sem.h> */
#else
            /* according to X/OPEN we have to define it ourselves */
            union semun {
             int val;                  /* value for SETVAL */
             struct semid_ds *buf;     /* buffer for IPC_STAT, IPC_SET */
             unsigned short *array;    /* array for GETALL, SETALL */
                                       /* Linux specific part: */
             struct seminfo *__buf;    /* buffer for IPC_INFO */
            };
#endif
	    union semun semun; 

/*	    First wait on semaphore 0 if it is not now 0.		*/

	    sops.sem_num = 0;
	    sops.sem_op = 0;
	    sops.sem_flg = 0;
//	      flogerror("waiting on descramble semaphore");
	    sem_ret = semop(sem_read,&sops,1);
	    if (sem_ret < 0) {
		flogerror("camread: semop synchonization semaphore failed: %s",
		    strerror(errno));
		flogerror("exiting!!");
		exit(1);
	    }

/*	    Once we wake up we set the second semaphore to 0 to signal	*/
/*	    the descramble routine that we are no longer blocked.	*/
	    semun.val = 0;
	    sem_ret = semctl(sem_read,1,SETVAL,semun);
	    if (sem_ret < 0) {
		flogerror("camread: semctl setting of semaphore 1 failed: %s",
		    strerror(errno));
		flogerror("exiting!!");
		exit(1);
	    }
	}

	if(ret < 0) {
	    char *errtxt;
	    char status[512];
	    if(ret == -ENODEV) {
		flogerror("read of image returned ENODEV.");
		return 0;
	    }
/*	    If we got an error then tell the driver we want the	*/
/*	    detailed error message and then read the message and*/
/*	    print it.						*/
	    errtxt = loglasterror();
	    sprintf(status,
		"ERROR Error >>%s<< while reading image data." "#n"
		"This probably means the image did not fully read out.",
		errtxt);
//	    save_error_status(status);
	    err = 1;		/* Cause init on next read.	*/
	    id = image[0]&0xff;
	    c = image[2]+256*image[3];
	    r = image[4]+256*image[5];
	    flogerror(
		"bad or lost image with %d rows and %d columns, image_type %d",
		    r,c,id);
/*	    Wait a little bit to give other camread branches a chance	*/
/*	    to register an error.					*/
	    usleep(200000);		/* 0.2 seconds			*/
            return 0;
        }

	if(ret < (fastfiber ? HEADER_SIZE :12)) {
            flogerror("read only %d bytes so dont have header",ret);
	    err = 1;
/*	    Wait a little bit to give other camread branches a chance	*/
/*	    to register an error.					*/
	    usleep(200000);		/* 0.2 seconds			*/
            return 0;
        }

//	flogerror("read image of %d bytes",ret);

	return 1;
}

void simulate_image(int simulation,int imtype,int headersize,int r,int c)
{
	int rn,cn,px,incr,val;

	switch (simulation) {
	case 0:
	    break;
	default:
/*	    Simulate image as a linear ramp over the full range of the	*/
/*	    possible data values. Each row looks the same.		*/
	    incr = 65535/c;
	    px = headersize;
	    for(rn=0; rn<r; rn++) {
		val = 0;
		for(cn=0; cn<c; cn++) {
		    sim_im[px] = val;
		    val += incr;
		    px++;
		}
	    }
	}
	return;
}

/*	Tell the device driver we want the last detailed error message.	*/
/*	and then read that message.					*/
/*	Log the message to our error logfile.				*/
char *loglasterror(void)
{
/*	NOTE: I had to make this array static. Otherwise camread would	*/
/*	exit on the return from this function. I don't understand this.	*/
/*	It seems like a GCC compiler bug (RedHat 9).  rjs		*/

	static char last_error[128];

	strcpy(last_error,"NO ERROR RETURNED\n");
	if(ioctl(im_fd,DPP_LAST_ERROR,0) != 0) {
	    flogerror("%s LAST_ERROR ioctl returned error: %s",
		__FUNCTION__,strerror(errno));
	    return("LAST_ERROR ioctl error");
	}
	if(read(im_fd,last_error,sizeof(last_error)) <= 0) {
	    flogerror("%s read of last error returned an error: %s",
		__FUNCTION__,strerror(errno));
	    return("read of last error returned an error");
	}
	flogerror("Error Details:%s",last_error);
	return(last_error);
}

int get_image_param(short *addr)
{
        int param;
        param = (*addr & 0xff);
        addr++;
        param |= (*addr & 0xff)<<8;
        return(param);
}
char *get_cam_hostname(char *host)
{
	static char ghostname[32];
	char hostname[32];
	char *hostp;

	if(*host == '\0')
		gethostname(hostp=hostname,sizeof(hostname));
	else
		hostp = host;
	read_remote(hostp,"gcamhost",ghostname,sizeof(ghostname));
	if(ghostname[0] == '\0')
		strcpy(ghostname,"ccdlab");
	return(ghostname);
}

int get_cam_portnum(void)
{
	char gportnum[32];
	int gport;

	read_remote("network","guiderport",gportnum,sizeof(gportnum));
	if(gportnum[0] == '\0')
		strcpy(gportnum,"5684");
	sscanf(gportnum,"%d",&gport);
	return(gport);
}

/*      Start up connection with the traffic controller.        */
void connect_to_traffic(char *proc)
{
        static char hostname[30];

/*      Get our local host name.                                */
        if(strlen(hostname) == 0)
                gethostname(hostname,sizeof(hostname));

/*      Try to make a connection to traffic.                    */
/*	first ry our host name. If that fails try localhost.	*/
        our_addr = tremoteopen(proc,hostname,&traffic_fd);
        if(our_addr >= 0) {
                traffic_ok = 1;
        } else {
	    our_addr = tremoteopen(proc,"localhost",&traffic_fd);
	    if(our_addr < 0) {
                flogerror(
        "Error: camread-%s can't connect to traffic",proc);
	    } else {
	        traffic_ok = 1;
	    }
	}

/*      We don't want to receive NO_INTEREST messages.          */
        if(traffic_ok) {
            mstart(TRAFFIC_CONTROLLER,our_addr,WANT_NO_INTEREST);
            mputl(0);
            msend(traffic_fd);
        }

        return;
}

int ConnectGcam(char **devicename)
{
	int fd;
	fd = open(fiberport,O_RDWR);
	*devicename = fiberport;
	return(fd);
}

/*	Fill in the elements of the hdata structure by		*/
/*	decoding the bytes in the image header.			*/

static void decode_header(short *image,struct header_data *hdata)
{
	static int old_id = -10000;
	int loghead = 0;

	memset(hdata,0,sizeof(struct header_data));

	logbytes(image,32*2);

/*	Bits 0-7 of first word contains the image type.		*/

	hdata->type = image[0]&0xff;

/*	Bits 8-15 of 1st word contains the header size (bytes).	*/

	hdata->size = (image[0]>>8)&0xff;

/*	Bits 0-6 of word 1 contains the image ID.		*/
	hdata->id = image[1]&0x7f;

/*	Bit 7 of word 1 contains the checksum flag. If set the	*/
/*	camera has supplied a checksum.				*/
	hdata->checksum_flag = ((image[1]&0x80) == 0x80) ? 1 : 0;

/*	Log when we first see a new image id.			*/
	if(old_id != hdata->id) {
	    flogerror("decoding new image id %d",hdata->id);
	    loghead = 1;
	    old_id = hdata->id;
	}

/*	Bits 0-7 of word 2 contains the low byte of the number	*/
/*	of columns.						*/
/*	Bits 0-7 of the word 3 contains the high byte of the	*/
/*	number of columns.					*/

/*	This value is the number of pixels per amplifier.	*/
#define TWOBYTES(i) ((0xff&image[i])|((0xff&image[i+1])<<8))
	hdata->per_amp.c = TWOBYTES(2);

/*	Bits 0-7 of word 4 contains the low byte of the		*/
/*	number of rows.						*/
/*	Bits 0-7 of word 5 contains the high byte of		*/
/*	the number of rows.					*/

/*	This value is the number of rows per amplifier.		*/
	hdata->per_amp.r = TWOBYTES(4);

/*	Perhaps the header will supply these later:		*/

	hdata->per_amp.rorg = hdata->per_amp.corg = 0;

/*	Exposure time is stored in low bytes of words 6, 7, and 8.	*/
	hdata->exposure = image[6] + ((image[7]&0xff)<<8) +
		((image[8]&0xff)<<16);

/*	Shutter state is stored in word 9.			*/

	hdata->shutter = image[9]&0xf;

/*	Bits 0-7 of word 10 contains the low byte of the number	*/
/*	of overscan columns.					*/
/*	Bits 0-7 of the word 11 contains the high byte of the	*/
/*	number of overscan columns.				*/

/*	This value is the number of pixels per amplifier.	*/

	hdata->over.c = TWOBYTES(10);

/*	Bits 0-7 of word 12 contains the low byte of the	*/
/*	number of rows.						*/
/*	Bits 0-7 of word 13 contains the high byte of		*/
/*	the number of rows.					*/

/*	This value is the number of rows per amplifier.		*/

	hdata->over.r = TWOBYTES(12);

	hdata->over.rorg = hdata->over.corg = 0;

/*	Bits 0-7 of word 14 contains the low byte of the 	*/
/*	column origin of the data window.			*/
/*	Bits 0-7 of the word 15 contains the high byte of the	*/
/*	column origin of the data window.			*/

	hdata->window.corg = TWOBYTES(14);

/*	Bits 0-7 of word 16 contains the low byte of the	*/
/*	row origin of the data window.				*/
/*	Bits 0-7 of word 17 contains the high byte of the	*/
/*	row origin of the data window.				*/

	hdata->window.rorg = TWOBYTES(16);

/*	Bits 0-7 of word 18 contains the low byte of the number	*/
/*	of window columns.					*/
/*	Bits 0-7 of the word 19 contains the high byte of the	*/
/*	number of window columns.				*/

	hdata->window.c = TWOBYTES(18);

/*	Bits 0-7 of word 20 contains the low byte of the	*/
/*	number of window rows.					*/
/*	Bits 0-7 of word 21 contains the high byte of		*/
/*	the number of window rows.				*/

	hdata->window.r = TWOBYTES(20);

	switch(hdata->type) {
	case 0:		/* One amp at (0,0)		*/
	case 50:
	case 1:		/* One amp at (C,0)		*/
	case 2:		/* One amp at (0,R)		*/
	case 3:		/* One amp at (C,R)		*/
		hdata->amps_per_row = 1;
		hdata->amps_per_column = 1;
		break;
	case 4:		/* Two amps on same serial at (0,0) and (C,0)	*/
		hdata->amps_per_row = 2;
		hdata->amps_per_column = 1;
		break;
	case 5:		/* Two amps opposite sides of CCD, (0,0) and (C,R)*/
	case 6:		/* Two amps opposite sides of CCD, (0,0) and (0,R)*/
	case 7:		/* Two amps opposite sides of CCD, (C,0) and (0,R) should be (C,R) and (0,R) ?*/
		hdata->amps_per_row = 2;
		hdata->amps_per_column = 1;
		break;
	case 8:		/* Four amplifiers.				*/
	case 10:	/* Four amplifiers in LL AO CCD configuration.	*/
		hdata->amps_per_row = 2;
		hdata->amps_per_column = 2;
		break;
	case 9:		/* Fours amplifiers per row in LL AO CCD	*/
			/* configuration.				*/
	case 64:
		hdata->amps_per_row = 4;
		hdata->amps_per_column = 1;
		break;
        case 11:
        case 12:
        case 13:	/* Two amp at (0,0) (0,R) (add by Roy)*/
        case 14:
		hdata->amps_per_row = 1;
		hdata->amps_per_column = 2;
		break;
	case 48:	/* Eight amps					*/
		hdata->amps_per_row = 4;
		hdata->amps_per_column = 2;
		break;
	default:
		flogerror("Unrecognized image type %d",hdata->type);
		flogerror("Assuming one amp per row or column.");
		hdata->amps_per_row = 1;
		hdata->amps_per_column = 1;
	}

/*	If the header size is at least (30*2) bytes then we have the	*/
/*	binning information passed to us.				*/
	if (header_size >= (30*2)) {
	    int bin,rbin,cbin,bc;

/*	    Byte 26 of the header contains the encoded binning.	*/

	    bin = 0xff&image[26];

/*	    If the high bit is set then we have separate row and	*/
/*	    column binning parameters. Otherwise row and column binning	*/
/*	    are the same and are encoded in the lower three bits.	*/

	    if((bin&0x80) == 0x80) {
		rbin = (bin&0x70)>>4;
		cbin = (bin&0x07);
	    } else {
		rbin = cbin = (bin&0x07);
	    }

/*	    The actual binning is 2^X where X is the number in the	*/
/*	    header.							*/

	    hdata->row_bin = 1;
	    hdata->col_bin = 1;
	    for(bc=0; bc<rbin; bc++)
	        hdata->row_bin = hdata->row_bin<<1;
	    for(bc=0; bc<cbin; bc++)
		hdata->col_bin = hdata->col_bin<<1;
/*
	    flogerror("From header got rbin=%d cbin=%d",
		hdata->row_bin,hdata->col_bin);
*/
	} else if (guider_mode) {
/*		For early versions of the guider camera software	*/
/*		we just assume binning=1 because that is all that is	*/
/*		allowed.						*/
		hdata->row_bin = hdata->col_bin = 1;
//	    flogerror("In guider mode assumed binning=1");
	} else {
		RetrieveBinning(hdata->id,&hdata->col_bin,&hdata->row_bin);
/*
	    flogerror("From OTHER FITS buffer got rbin=%d cbin=%d",
		hdata->row_bin,hdata->col_bin);
*/
	}
		
/*	Compute size of transmitted image (before the window is		*/
/*	extracted.)							*/
	hdata->image.r = hdata->per_amp.r*hdata->amps_per_column;
	hdata->image.c = hdata->per_amp.c*hdata->amps_per_row;

/*	Perhaps the header will supply these later:		*/

	hdata->image.rorg = hdata->image.corg = 0;

/*	Compute the size of the windowed image, including overscan	*/

	hdata->final.r = hdata->window.r +
		hdata->over.r*hdata->amps_per_column;
	hdata->final.c = hdata->window.c +
		hdata->over.c*hdata->amps_per_row;

/*	The 'final' origin is the origin of the window in CCD pixel	*/
/*	coordinates.							*/

/*	Columns: (1) If we are reading from a left-side amplifier then	*/
/*	we will not be flipping the image so we add the window origin	*/
/*	(which is relative to the transmitted image) to the origin of	*/
/*	the transmitted	image (which is in CCD pixels). We want this	*/
/*	final origin to be in CCD pixels so we must multipy the window	*/
/*	origin by the binning factor.					*/
/*	Columns: (2) If we are reading from a right-side amplifier then	*/
/*	we will be flipping the image. So we must translate the origin	*/
/*	to be relative to the left edge of the flipped image. Do this	*/
/*	by subtracting the pixel number of the left-most pixel from the	*/
/*	number of columns in the sensor.				*/

/*	Rows: (1) If we are reading from a top-row amplifier then	*/
/*	we will not be flipping the image so we add the window origin	*/
/*	(which is relative to the transmitted image) to the origin of	*/
/*	the transmitted	image (which is in CCD pixels). We want this	*/
/*	final origin to be in CCD pixels so we must multipy the window	*/
/*	origin by the binning factor.					*/
/*	Rows: (2) If we are reading from a bottom-row amplifier then	*/
/*	we will be flipping the image. So we must translate the origin	*/
/*	to be relative to the upper edge of the flipped image. Do this	*/
/*	by subtracting the pixel number of the upper-most pixel from the*/
/*	number of rows in the sensor.					*/

/*	Bits 0-7 of word 22 contains the low byte of the	*/
/*	transmitted image column origin.			*/
/*	Bits 0-7 of word 23 contains the high byte of the	*/
/*	transmitted image column origin.			*/

/*	Bits 0-7 of word 24 contains the low byte of the	*/
/*	transmitted image row origin.				*/
/*	Bits 0-7 of word 25 contains the high byte of the	*/
/*	transmitted image row origin.				*/

	switch(hdata->type) {

	case 0:		/* One amp at (0,0)		*/
	case 4:		/* Two amps on same serial at (0,0) and (C,0)	*/
	case 5:		/* Two amps opposite sides of CCD, (0,0) and (C,R)*/
	case 6:		/* Two amps opposite sides of CCD, (0,0) and (0,R)*/ /*(0,R) (C,R)?? */ 
	case 8:		/* Four amplifiers.				*/
	case 9:		/* Four amplifiers in LL AO CCD configuration.	*/
	case 10:	/* Fours amplifiers per row in LL AO CCD	*/
			/* configuration.				*/
        case 13:	/* Two amps at (0,0) and (0,R) */
        case 11:
        case 12:
	case 48:
	case 64:
	    hdata->final.corg = hdata->col_bin*hdata->window.corg +
		TWOBYTES(22);
	    hdata->final.rorg = hdata->row_bin*hdata->window.rorg + 
		TWOBYTES(24);
	    break;

	case 1:		/* One amp at (C,0)		*/
	case 50:
	case 7:		/* Two amps opposite sides of CCD, (C,0) and (0,R)*/
        case 14:
/*	    If col_size is not defined (<=0) that means we didn't find	*/
/*	    col_size in the database. So just pretend we aren't going to*/
/*	    flip the image. This isn't the correct result, but we don't */
/*	    have enough info to compute the correct result. The lack	*/
/*	    of the col_size in the database has already been logged.	*/
	    if (col_size > 0) {
	        hdata->final.corg = col_size + hdata->over.c -
		(hdata->col_bin*(hdata->window.corg + hdata->final.c) +
		TWOBYTES(22));
	    } else {
	        hdata->final.corg = hdata->col_bin*hdata->window.corg +
		TWOBYTES(22);
	    }
	    hdata->final.rorg = hdata->row_bin*hdata->window.rorg + 
		TWOBYTES(24);
	    break;

	case 2:		/* One amp at (0,R)		*/
	    hdata->final.corg = hdata->col_bin*hdata->window.corg +
		TWOBYTES(22);
/*	    Note comment about col_size above. Same comment applies here. */
	    if (row_size > 0) {
		hdata->final.rorg = row_size + hdata->over.r - 
		(hdata->row_bin*(hdata->window.rorg + hdata->final.r) +
		TWOBYTES(24));
	    } else {
	        hdata->final.rorg = hdata->row_bin*hdata->window.rorg + 
		TWOBYTES(24);
	    }
	    break;

	case 3:		/* One amp at (C,R)		*/
	    if (col_size > 0) {
		hdata->final.corg = col_size + hdata->over.c -
		(hdata->col_bin*(hdata->window.corg + hdata->final.c) +
		TWOBYTES(22));
	    } else {
	        hdata->final.corg = hdata->col_bin*hdata->window.corg +
		TWOBYTES(22);
	    }
	    if (row_size > 0) {
		hdata->final.rorg = row_size + hdata->over.r - 
		(hdata->row_bin*(hdata->window.rorg + hdata->final.r) +
		TWOBYTES(24));
	    } else {
	        hdata->final.rorg = hdata->row_bin*hdata->window.rorg + 
		TWOBYTES(24);
	    }
	    break;

	default:
		flogerror("Unrecognized image type %d",hdata->type);
		flogerror("Assuming natural CCD order.");
	    hdata->final.corg = hdata->col_bin*hdata->window.corg +
		TWOBYTES(22);
	    hdata->final.rorg = hdata->row_bin*hdata->window.rorg + 
		TWOBYTES(24);
	    break;
	}


	if(loghead) {
	    logdecoded(hdata);
//	    logbytes(image,32);
	}
}

void logbytes(void *data,int ndata)
{
 	unsigned char *dp = (unsigned char *)data;
	int n;

	for(n=0; n<ndata; n++)
	    flogerror("data byte %d = %x",n,*(dp+n));
}

void logdecoded(struct header_data *hdata)
{
	flogerror("header data:");
	flogerror("    type %d id %d amp/r %d amp/c %d",
		hdata->type,hdata->id,hdata->amps_per_row,
		hdata->amps_per_column);
	flogerror("    Per-amp rxc %dx%d r0,c0 %d,%d",
		hdata->per_amp.r,hdata->per_amp.c,
		hdata->per_amp.rorg,hdata->per_amp.corg);
	flogerror("    image   rxc %dx%d r0,c0 %d,%d",
		hdata->image.r,hdata->image.c,
		hdata->image.rorg,hdata->image.corg);
	flogerror("    window  rxc %dx%d r0,c0 %d,%d",
		hdata->window.r,hdata->window.c,
		hdata->window.rorg,hdata->window.corg);
	flogerror("    final  rxc %dx%d r0,c0 %d,%d",
		hdata->final.r,hdata->final.c,
		hdata->final.rorg,hdata->final.corg);
	flogerror("    over    rxc %dx%d",hdata->over.r,hdata->over.c);
}

/*      Use sqlget to retrieve a value from the database.		*/
/*      The 'format' string tells us how to convert the string we get   */
/*      from the database.                                              */
/*      Return -1 if some error occurs (and log an error message).      */
/*      Return 1 and the converted value in '*item' if we encounter     */
/*      no problems.                                                    */

int dbget(char *format,char *dbitem,void *item)
{
	int ncon;
        char dbname[32],dbtable[32],dbcol[32],dbrow[32];
        int ret;
        struct tuple value[2] = {{0,0,(char *)0,(char *)0},
                                 {0,0,(char *)0,(char *)0}};
	static sem_t sem;
        static int sem_is_init = 0;

/*	Access to the database does not seem to be thread-safe.     */
/*	So here we make sure that only one thread is talking to     */
/*	the database at one time.                                   */

/*	Initialize our thread semaphore if that has not been done.  */

	if(sem_is_init == 0) {
            sem_is_init++;
            sem_init(&sem,0,1);
	}

/*	Wait if we are not the first thread in this routine.        */
/*	If we do block here, a sem_post(&sem) will free us later.   */

	sem_wait(&sem);

/*      First break the source location field into its parts.   */
	sscanf(dbitem,
	"%[a-zA-Z0-9_-].%[a-zA-Z0-9_-].%[a-zA-Z0-9_-].%[a-zA-Z0-9_-]",
		dbname,dbtable,dbrow,dbcol);


        ret = sqlget("select",dbname,dbtable,dbrow,dbcol,"__STD",&value[0],2);
        if ((ret < 0) || (value[0].value == NULL)) {
            flogerror("Database error getting %s: %s",dbitem,value[0].value);
	    freetuple(&value[0]);
	    sem_post(&sem);
	    return -1;
        }

/*	If the format is the special case %S then just copy the		*/
/*	entire string just read.					*/
	if (strcmp("%S",format) == 0) {
	    int nc;
	    int trim = 1;
	    char c;
/*	    Trim off trailing \r or \n					*/
	    while(trim) {
	        nc = strlen(value[0].value)-1;
	        if(nc >= 0) {
		    c = value[0].value[nc];
	    	    if((c == '\r') || (c == '\n')) {
			value[0].value[nc] = '\0';
		    } else {
			trim = 0;
		    }
		} else {
		    trim = 0;
		}
	    }
	    strcpy(item,value[0].value);
	    freetuple(&value[0]);
	    sem_post(&sem);
	    return 1;
	}

/*	Convert the result according to the specified format.		*/
        if((ncon=sscanf(value[0].value,format,item)) != 1) {
            flogerror(
	"dbget %s failed sscanf of %s with format %s: return was %d",
		dbitem,value[0].value,format,ncon);
	    freetuple(&value[0]);
	    sem_post(&sem);
            return -1;
        }
	freetuple(&value[0]);
	sem_post(&sem);
        return 1;
}

/*      Use the dbshow command to retrieve a value from the database.   */
/*	The only reason to use dbshow instead of dbget is that dbshow	*/
/*	works even for the superuser.					*/
/*      The 'format' string tells us how to convert the string we get   */
/*      from the database.                                              */
/*      Return -1 if some error occurs (and log an error message).      */
/*      Return 1 and the converted value in '*item' if we encounter     */
/*      no problems.                                                    */

int dbshow(char *format,char *dbitem,void *item)
{
        FILE *fp;
        char line[512];
        char cmd[256];
        static char *dbcmd = NULL;
	char *db_env;
	int ncon;

/*	If the path for the dbshow command is not set then look for	*/
/*	the path in the environment. Failing that, use a default.	*/
	if(dbcmd == NULL) {
	    db_env = getenv("DBSHOW");
	    if(db_env == NULL)
		dbcmd = UCAM_BIN_DIR "/dbshow";
	    else
		dbcmd = db_env;
	}

/*	Construct the command to execute.				*/
        sprintf(cmd,"%s %s",dbcmd,dbitem);

/*	Run the command and read the result.				*/
        fp = popen(cmd,"r");
        if (fp == NULL) {
            flogerror("dbshow (%s) failed: %s\n",cmd,strerror(errno));
            return -1;
        }
	memset(line,0,sizeof(line));
        fgets(line,sizeof(line),fp);
        pclose(fp);

/*	If the format is the special case %S then just copy the		*/
/*	entire string just read.					*/
	if (strcmp("%S",format) == 0) {
	    int nc;
	    int trim = 1;
	    char c;
/*	    Trim off trailing \r or \n					*/
	    while(trim) {
	        nc = strlen(line)-1;
	        if(nc >= 0) {
		    c = line[nc];
	    	    if((c == '\r') || (c == '\n')) {
			line[nc] = '\0';
		    } else {
			trim = 0;
		    }
		} else {
		    trim = 0;
		}
	    }
	    strcpy(item,line);
	    return 1;
	}

/*	Convert the result according to the specified format.		*/
        if((ncon=sscanf(line,format,item)) != 1) {
            flogerror(
		"dbshow failed sscanf of %s with format %s: return was %d",
		line,format,ncon);
            return -1;
        }
        return 1;
}

/*      Use the dbset command to save a status value to the database.	*/
/*      Return -1 if some error occurs (and log an error message).      */

void save_error_status(char * status)
{
        FILE *fp;
        char line[512];
        char cmd[512];
        static char *dbcmd = NULL;
	char *db_env;

/*	If the path for the dbset command is not set then look for	*/
/*	the path in the environment. Failing that, use a default.	*/
	if(dbcmd == NULL) {
	    db_env = getenv("DBSET");
	    if(db_env == NULL)
		dbcmd = UCAM_BIN_DIR "/dbset";
	    else
		dbcmd = db_env;
	}

/*	Construct the command to execute.				*/
/*	If the image_aborted flag is set then turn the message into	*/
/*	a warning rather than an error.					*/
	if (image_aborted) {
	    sprintf(line,
		"imageid.variables.camread_status.value \"ABORT_%s\"",status);
	    image_aborted = 0;
	} else {
	    sprintf(line,
		"imageid.variables.camread_status.value \"%s\"",status);
	}
        sprintf(cmd,"%s -ok %s",dbcmd,line);

/*	Run the command and read the result.				*/
        fp = popen(cmd,"r");
        if (fp == NULL) {
            flogerror("%s: dbset (%s) failed: %s\n",__FUNCTION__,
		cmd,strerror(errno));
            return;
        }
        fgets(line,sizeof(line),fp);
        pclose(fp);

}

/*	Set our UID to a user known to the database.			*/
static void set_dbuid(void)
{
	uid_t euid,uid;
	uid_t db_uid;
	struct passwd *pw = NULL;
	char *db_pwname;
	int ret;

/*      Make sure our uid is one that can access the database.		*/
	uid = getuid();
	euid = geteuid();

	db_pwname = getenv("DB_USER");
	if(db_pwname == NULL) {
	    db_pwname = "ccdev";		/* Fallback for Lick	*/
	}

/*	Get the UID of a user name know to the database.		*/

	pw = getpwnam(db_pwname);

/*	If the specified user isn't in the password file then we hope	*/
/*	our current uid will work.					*/
	if(pw == NULL) {
	    db_uid = uid;	/* Hope this works!	*/

/*	    Look up our user name.					*/
	    pw = getpwuid(uid);
	    if(pw == NULL)
		db_pwname = "unknown";
	    else
		db_pwname = pw->pw_name;
	} else {
	    db_uid = (uid_t)pw->pw_uid;
	}
	ret = setuid(db_uid);
	if(ret < 0) {
	    flogerror("Unable to setuid(%d): user %s: %s",db_uid,db_pwname,
		strerror(errno));
	}
}


/*	Given the version name, which was read previously from the	*/
/*	database, read all of the version-dependent information we need.*/

void init_version_data(void)
{
	char result[64];
	int pn;
	int ret;

/*	Given our version, get the telescope position port number	*/
	sprintf(result,"telefocus.configs.%s.position_port",version);
	ret = dbget("%d",result,&pn);
	if (ret == -1)
	    telpos_port = -1;
	else
	    telpos_port = pn;
	flogerror("Running version %s with telpos port %d",
		version,telpos_port);

/*	Get the simulator flag.						*/
	sprintf(result,"telefocus.configs.%s.simulation",version);
	flogerror("Getting simulation flag %s",result);
	ret = dbget("%d",result,&pn);
        if (ret == -1)
	    simulation = 0;
	else
	    simulation = pn;
	flogerror("Doing simulation %d (0==real data)",simulation);

/*	Get the fiber device name.					*/
	sprintf(result,"telefocus.configs.%s.fiberport",version);
	flogerror("Getting fiberport device name: %s",result);
	ret = dbget("%s",result,result);
	if(ret == -1)
		strcpy(fiberport,"/dev/cyp0");
	else
		strcpy(fiberport,result);

/*      Get the data interface configuration flag. This flag tells      */
/*      us if we need to do any special configurations of the fiber	*/
/*	interface. For the dual-channel USB interface this tells us the	*/
/*      channel or channels used for the incoming data. Also we		*/
/*	know how to decode the incoming data in this case.		*/

        sprintf(result,"telefocus.configs.%s.iflag",version);
        flogerror("Getting fiber interface flag: %s",result);
        ret = dbget("%d",result,&pn);
        if(ret == -1) {
                interface_flag = 0;     /* Assume old interface         */
		flogerror("Failed to read %s from database",result);
        } else {
                interface_flag = pn;
	}
        flogerror("Using interface flag %d: %s",interface_flag,
		interface_meaning(interface_flag));

/*	Get the unrecorded image destination.				*/
	sprintf(result,"telefocus.configs.%s.unrecorded",version);
	flogerror("Getting unrecorded image destination");
	ret = dbget("%s",result,unrecorded_file);
	if ((unrecorded_file[0] == '\0') ||
	    (strcmp(unrecorded_file,"NOT")==0)) {
		record_unrecorded = 0;
		flogerror("Unrecorded images will not be recorded anywhere.");
	} else {
		unrecorded_file[sizeof(unrecorded_file)-1] = '\0';
		record_unrecorded = 1;
		flogerror("Unrecorded images will be recorded in file %s",
			unrecorded_file);
	}

/*	Get the image header size.					*/
	sprintf(result,"telefocus.configs.%s.headsize",version);
	flogerror("Getting header size");
        ret = dbget("%d",result,&pn);
        if(ret == -1) {
		header_size = 26*2;
                flogerror(
	    "Database error: Header size assumed to be %d",header_size/2);
        } else {
		header_size = pn*2;
                flogerror("Header size from database = %d",pn);
        }

/*	Get the number of rows in the sensor.				*/

	sprintf(result,"telefocus.configs.%s.maxrows",version);
	flogerror("Getting sensor number of rows");
        ret = dbget("%d",result,&pn);
        if(ret == -1) {
		row_size = -1;
                flogerror(
	    "Database error: row_size assumed to be %d",row_size);
        } else {
		row_size = pn;
                flogerror("row_size from database = %d",pn);
        }

/*	Get the number of columns in the sensor.			*/

	sprintf(result,"telefocus.configs.%s.maxcols",version);
	flogerror("Getting sensor number of columns");
        ret = dbget("%d",result,&pn);
        if(ret == -1) {
		col_size = -1;
                flogerror(
	    "Database error: col_size assumed to be %d",col_size);
        } else {
		col_size = pn;
                flogerror("col_size from database = %d",pn);
        }

}

/*	Return a string that describes the meaning of the interface	*/
/*	flag value.							*/
static char *interface_meaning(int val)
{
	char *meaning;
	switch(val) {
	case 0:
		meaning = "Standard single-channel";
		break;
	case 1:
		meaning = "Dual channel interface, using channel A";
		break;
	case 2:
		meaning = "Dual channel interface, using channel B";
		break;
	case 3:
		meaning = "Dual channel interface, using channels A and B";
		break;
	default:
		meaning = "Unknown code";
	}
	return(meaning);
}


/*	Given the interface flag do any necessary interface setup.	*/
static void interface_setup(int val)
{
	char *meaning;
	int chan;
	int ret;
 	meaning = interface_meaning(val);
	switch(val) {
	case 0:
		break;		/*Nothing special to do.		*/
	case 1:
/*	    Configure USB interface to use fiber channel A		*/
	    chan = USB_FIBER_CHAN_A;
	    ret = ioctl(im_fd,USBFO_USB_CTL,&chan);
	    if (ret) {
		flogerror("intface_setup USBFO_USB_CTL ioctl error:");
		(void)loglasterror();
	    }
	    break;
	case 2:
/*	    Configure USB interface to use fiber channel B		*/
	    chan = USB_FIBER_CHAN_B;
	    ret = ioctl(im_fd,USBFO_USB_CTL,&chan);
	    if (ret) {
		flogerror("intface_setup USBFO_USB_CTL ioctl error:");
		(void)loglasterror();
	    }
	    break;
	case 3:
/*	    Configure USB interface to use combined fiber channels A	*/
/*	    and B.							*/
	    chan = USB_FIBER_CHAN_A|USB_FIBER_CHAN_B;
	    ret = ioctl(im_fd,USBFO_USB_CTL,&chan);
	    if (ret) {
		flogerror("intface_setup USBFO_USB_CTL ioctl error:");
		(void)loglasterror();
	    }
	    break;
	default:
/*	    Nothing yet defined for any other values.			*/
	    break;
	}
}

/*	Given a process name, look up in the database the name of	*/
/*	the log file used by that process. We always return some name,	*/
/*	even if the database access fails for some reason.		*/

static char *getlogfile(char *process_name)
{
	char db_log_entry[100];
	char db_log_item[100];
	char *db_return;
	int ret;

/*	Construct the database name of our logfile.			*/
        sprintf(db_log_entry,"telefocus.logfiles.%s.%s",version,process_name);

/*	Use the dbshow command to access the database. This works even	*/
/*	if our uid is not known to the database because dbshow is.	*/

        ret = dbshow("%s",db_log_entry,db_log_item);
        if(ret == -1) {
/*		If a logfile name is not defined then define one so	*/
/*		we can log a message.					*/
		if(strcmp(log_filename,"") == 0)
		    log_filename = "/tmp/camread_initiallog";
/*		Pick a backup logfile name based on our process name.	*/
		sprintf(db_log_item,"/tmp/%slog",process_name);
		flogerror("Failed to read %s from database",db_log_entry);
        }

/*	Return a copy of the logfile name.				*/
	db_return = strdup_envar(db_log_item);

	return db_return;
}

static char *local_strndup(char *string,int nc)
{
	char *cp;
	char *ci;

	if(nc <= 0) {
	    flogerror("strndup called with byte count %d",nc);
	    nc = 0;
	}
	ci = cp = malloc(nc+1);
	if(ci == NULL) {
	    flogerror("Can't allocate memory (%d bytes) for strndup"
		"\n    Reason: %s",nc+1,strerror(errno));
	    return(ci);
	}
	while((nc > 0) && (*string != '\0')) {
	   *cp = *string;
	   cp++;
	   string++;
	   nc--;
	}
	*cp = '\0';
	return(ci);
}

/*	If we are at the start of a readout (current_pixels==0) or	*/
/*	we are at the end of a readout (current_pixels >= total_pixels)	*/
/*	or sufficient time has elaped since the last report for the	*/
/*	current readout then send a new readout progress report.	*/

static void send_progress_report(int total_pixels,int current_pixels)
{
	static struct timeval tv0,tstart;
	static struct timeval dtv0;
	struct timeval tv1;
	int ret,dt,t_elapse;
	int send = 0;
	int send_display = 0;

	if(current_pixels == 0) {

/*	    Capture the current time.					*/
	    ret = gettimeofday(&tv1,NULL);
	    if(ret < 0) {
		flogerror("Failed gettimeofday %s,%s; %s",
			__FUNCTION__,__LINE__,strerror(errno));
	    }
	    tstart.tv_sec = tv1.tv_sec;
	    tstart.tv_usec = tv1.tv_usec;
	    dtv0.tv_sec = tv1.tv_sec;
	    dtv0.tv_usec = tv1.tv_usec;

/*	    Always send report at start of image.			*/
	    send = 1;

	} else {

	    if(current_pixels >= total_pixels) {
	        current_pixels = total_pixels;
	    }

/*	    Capture the current time.					*/
	    ret = gettimeofday(&tv1,NULL);
	    if(ret < 0) {
		flogerror("Failed gettimeofday %s,%s; %s",
		    __FUNCTION__,__LINE__,strerror(errno));
	    }

/*	    Compute the time elapsed since last report.			*/
/*	    Units of dt are milliseconds.				*/

	    dt = (tv1.tv_sec - tv0.tv_sec)*1000;
	    dt += (tv1.tv_usec - tv0.tv_usec)/1000;

/*	    Send a new report if at least REPORT_INTERVAL millisceonds	*/
/*	    have elapsed since last report or if we are at the end of	*/
/*	    the image.							*/

#define REPORT_INTERVAL 500
	    if(current_pixels >= total_pixels) {
/*	        Always send report at end of image.			*/
	        send = 1;
	    } else if (dt >= REPORT_INTERVAL) {
/*		Report if enough time has elapsed since last report	*/
		send = 1;
	    }

/*	    Compute elapsed time for the image display. If the elapsed	*/
/*	    is greater than the redisplay interval then request a new	*/
/*	    display update.						*/

	    dt = (tv1.tv_sec - dtv0.tv_sec)*1000;
	    dt += (tv1.tv_usec - dtv0.tv_usec)/1000;
	    if((current_pixels < total_pixels) &&
	       (display_interval != 0) && (dt >= display_interval)) {

/*		Request a new display update.				*/
		send_display = 1;
	    }

	}

/*	If requested then send the total number of bytes in the image,	*/
/*	the number of bytes processed so far, and the elapsed time.	*/
	if(send) {
		tv0.tv_sec = tv1.tv_sec;
		tv0.tv_usec = tv1.tv_usec;
		t_elapse = (tv1.tv_sec - tstart.tv_sec)*1000;
		t_elapse += (tv1.tv_usec - tstart.tv_usec)/1000;
		mstart(BROADCAST,our_addr,report_progress_msg);
		mputl(total_pixels);	/* Total pixels in image.	*/
		mputl(current_pixels);	/* Pixels processed so far.	*/
		mputl(t_elapse);	/* Elapsed readout time in ms.	*/
		msend(traffic_fd);	/* Send it			*/
	}

	if(send_display) {

	    dtv0.tv_sec = tv1.tv_sec;
	    dtv0.tv_usec = tv1.tv_usec;
	    if(send_display_newim_msg(msid) < 0) {
		flogerror("Unable to send NEWIM message: %s",
		    strerror(errno));
	    }
	}
}

int image_is_background(struct header_data *hdata)
{
	return 0;
}

/*	Save the current image by accumulating it into the bkAccImg	*/
/*	buffer and then creating a normalized image in bkNormImg.	*/

void save_background(struct header_data *hdata)
{
	unsigned short *from,*to;
	int pixels,bytes;
	register int pn;
	register unsigned int *acc;

/*	This is the first background image if last_was_background is 0	*/
	
	if(last_was_background == 0) {

/*	    Just copy the image and header to both the accumulated	*/
/*	    buffer and the normalized buffer.				*/
	    bkCnt = 1;

	    from = (unsigned short *)mem_image;
            pixels = hdata->image.r*hdata->image.c;
	    bytes = pixels*2;

	    bkAccHeader = bkAccShad;
	    memcpy((void *)bkAccHeader,(void *)hdata,
		sizeof(struct header_data));

	    bkAccImg = bkAccHeader + sizeof(struct header_data);
	    memcpy((void *)bkAccImg,(void *)from,bytes);

	    bkNormHeader = bkNormShad;
	    memcpy((void *)bkNormHeader,(void *)hdata,
		sizeof(struct header_data));
	    bkNormImg = bkNormShad + sizeof(struct header_data);
	    memcpy((void *)bkNormImg,(void *)from,bytes);

	} else {

/*	   We add the new image to the accumulated image and we		*/
/*	   recompute the normalized image.				*/

	    bkCnt++;		/* Number of images added to bkAccImg.	*/

	    from = (unsigned short *)mem_image;
            pixels = hdata->image.r*hdata->image.c;

	    acc = (unsigned int *)bkAccImg;

/*	    The accumulated image is 32-bit unsigned so we can		*/
/*	    accumulate many images.					*/
	    for(pn=0; pn<pixels; pn++) {
		*acc += (unsigned int)(*from);
		acc++;
		from++;
	    }

/*	    Recompute the normalized image. This is just the accumulated*/
/*	    image divided by the number of accumulated images.		*/

	    acc = (unsigned int *)bkAccImg;
	    to = (unsigned short *)bkNormImg;
	    for(pn=0; pn<pixels; pn++) {
		*to = (unsigned short)((*acc)/bkCnt);
		to++;
		acc++;
	    }
	}
	last_was_background = 1;

/*	The image to use for later processing or display is the		*/
/*	normalized image.						*/

	process_image = bkNormImg;
	process_header = bkNormHeader;
	process_buffer = bkNormShad;
	process_shid = bkNormShid;

	return;
}

/*	If there is a normalized background image then subtract it	*/
/*	from the image stored in buffer mem_image.			*/
static void subtract_background(char *mem_image,struct header_data *hdata)
{
	unsigned short *from,*to;
	int pixels;
	register int pn;
	float factor;

	if(subtractBackground == 0) return;
	if(bkCnt <= 0) return;
	factor = headers_match(hdata,(struct header_data *)bkNormHeader);
	if(factor < 0.0) return;

	from = (unsigned short *)bkNormImg;
	to = (unsigned short *)mem_image;
        pixels = hdata->image.r*hdata->image.c;

/*	Subtract the normalized background image from the raw image.	*/
/*	The background image is scaled by the ratio of exposure times.	*/
/*	100 is added to the result to help eliminate pixels less than 0	*/
/*	which are difficult to store back into an unsigned integer.	*/

	for(pn=0; pn<pixels; pn++) {
	    register float diff;
	    diff = 100.0 + (float)(*to) - (factor * (float)(*from));
	    if (diff < 0.0) diff = 0.0;		/* Truncate negatives.	*/
	    *to++ = (unsigned short)diff;
	    from++;
	}

}

/*	Test to see if the two images described by the two headers are	*/
/*	the same readout window. If they are not then -1.0 is returned.	*/
/*	If they are the same then the ratio of the exposure times is	*/
/*	returned.							*/
float headers_match(struct header_data *hdata,struct header_data *bk)
{
	if(memcmp((void *)(&hdata->image),(void *)(&bk->image),
	    sizeof(struct window)) != 0) {
		return (-1.0);
	}
	if(memcmp((void *)(&hdata->window),(void *)(&bk->window),
	    sizeof(struct window)) != 0) {
		return (-1.0);
	}
	return(1.0);
/*	return((float)hdata->exposure/(float)bk->exposure); */

}
